{ TelesisEnh_BaseReg.pas                          SearchTag: ~~EnhComponents~~

 Copyright 1999-2001 - Telesis Computing Pty Ltd.  Australia.
Contact: Geoff Worboys - geoff@telesiscomputing.com.au }

{: @summary Registration unit for components defined via TelesisEnh_Base.pas.
<br>
@desc See <See Unit=TelesisEnh_Base>.
<br>
<br>
<strong>History</strong><br>
<br>
13-Aug-2001 - Eliminated component save-to-file feature for versions prior
to VCL50 as the required functions were not available.<br>
<br>
05-Aug-2001 - Support for Delphi6 (courtesy Luiz Alves).<br>
<br>
27-Feb-2001 - Documentation.<br>
<br>
<br>
<strong>Author: </strong>Geoff Worboys - Telesis Computing Pty Ltd.<br>
<br>}
unit TelesisEnh_BaseReg;

{$INCLUDE Telesis.inc}


interface

uses SysUtils,
  Classes,
  {$IFDEF __VCL60_OR_LATER__}
  DesignIntf, DesignMenus, DesignEditors, DesignWindows,
 {$ELSE}
  DsgnIntf,
 {$ENDIF}
  TypInfo;

type

  TTelesisEnh_VersionProperty = class(TStringProperty)
    function GetAttributes: TPropertyAttributes; override;
    function GetValue: string; override;
    procedure SetValue(const Value: string); override;
    procedure Edit; override;
  end;

  TMaskEnhProperty = class(TPropertyEditor)
  public
    procedure Edit; override;
    function GetAttributes: TPropertyAttributes; override;
    function GetValue: string; override;
    procedure SetValue(const Value: string); override;
  end;

  TVariantEnhProperty = class(TPropertyEditor)
  public
    function AllEqual: Boolean; override;
    function GetValue: string; override;
    procedure SetValue(const Value: string); override;
  end;


procedure Register;


implementation

// Bring in the palette resources
{$R Telesis_EditEnh.res}
{$R Telesis_SpinEditEnh.res}
{$R Telesis_ComboEnh.res}
{$R Telesis_MemoEnh.res}
{$R Telesis_CheckBoxEnh.res}
{$R Telesis_CheckListEnh.RES}
{$R Telesis_DateTimeEnh.res}


uses
  Windows,
  Forms,
  Dialogs,
  Clipbrd,
  Registry,
  Telesis_Constants,
  Telesis_Locale,
  Telesis_MaskEnh,
  Telesis_MaskEnhEditor,
  Telesis_EditEnh,
  Telesis_SpinEditEnh,
  Telesis_ComboEnh,
  Telesis_MemoEnh,
  Telesis_CheckBoxEnh,
  Telesis_CheckListEnh,
  Telesis_DateTimeEnh,
  Telesis_Utils;

procedure Register;
begin
  RegisterPropertyEditor(TypeInfo(TTelesisEnh_Version), nil, '',
    TTelesisEnh_VersionProperty);
  // Base Enh Component Registrations
  RegisterComponents('Enhanced', [
    // Specific/Individual Components...
    TEditEnh,
    TSpinEditEnh,
    TComboBoxEnh,
    TMemoEnh,
    TCheckBoxEnh,
    TCheckListBoxEnh,
    TDateTimeEnh,
    TDateTimePnlEnh]);
  RegisterPropertyEditor(TypeInfo(TMaskEnhString),
    TCustomEditEnh,
    'EditMask',
    TMaskEnhProperty);
  RegisterPropertyEditor(TypeInfo(Variant),
    TCustomEditEnh,
    '',
    TVariantEnhProperty);
end;



function TTelesisEnh_VersionProperty.GetAttributes: TPropertyAttributes;
begin
  // Done show on multi-select, can only save/load one component at a time.
  Result := inherited GetAttributes + [paReadOnly, paDialog] - [paMultiSelect];
end;

function TTelesisEnh_VersionProperty.GetValue: string;
begin
  Result := TelesisEnh_Version_String;
end;

procedure TTelesisEnh_VersionProperty.SetValue(const Value: string);
begin
  inherited SetValue('');
end;

procedure TTelesisEnh_VersionProperty.Edit;
var
  t, s, f: string;
begin
  t := 'About: ' + TelesisEnh_Product_Name;
  {$IFDEF __VCL50_OR_LATER__}
  s := 'Version: ' + TelesisEnh_Version_String + #13#10 +
    'Date: ' + DateTimeToStr(TelesisEnh_Version_Date) + #13#10#13#10 +
    TelesisEnh_Copyright_String + #13#10 +
    'Contact: ' + TelesisEnh_Email + #13#10#13#10 +
    TelesisEnh_SaveLoadMsg;
  if GetComponent(0) is TComponent then
    f := TComponent(GetComponent(0)).Name + '.rc'
  else
    f := 'Component.rc';
  case MessageBox(0, PChar(s), PChar(t),
    MB_YESNOCANCEL or MB_ICONINFORMATION or MB_DEFBUTTON3) of
    IDNO:
      begin
        // Save previous entry
        if Clipboard.HasFormat(CF_TEXT) then
          s := Clipboard.AsText
        else
          s := '';
        try
          t := ReadStringFromFile(f, true);
          if t <> '' then
          begin
            Clipboard.AsText := t;
            Designer.PasteSelection;
          end;
        finally
          // Restore saved entry...
          if s <> '' then
            Clipboard.AsText := s;
        end;
      end;
    IDYES:
      begin
        // Save previous entry
        if Clipboard.HasFormat(CF_TEXT) then
          s := Clipboard.AsText
        else
          s := '';
        try
          Designer.CopySelection;
          if Clipboard.HasFormat(CF_TEXT) then
            t := Clipboard.AsText
          else
            t := '';
          if t <> '' then
            SaveStringToFile(t, f, true);
        finally
          // Restore saved entry...
          if s <> '' then
            Clipboard.AsText := s;
        end;
      end;
  end;
 {$ELSE}
  s := 'Version: ' + TelesisEnh_Version_String + #13#10 +
    'Date: ' + DateTimeToStr(TelesisEnh_Version_Date) + #13#10#13#10 +
    TelesisEnh_Copyright_String + #13#10 +
    'Contact: ' + TelesisEnh_Email;
    MessageBox(0, PChar(s), PChar(t), MB_OK or MB_ICONINFORMATION);
 {$ENDIF}
end;


function TMaskEnhProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog];
end;

function TMaskEnhProperty.GetValue: string;
var
  EditEnh: TCustomEditEnh;
begin
  if GetComponent(0) is TCustomEditEnh then
  begin
    EditEnh := GetComponent(0) as TCustomEditEnh;
    if assigned(EditEnh.Mask) then
      Result := EditEnh.Mask.Mask
    else
      Result := '';
  end;
end;

procedure TMaskEnhProperty.SetValue(const Value: string);
var
  EditEnh: TCustomEditEnh;
begin
  if GetComponent(0) is TCustomEditEnh then
  begin
    EditEnh := GetComponent(0) as TCustomEditEnh;
    EditEnh.SetEditMaskEx(Value);
  end;
end;

procedure TMaskEnhProperty.Edit;
var
  EditEnh: TCustomEditEnh;
begin
  if GetComponent(0) is TCustomEditEnh then
  begin
    EditEnh := GetComponent(0) as TCustomEditEnh;
    with TMaskEnhEditorForm.Create(Application) do
    try
      RegistryKey := Telesis_BaseRegKey + '\MaskEnh';
      AssignMask(EditEnh.Mask);
      if ShowModal = idOK then
      begin
        EditEnh.SetEditMaskEx(InputTestEB.GetEditMaskEx);
        Self.Designer.Modified;
      end;
    finally
      Free;
    end;
  end;
end;


function TVariantEnhProperty.AllEqual: Boolean;
var
  I: Integer;
  V: Variant;
begin
  Result := False;
  if PropCount > 1 then
  begin
    V := GetVarValue;
    for I := 1 to PropCount - 1 do
      if GetVarValueAt(I) <> V then Exit;
  end;
  Result := True;
end;

function TVariantEnhProperty.GetValue: string;
begin
  Result := GetVarValue;
end;

procedure TVariantEnhProperty.SetValue(const Value: string);
begin
  SetVarValue(Value);
end;

end.