{ Telesis_Constants.pas                           SearchTag: ~~EnhComponents~~

 Copyright 1999-2006 - Telesis Computing Pty Ltd.  Australia.
Contact: Geoff Worboys - geoff@telesiscomputing.com.au }

{: @summary Various constants (not locale specific) as well as some version
information functions.
<br>
@desc Also see locale specific constants in <See Unit=Telesis_Locale>.
<br>
<br>
<strong>History</strong><br>
<br>
26-Oct-2007 - Version Update.<br>
<br>
01-Aug-2001 - Documentation.<br>
<br>
<br>
<strong>Author: </strong>Geoff Worboys - Telesis Computing Pty Ltd.<br>
<br>}
unit Telesis_Constants;

{$INCLUDE Telesis.inc}

interface

uses SysUtils,
  Windows,
  Classes;

const
  {: Version information as a integer.}
  TelesisEnh_Version: integer =  20020005;
  //                            |||||^^^^ Build
  //                            ||^^^ Minor
  //                            ^^ Major
  {: Version status string.}
  TelesisEnh_Version_Status: string = '';

  {: Version date information - day of month (1..31).}
  TelesisEnh_Version_Day: smallint = 31;
  {: Version date information - month number (1..12).}
  TelesisEnh_Version_Month: smallint = 7; // July
  {: Version date information - four digit year.}
  TelesisEnh_Version_Year: smallint = 2011;

  {: EnhComponents support contact email address.}
  TelesisEnh_Email: string = 'enh.info@telesiscomputing.com.au';

  {: Base registry key where Telesis Computing installation data is stored.}
  Telesis_BaseRegKey: string = 'Software\Telesis Computing';

  {: EnhComponents Product descriptive name.}
  TelesisEnh_Product_Name: string = 'Telesis Enhanced Components';
  {: EnhComponents Product copyright string.}
  TelesisEnh_Copyright: string = ' Copyright 1999-2011';
  {: EnhComponents Product author.}
  TelesisEnh_Author: string = 'Telesis Computing Pty Ltd';
  {: EnhComponents Country of Origin.}
  TelesisEnh_Location: string = 'Australia';


  {: @summary Name of environment variable carrying the session name.
  <br>
  @desc Microsoft Terminal Server supports multiple sessions per server.  If
  an application need to identify a particular instance, it must not only
  supply the workstation name, but also the session identifier.  There is
  probably an API to provide this information, but it is also available from
  an environment variable 'WINSTATIONNAME' in old terminal servers and
  from SESSIONNAME in newer (2003 and later) terminal servers.}
  MTS_SessionEnvironmentVar: string = 'WINSTATIONNAME';
  MTS_SessionEnvironmentVar2: string = 'SESSIONNAME';
  {: @summary Default session name.
  <br>
  @desc If no session environment variable is provided (see
  MTS_SessionEnvironmentVar), then this string is used
  to describe the default (CONSOLE) session.}
  SessionDefaultName: string = 'CONSOLE';


  // RELATED ITEMS

function TelesisEnh_Version_Major: smallint;
function TelesisEnh_Version_Minor: smallint;
function TelesisEnh_Version_Build: smallint;
function TelesisEnh_Version_String: string;
function TelesisEnh_Version_Date: TDateTime;
function TelesisEnh_Copyright_String: string;

type
  {: Specially defined string type for reporting version information as
  a property.}
  TTelesisEnh_Version = type string;

  {: Specially defined string type for internal resource processing.}
  TTelesisEnh_Resource = type string;


const
{: @summary EditEnh Validation Error Codes.
<br>
@desc Error codes provided in MaskEnh and EditEnh validation processing, plus
some additional error codes used by DateTime and SpinEdit controls.
<br>
@enum E_EDITENH_NOERROR (and E_MASKENH_NOERROR)
Value = 0, signifies no error.
<br>
@enum E_MASKENH_CHAR_INVALID
Insert/Replace character invalid for position.
<br>
@enum E_MASKENH_TRUNCATED
Input string was too long for mask.
<br>
@enum E_MASKENH_INVALID
Input string does not fit mask requirements.
<br>
@enum E_MASKENH_INVALIDTRUNC
Input string was truncated, or some characters ignored due to invalid
mask match.
<br>
@enum E_MASKENH_CHAR_TRUNCATED
Insert/Replace attempted beyond end of mask.
<br>
@enum E_EDITENH_USER_ERROR (and E_MASKENH_USERERROR)
Base from which to create custom error codes within EditEnh.OnValidateEdit.
eg.<br>
<br>
const<br>
  E_MY_OWN_ERROR = E_EDITENH_USER_ERROR + 1;
<br>
}
  E_EDITENH_NOERROR = 0;
  E_MASKENH_NOERROR = E_EDITENH_NOERROR; // for backwards compatibility
  E_MASKENH_CHAR_INVALID = 1;
  E_MASKENH_TRUNCATED = 2;
  E_MASKENH_INVALID = 3;
  E_MASKENH_INVALIDTRUNC = 4;
  E_MASKENH_CHAR_TRUNCATED = 5;

  E_COMBOENH_INVALID_PARENT = 21;

  E_DATEENH_VALUE_NOT_FOUND = 41;    // internal use only
  //E_DATEENH_DATETIME_IS_SET = 42;    // internal use only - no longer used
  E_DATEENH_INVALID_LAYOUT = 43;     // invalid or blank Layout string
  E_DATEENH_INPUT_TOO_SHORT = 44;    // required input values not completed
  E_DATEENH_NO_AM_OR_PM = 45;        // no am/pm input and 'A' specified
  E_DATEENH_INVALID_AM_TIME = 46;    // time invalid for AM entry
  E_DATEENH_INVALID_PM_TIME = 47;    // time invalid for PM entry
  E_DATEENH_INVALID_DATE = 48;       // invalid date input, eg wrong # of days
  E_DATEENH_MISMATCHED_DAYNAME = 49; // input of day name did not match date
  E_DATEENH_INVALID_DAY_DATE = 50;   // the day date number was invalid
  E_DATEENH_INVALID_MONTH = 51;      // the month number or str was invalid
  E_DATEENH_INVALID_YEAR = 52;       // the year number was invalid
  E_DATEENH_INVALID_HOUR = 53;       // the hour number was invalid
  E_DATEENH_INVALID_MINUTE = 54;     // the minute number was invalid
  E_DATEENH_INVALID_SECOND = 55;     // the second number was invalid
  E_DATEENH_INVALID_POINTS = 56;     // the points of seconds number was invalid
  E_DATEENH_INVALID_AMPM = 57;       // the am/pm string was invalid

  E_EDITENH_USER_ERROR = 301;
  E_MASKENH_USERERROR = E_EDITENH_USER_ERROR; // for backwards compatibility


function GetEditEnhErrorMsg( ErrCode: integer ): string;

implementation

uses Telesis_Locale;

{: Returns EnhComponents major version information.}
function TelesisEnh_Version_Major: smallint;
begin
  Result := smallint( (TelesisEnh_Version div 10000000) );
end;

{: @summary Returns EnhComponents minor version information.
<br>
@desc By convention, an Odd value indicates a testing/beta release.}
function TelesisEnh_Version_Minor: smallint;
var
  tmpVer: integer;
begin
  tmpVer := TelesisEnh_Version;
  tmpVer := tmpVer - ((tmpVer div 10000000) * 10000000);
  Result := smallint( (tmpVer div 10000) );
end;

{: @summary Build number of the current EnhComponent version.
<br>
@desc Value is reset with every change in minor version.}
function TelesisEnh_Version_Build: smallint;
var
  tmpVer: integer;
begin
  tmpVer := TelesisEnh_Version;
  tmpVer := tmpVer - ((tmpVer div 10000) * 10000);
  Result := smallint( tmpVer );
end;

{: @summary Full EnhComponents version details returned as a string.
<br>
@desc String returned as "Major.Minor.Build.Status".}
function TelesisEnh_Version_String: string;
begin
  Result := IntToStr(TelesisEnh_Version_Major) + '.' +
    IntToStr(TelesisEnh_Version_Minor) + '.' +
    IntToStr(TelesisEnh_Version_Build);
  if TelesisEnh_Version_Status <> '' then
    Result := Result + '.' + TelesisEnh_Version_Status;
end;

{: EnhComponents Version date.}
function TelesisEnh_Version_Date: TDateTime;
begin
  Result := EncodeDate(TelesisEnh_Version_Year,
    TelesisEnh_Version_Month,
    TelesisEnh_Version_Day);
end;

{: EnhComponents Author and Copyright string.}
function TelesisEnh_Copyright_String: string;
begin
  Result := TelesisEnh_Copyright + ' - ' + TelesisEnh_Author;
end;


function GetEditEnhErrorMsg( ErrCode: integer ): string;
begin
  case ErrCode of
    E_EDITENH_NOERROR:             Result := ENHERR_EDITENH_NOERROR;
    E_MASKENH_CHAR_INVALID:        Result := ENHERR_MASKENH_CHAR_INVALID;
    E_MASKENH_TRUNCATED:           Result := ENHERR_MASKENH_TRUNCATED;
    E_MASKENH_INVALID:             Result := ENHERR_MASKENH_INVALID;
    E_MASKENH_INVALIDTRUNC:        Result := ENHERR_MASKENH_INVALIDTRUNC;
    E_MASKENH_CHAR_TRUNCATED:      Result := ENHERR_MASKENH_CHAR_TRUNCATED;
    E_DATEENH_VALUE_NOT_FOUND:     Result := ENHERR_DATEENH_VALUE_NOT_FOUND;
    //E_DATEENH_DATETIME_IS_SET:     Result := ENHERR_DATEENH_DATETIME_IS_SET;
    E_DATEENH_INVALID_LAYOUT:      Result := ENHERR_DATEENH_INVALID_LAYOUT;
    E_DATEENH_INPUT_TOO_SHORT:     Result := ENHERR_DATEENH_INPUT_TOO_SHORT;
    E_DATEENH_NO_AM_OR_PM:         Result := ENHERR_DATEENH_NO_AM_OR_PM;
    E_DATEENH_INVALID_AM_TIME:     Result := ENHERR_DATEENH_INVALID_AM_TIME;
    E_DATEENH_INVALID_PM_TIME:     Result := ENHERR_DATEENH_INVALID_PM_TIME;
    E_DATEENH_INVALID_DATE:        Result := ENHERR_DATEENH_INVALID_DATE;
    E_DATEENH_MISMATCHED_DAYNAME:  Result := ENHERR_DATEENH_MISMATCHED_DAYNAME;
    E_DATEENH_INVALID_DAY_DATE:    Result := ENHERR_DATEENH_INVALID_DAY_DATE;
    E_DATEENH_INVALID_MONTH:       Result := ENHERR_DATEENH_INVALID_MONTH;
    E_DATEENH_INVALID_YEAR:        Result := ENHERR_DATEENH_INVALID_YEAR;
    E_DATEENH_INVALID_HOUR:        Result := ENHERR_DATEENH_INVALID_HOUR;
    E_DATEENH_INVALID_MINUTE:      Result := ENHERR_DATEENH_INVALID_MINUTE;
    E_DATEENH_INVALID_SECOND:      Result := ENHERR_DATEENH_INVALID_SECOND;
    E_DATEENH_INVALID_POINTS:      Result := ENHERR_DATEENH_INVALID_POINTS;
    E_DATEENH_INVALID_AMPM:        Result := ENHERR_DATEENH_INVALID_AMPM;
  else
    Result := ENHERR_UNKNOWN_EDIT_ERROR;
  end;
end;

end.