{ Telesis_IB_BitCheckBoxEnh.pas                   SearchTag: ~~EnhComponents~~

 Copyright 1999-2001 - Telesis Computing Pty Ltd.  Australia.
Contact: Geoff Worboys - geoff@telesiscomputing.com.au }

{: @summary Specialised variation of a checkbox used to control the value of a
single bit from an IBO field.
<br>
@desc ...
<br>
<br>
<strong>History</strong><br>
<br>
22-Jun-2006 - AdvSearchData was not being cleared (problems with updates
not listed in history from a few months ago).<br>
<br>
7-Feb-2005 - Search mode was not working properly.<br>
<br>
29-May-2002 - Published some properties and events from base class.<br>
<br>
17-Dec-2001 - Published some properties and events from base class.<br>
<br>
23-Aug-2001 - Additional csDestroying checks.<br>
<br>
02-Aug-2001 - Documentation.<br>
<br>
<br>
<strong>Author: </strong>Geoff Worboys - Telesis Computing Pty Ltd.<br>
<br>}
unit Telesis_IB_BitCheckBoxEnh;

{$INCLUDE IB_Directives.inc}

interface

uses SysUtils,
  Messages,
  Windows,
  Classes,
  Graphics,
  Menus,
  Controls,
  Forms,
  StdCtrls,
  Mask,
  ExtCtrls,
  Telesis_CheckBoxEnh,
  IB_Session,
  IB_Components,
  IB_Controls,
  IB_FieldDataLink,
  Telesis_Bits,
  IBC_Label;

type

  {$DEFINE IBC_BitCheckBoxEnh}
  {$DEFINE IB_UserClass}
  {$DEFINE IB_FieldControl}

  TIB_CustomBitCheckBoxEnh = class(TCustomCheckBoxEnh)
    {$DEFINE IB_AutoLabel}
    {$I IB_WinControl.INT}
    {$UNDEF IB_AutoLabel}
  private
    FBitPosition: TBitPosition;
    FBitCompare: TBitComparison;
    FFollowerBit: smallint;
    FPreventGrayed: boolean;
    FInternalUpd: integer;
    FInUpdate: boolean;
    //??FDisplayLabelAsCaption: Boolean;
    //??procedure SetDisplayLabelAsCaption(AValue: Boolean);
    function GetFieldState: TCheckBoxState; virtual;
    procedure SetBitPosition(AValue: TBitPosition);
    procedure PrepareBitsSQL(Sender: TIB_DataLink; ADataSource: TIB_DataSource);
    //function GetParamName: string;
    //procedure SetParamName(AVal: string);

  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure Loaded; override;
    function IsMaskStored: boolean; override;
    function GetDBDisplayField: TObject; override;
    function IBG_GetAutoLabel: TObject; override;
    procedure Toggle; override;
    procedure KeyPress(var Key: Char); override;

    //property ParamName: string read GetParamName write SetParamName;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Click; override;

    property State stored false;

    property BitPosition: TBitPosition read FBitPosition write SetBitPosition
      default 0;
    property BitCompare: TBitComparison read FBitCompare write FBitCompare
      default bcAtLeast;
    {: The FollowerBit, if defined within constraints of TBitPosition, will
    cause the specified bit to follow the state of the defined BitPosition
    when its state is changed via interaction with this control.  By default
    this setting is -1, which means that there is no follower bit.  This is
    property useful if two bit flags are interdependant. }
    property FollowerBit: smallint read FFollowerBit write FFollowerBit
      default -1;
    property PreventGrayed: boolean read FPreventGrayed write FPreventGrayed
      default false;
    //??property DisplayLabelAsCaption: boolean read FDisplayLabelAsCaption write
    //??  SetDisplayLabelAsCaption default false;

  published

  end;                                  // TIB_CustomBitCheckBoxEnh


  TIB_BitCheckBoxEnh = class(TIB_CustomBitCheckBoxEnh)
    {$DEFINE IB_AutoLabel}
    {$I IB_WinControl.PBL}
    {$UNDEF IB_AutoLabel}
  published
    //property ParamName;
    property BitCompare;
    property BitPosition;
    property FollowerBit;
    property PreventGrayed;
    //?? property DisplayLabelAsCaption;
    property Ctl3DShallow;
    //property FillWithTextColor;
    {$IFNDEF HELPSCAN}
    property Alignment;
    //??? property Caption;
    property ActiveWhenReadOnly;
    property AutoDropDown;
    property DropDownColor;
    property DropDownFont;
    property DropDownFontAuto;
    property Checked;
    property State;
    property GlyphAlign;
    property GlyphSize;
    property OnButton1;
    property OnButton2;
    property OnButton3;
    property OnDrawButton;
    property AfterValidateEdit;
    property OnValidateEdit;
    property OnValidateError;
    property OnGlyphClick;
    property OnDrawGlyph;
    {$ENDIF}
  end;                                  // TIB_BitCheckBoxEnh



implementation

{$DEFINE IB_AutoLabel}
{$I IB_WinControl.IMP}
{$UNDEF IB_AutoLabel}


constructor TIB_CustomBitCheckBoxEnh.Create(AOwner: TComponent);
begin
  FIB_FieldDataLink := TIB_FieldDataLink.Create(Self);
  FAutoLabel := TIB_AutoLabel.Create(Self);
  FInternalUpd := 0;
  FInUpdate := false;
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csReplicatable];
  DataLink.Control := Self;
  DataLink.OnStateChanged := StateChanged;
  DataLink.OnDataChange := DataChange;
  DataLink.OnUpdateData := Self.UpdateData;
  OnPrepareSQL := PrepareBitsSQL;
  FBorderStyle := bsSingle;
  //??FDisplayLabelAsCaption := false;
  FBitPosition := 0;
  FBitCompare := bcAtLeast;
  FFollowerBit := -1;
  FPreventGrayed := false;
end;

destructor TIB_CustomBitCheckBoxEnh.Destroy;
begin
  FAutoLabel.Free;
  FAutoLabel := nil;
  DataSource := nil;
  DataLink.Control := nil;
  DataLink.OnStateChanged := nil;
  DataLink.OnDataChange := nil;
  DataLink.OnUpdateData := nil;
  inherited Destroy;
end;

procedure TIB_CustomBitCheckBoxEnh.Loaded;
begin
  inherited Loaded;
  SysStateChanged;
  SysDataChange(nil);
  RemakeAutoLabel;
end;

// Make sure ControlIsModified is set true if required...
//  NOTE: see procedure TIB_RadioGroup.Change; for example.
//  NOTE: this will produce multiple datachange events in some cases!!!!!!!!

procedure TIB_CustomBitCheckBoxEnh.Click;
begin
  with DataLink do
    if Editing then ControlIsModified := true;
  inherited Click;
  DataLink.UpdateRecord;
end;

procedure TIB_CustomBitCheckBoxEnh.SysStateChanged;
begin
  // No point doing anything if we disappearing anyway.
  if csDestroying in ComponentState then
    Exit;
  with DataLink do
  begin
    if ColorScheme then
    begin
      inherited Color := Color;
    end;
    inherited AllowGrayed := ((State = dssSearch) or IsParamEdit) and not PreventGrayed;
    //??if FDisplayLabelAsCaption and Assigned(Field) then
    //??  Caption := Field.DisplayName;
    UpdateTabStop;
  end;
end;

procedure TIB_CustomBitCheckBoxEnh.SysDataChange;
begin
  // No point doing anything if we disappearing anyway.
  if csDestroying in ComponentState then
    Exit;
  State := GetFieldState;
end;

procedure TIB_CustomBitCheckBoxEnh.SysUpdateData;
var
  Hold: integer;
begin
  if FInUpdate then
    Exit;
  with DataLink do
  try
    FInUpdate := true;
    if Self.State = cbGrayed then
    begin
      FieldText := '';
      if (State = dssSearch) or SearchAlways then
        AdvSearchData := '';
    end
    else
    begin
      if Length(FieldText) > 0 then
      begin
        try
          Hold := StrToInt(FieldText)
        except
          Hold := 0;
        end;
      end
      else
        Hold := 0;
      if Checked then
        BitTurnOn(Hold, FBitPosition)
      else
        BitTurnOff(Hold, FBitPosition);
      // If change due to user interaction (from Toggle call)
      // then check the FollowerBit...
      if (FInternalUpd = 1) and (DataLink.State in [dssEdit,dssInsert]) and (
        (FollowerBit >= Ord(Low(TBitPosition))) and
        (FollowerBit <= Ord(High(TBitPosition))) ) then
      begin
        if Checked then
          BitTurnOn(Hold, TBitPosition(FollowerBit))
        else
          BitTurnOff(Hold, TBitPosition(FollowerBit));
      end;
      FieldText := IntToStr(Hold);
      if (State = dssSearch) or SearchAlways then
        AdvSearchData := FieldText; // for special prepare processing
    end;
  finally
    FInUpdate := false;
  end;
end;

{//?? procedure TIB_CustomBitCheckBoxEnh.SetDisplayLabelAsCaption(AValue: boolean);
begin
  if FDisplayLabelAsCaption <> AValue then
  begin
    FDisplayLabelAsCaption := AValue;
    if FDisplayLabelAsCaption then
    begin
      SysStateChanged;
    end;
  end;
end;}

function TIB_CustomBitCheckBoxEnh.GetFieldState: TCheckBoxState;
var
  Hold: integer;
begin
  with DataLink do
  begin
    if not Assigned(Field) then
      Result := cbGrayed
    else
    begin
      if Length(FieldText) > 0 then
      begin
        try
          Hold := StrToInt(FieldText)
        except
          Hold := 0;
        end;
        if BitIsSet(Hold, BitPosition) then
          Result := cbChecked
        else
          Result := cbUnchecked;
      end
      else
        Result := cbGrayed;
    end;
  end;
end;

function TIB_CustomBitCheckBoxEnh.IsMaskStored: boolean;
begin
  // IBO does not store masks with controls, only with the data definition
  Result := false;
end;

function TIB_CustomBitCheckBoxEnh.GetDBDisplayField: TObject;
begin
  Result := DataLink.Field;
end;

function TIB_CustomBitCheckBoxEnh.IBG_GetAutoLabel: TObject;
begin
  Result := FAutoLabel;
end;

procedure TIB_CustomBitCheckBoxEnh.Toggle;
begin
  Inc(FInternalUpd);
  try
    if DataLink.IsParamEdit or DataLink.Modify then
    begin
      inherited Toggle;
      DataLink.ControlIsModified := true;
    end;
    SysUpdateData(Field);
  finally
    Dec(FInternalUpd);
  end;
end;

procedure TIB_CustomBitCheckBoxEnh.KeyPress(var Key: Char);
begin
  inherited KeyPress(Key);
  case Key of
    #8, ' ':
      if DataLink.IsParamEdit or DataLink.Modify then
      begin
        DataLink.ControlIsModified := true;
      end;
    #27: SysDataChange(nil);
  end;
end;

procedure TIB_CustomBitCheckBoxEnh.CMEnter(var Message: TCMEnter);
begin
  DataLink.SetFocus;
  inherited;
end;

procedure TIB_CustomBitCheckBoxEnh.CMExit(var Message: TCMExit);
begin
  try
    DataLink.UpdateRecord;
  except
    SetFocus;
    raise;
  end;
  inherited;
end;

procedure TIB_CustomBitCheckBoxEnh.CreateParams(var Params: TCreateParams);
const
  BorderStyles: array[TBorderStyle] of integer = (0, WS_BORDER);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := longint(Style) or longint(BorderStyles[FBorderStyle]);
    if NewStyleControls and Ctl3D and (FBorderStyle = bsSingle) then
    begin
      Style := Style and not WS_BORDER;
      ExStyle := ExStyle or WS_EX_CLIENTEDGE;
    end;
  end;
end;

procedure TIB_CustomBitCheckBoxEnh.SetBitPosition(AValue: TBitPosition);
begin
  if AValue <> FBitPosition then
  begin
    FBitPosition := AValue;
    SysDataChange(Field);
  end;
end;

procedure TIB_CustomBitCheckBoxEnh.PrepareBitsSQL(Sender: TIB_DataLink;
  ADataSource: TIB_DataSource);
var
  Hold: integer;
  tmpstr: string;
begin
  with DataLink do
  begin
    if (State = dssSearch) and (Length(FieldText) > 0) then
      tmpstr := FieldText
    else
      tmpstr := AdvSearchData;
    if Length(tmpstr) > 0 then
    begin
      try
        Hold := StrToInt(tmpstr);
      except
        Hold := 0;
      end;
      if Hold = 0 then
      begin
        BitTurnOn(Hold, FBitPosition);
        Dataset.SQLWhereItems.Add(GetBitsWhereSQL(FieldName, Hold, bcNoneOf));
      end
      else
        Dataset.SQLWhereItems.Add(GetBitsWhereSQL(FieldName, Hold, BitCompare));
    end;
  end;
end;

//function TIB_CustomBitCheckBoxEnh.GetParamName: string;
//begin
//  Result := DataLink.ParamName;
//end;
//
//procedure TIB_CustomBitCheckBoxEnh.SetParamName(AVal: string);
//begin
//  DataLink.ParamName := AVal;
//end;

end.