{ Telesis_IB_BitCheckListEnh.pas                  SearchTag: ~~EnhComponents~~

 Copyright 1999-2001 - Telesis Computing Pty Ltd.  Australia.
Contact: Geoff Worboys - geoff@telesiscomputing.com.au }

{: @summary IBO aware variation of a CheckList used to set/clear bits in a
field.
<br>
@desc ...
<br>
<br>
<strong>History</strong><br>
<br>
22-Jun-2006 - AdvSearchData was not being cleared (problems with updates
not listed in history from a few months ago).<br>
<br>
20-Jun-2005 - Make DataLink available (mainly for ControlIsModified).<br>
<br>
27-Feb-2001 - Documentation.<br>
<br>
23-Aug-2001 - Additional csDestroying checks<br>
<br>
<br>
<strong>Author: </strong>Geoff Worboys - Telesis Computing Pty Ltd.<br>
<br>}
unit Telesis_IB_BitCheckListEnh;

{$INCLUDE IB_Directives.inc}

interface

uses SysUtils,
  Messages,
  Windows,
  Classes,
  Graphics,
  Menus,
  Controls,
  Forms,
  StdCtrls,
  Mask,
  ExtCtrls,
  IB_Session,
  IB_Components,
  IB_Controls,
  IB_FieldDataLink,
  Telesis_Bits,
  Telesis_CheckListEnh,
  IBC_Label;

type
  {$DEFINE IBC_BitCheckListEnh}
  {$DEFINE IB_UserClass}
  {$DEFINE IB_FieldControl}

  // I have to use IB_Control not IB_WinControl because the default IB_WinControl
  // stuff does some BorderStyle handling which is inappropriate.  Because of
  // this I need to include the bits of IB_WinControl that I do want...
  TIB_CustomBitCheckListEnh = class(TCustomCheckListBoxEnh)
    {$DEFINE IB_AutoLabel}
    {$DEFINE IB_WinControl}
    {$I IB_Control.INT}
    {$UNDEF IB_WinControl}
    {$UNDEF IB_AutoLabel}
    // From IB_WinControl.INT
  private
    FItemsWereLoaded: boolean;
    FItemListAutoLoad: boolean;
    FTabStopAutomation: boolean;
    FDynamicFieldUpdate: boolean;
    FPreventGrayed: boolean;
    function StoreTabStop: boolean;
    procedure SetTabStopAutomation( AValue: boolean );
    procedure UpdateTabStop;
    function GetParamName: string;
    procedure SetParamName(AVal: string);
    procedure SetPreventEditing(AValue: Boolean);
    procedure SetPreventInserting(AValue: Boolean);
    procedure SetPreventSearching(AValue: Boolean);
    function GetPreventEditing: Boolean;
    function GetPreventInserting: Boolean;
    function GetPreventSearching: Boolean;
    procedure CMEnter(var Message: TCMEnter); message CM_ENTER;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure CMEnabledChanged(var Message: TMessage); message
      CM_ENABLEDCHANGED;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
  protected
    function GetReadOnly: boolean; virtual;
    procedure SetReadOnly(AValue: Boolean); virtual;
    procedure SysUpdateData(IB_Field: TIB_Column); virtual;
    procedure UpdateData(Sender: TIB_DataLink;
      IB_DataSource: TIB_DataSource;
      IB_Field: TIB_Column);
  public
    function CanModify: Boolean; virtual;
  public
    property ItemListAutoLoad: boolean read FItemListAutoLoad
     write FItemListAutoLoad default true;
    property ReadOnly: boolean read GetReadOnly
    write SetReadOnly
      default false;
    property PreventEditing: boolean read GetPreventEditing
    write SetPreventEditing
      default false;
    property PreventInserting: boolean read GetPreventInserting
    write SetPreventInserting
      default false;
    property PreventSearching: boolean read GetPreventSearching
    write SetPreventSearching
      default false;
    property Ctl3D default true;
    property TabStopAutomation: boolean read FTabStopAutomation
                                        write SetTabStopAutomation
                                        default false;
    property TabStop stored StoreTabStop;

    // My Own Stuff...
  private
    FBitPosition: TBitPosition;
    FBitCompare: TBitComparison;
    FItemListField: string;
    FItemListLink: TIB_FieldDataLink;
    procedure SetBitPosition(AValue: TBitPosition);
    procedure PrepareBitsSQL(Sender: TIB_DataLink; ADataSource: TIB_DataSource);
    function GetItemListDataField: string;
    procedure SetItemListDataField(AValue: string);
    function GetItemListDataSource: TIB_DataSource;
    procedure SetItemListDataSource(AValue: TIB_DataSource);

  protected
    procedure Loaded; override;
    procedure ClickCheck; override;
    procedure LoadItemList; virtual;
    function AreItemsStored: boolean; virtual;
    procedure PreparedChanged(Sender: TIB_DataLink; IB_DataSource:
      TIB_DataSource); virtual;

    property Items stored AreItemsStored;
    property BitPosition: TBitPosition read FBitPosition write SetBitPosition
      default 0;
    property BitCompare: TBitComparison read FBitCompare write FBitCompare
      default bcAtLeast;
    property DynamicFieldUpdate: boolean read FDynamicFieldUpdate
      write FDynamicFieldUpdate default false;
    property PreventGrayed: boolean read FPreventGrayed
      write FPreventGrayed default false;
    property ItemListField: string read FItemListField write FItemListField;
    property ItemListDataField: string read GetItemListDataField write
      SetItemListDataField;
    property ItemListDataSource: TIB_DataSource read GetItemListDataSource write
      SetItemListDataSource;
    property ParamName: string read GetParamName write SetParamName;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    //procedure Click; override;
    function GetDisplayString(ForVal: integer): string;

  end;                                  // TIB_CustomBitCheckListEnh


  // I can publish IB_WinControl stuff, no special ListBox handling...
  TIB_BitCheckListEnh = class(TIB_CustomBitCheckListEnh)
    {$DEFINE IB_AutoLabel}
    {$I IB_WinControl.PBL}
    {$UNDEF IB_AutoLabel}
  public
    {$IFNDEF HELPSCAN}
    property Checked;
    property State;
    property DataLink;
    {$ENDIF}
  published
    property BitCompare;
    property BitPosition;
    //property Ctl3DShallow;
    property DynamicFieldUpdate;
    property PreventGrayed;
    property ItemListField;
    property ItemListDataField;
    property ItemListDataSource;
    property ItemListAutoLoad;
    //property ParamName;
    {$IFNDEF HELPSCAN}
    property Columns;
    property ItemHeight;
    property Items;
    property OnClickCheck;
    {$ENDIF}
  end;                                  // TIB_BitCheckListEnh


  TLoad_IB_BitCheckListEnhItems = procedure( Sender: TIB_CustomBitCheckListEnh;
    const AItems: TStrings; const DataField: TIB_Column; var WasLoaded: boolean );


var
  Load_IB_BitCheckListEnhItems: TLoad_IB_BitCheckListEnhItems;

implementation

{$DEFINE IB_AutoLabel}
{$DEFINE IB_WinControl}
{$I IB_Control.IMP}
{$UNDEF IB_WinControl}
{$UNDEF IB_AutoLabel}
// From IB_WinControl.IMP

procedure TIB_CustomBitCheckListEnh.SetTabStopAutomation( AValue: boolean );
begin
  if FTabStopAutomation <> AValue then
  begin
    FTabStopAutomation := AValue;
    if FTabStopAutomation then
      UpdateTabStop
    else
      TabStop := true; // return to original default
  end;
end;

procedure TIB_CustomBitCheckListEnh.UpdateTabStop;
begin
  // Only perform update if automation is enabled
  if FTabStopAutomation then
  begin
    // We only play with TabStop when in a valid state.  Otherwise we
    // can end up setting all controls to readonly at form startup, which
    // may alter which control is selected.
    if DataLink.State in [dssSearch,dssBrowse,dssEdit,dssInsert,dssDelete] then
      TabStop := not DataLink.ReadOnly;
  end;
end;

function TIB_CustomBitCheckListEnh.StoreTabStop: boolean;
begin
  // Dont store TabStop when automating its value
  if FTabStopAutomation then
    Result := false
  // if no automation then store when not true (the usual default)
  else
    Result := (TabStop <> true);
end;

function TIB_CustomBitCheckListEnh.CanModify: Boolean;
begin
  Result := (DataLink.IsParamEdit or DataLink.CanModify);
end;

function TIB_CustomBitCheckListEnh.GetReadOnly: Boolean;
begin
  Result := DataLink.ControlIsReadOnly;
end;

procedure TIB_CustomBitCheckListEnh.SetReadOnly(AValue: Boolean);
begin
  DataLink.ControlIsReadOnly := AValue;
end;

function TIB_CustomBitCheckListEnh.GetPreventEditing: Boolean;
begin
  Result := DataLink.ControlPreventsEditing;
end;

procedure TIB_CustomBitCheckListEnh.SetPreventEditing(AValue: Boolean);
begin
  DataLink.ControlPreventsEditing := AValue;
end;

function TIB_CustomBitCheckListEnh.GetPreventInserting: Boolean;
begin
  Result := DataLink.ControlPreventsInserting;
end;

procedure TIB_CustomBitCheckListEnh.SetPreventInserting(AValue: Boolean);
begin
  DataLink.ControlPreventsInserting := AValue;
end;

function TIB_CustomBitCheckListEnh.GetPreventSearching: Boolean;
begin
  Result := DataLink.ControlPreventsSearching;
end;

procedure TIB_CustomBitCheckListEnh.SetPreventSearching(AValue: Boolean);
begin
  DataLink.ControlPreventsSearching := AValue;
end;

procedure TIB_CustomBitCheckListEnh.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  SysStateChanged;
end;

procedure TIB_CustomBitCheckListEnh.UpdateData(Sender: TIB_DataLink;
  IB_DataSource: TIB_DataSource;
  IB_Field: TIB_Column);
begin
  SysUpdateData(IB_Field);
end;

procedure TIB_CustomBitCheckListEnh.WMSetFocus(var Message: TWMSetFocus);
begin
  inherited;
  DataLink.SetFocus;
end;



// My Own Stuff

constructor TIB_CustomBitCheckListEnh.Create(AOwner: TComponent);
begin
  FIB_FieldDataLink := TIB_FieldDataLink.Create(Self);
  FItemListLink := TIB_FieldDataLink.Create(Self);
  FAutoLabel := TIB_AutoLabel.Create(Self);
  FDynamicFieldUpdate := false;
  FPreventGrayed := false;
  inherited Create(AOwner);
  FItemListAutoLoad := true;
  FItemsWereLoaded := false;
  ControlStyle := ControlStyle + [csReplicatable];
  DataLink.Control := Self;
  DataLink.OnStateChanged := StateChanged;
  DataLink.OnDataChange := DataChange;
  DataLink.OnPreparedChanged := PreparedChanged;
  FItemListLink.OnDataChange := DataChange;
  DataLink.OnUpdateData := Self.UpdateData;
  FBitPosition := 0;
  FBitCompare := bcAtLeast;
  OnPrepareSQL := PrepareBitsSQL;
end;

destructor TIB_CustomBitCheckListEnh.Destroy;
begin
  FAutoLabel.Free;
  FAutoLabel := nil;
  ItemListDataSource := nil;
  FItemListLink.OnDataChange := nil;
  DataSource := nil;
  DataLink.Control := nil;
  DataLink.OnStateChanged := nil;
  DataLink.OnDataChange := nil;
  DataLink.OnUpdateData := nil;
  inherited Destroy;
end;

procedure TIB_CustomBitCheckListEnh.Loaded;
begin
  inherited Loaded;
  SysStateChanged;
  SysDataChange(nil);
  RemakeAutoLabel;
  LoadItemList;
end;

procedure TIB_CustomBitCheckListEnh.ClickCheck;
begin
  if DataLink.Modify then
  begin
    inherited ClickCheck;
    DataLink.ControlIsModified := true;
    if DynamicFieldUpdate then
      DataLink.UpdateRecord;
  end
  else
    SysDataChange(Field);
end;

procedure TIB_CustomBitCheckListEnh.SysStateChanged;
begin
  // No point doing anything if we disappearing anyway.
  if csDestroying in ComponentState then
    Exit;
  with DataLink do
  begin
    if ColorScheme then
    begin
      inherited Color := Color;
    end;
    inherited AllowGrayed := ((State = dssSearch) or IsParamEdit) and not PreventGrayed;
    UpdateTabStop;
  end;
end;

procedure TIB_CustomBitCheckListEnh.LoadItemList;
var
  WasLoaded: boolean;
begin
  // If no specific ItemListField and no specific items defined
  // try to load from global function
  if DataLink.Prepared and (Length(ItemListField) <= 0) and
    (ItemListAutoLoad) and Assigned(Load_IB_BitCheckListEnhItems) then
  begin
    Load_IB_BitCheckListEnhItems(Self, Items, Field, WasLoaded );
    FItemsWereLoaded := WasLoaded;
  end
  else
    inherited;
end;

function TIB_CustomBitCheckListEnh.AreItemsStored: boolean;
begin
  if ItemListAutoLoad and FItemsWereLoaded then
    Result := false
  else
    Result := true;
end;

procedure TIB_CustomBitCheckListEnh.PreparedChanged(Sender: TIB_DataLink;
  IB_DataSource: TIB_DataSource);
begin
  LoadItemList;
end;

procedure TIB_CustomBitCheckListEnh.SysDataChange;
var
  i, Val: integer;
begin
  // No point doing anything if we disappearing anyway.
  if csDestroying in ComponentState then
    Exit;
  with DataLink do
  begin
    // First try and initialise the item list...
    // If editing we try and get from the ItemListLink first
    if Editing then
    begin
      if FItemListLink.Active then
      begin
        Self.Items.Text := FItemListLink.Field.AsString;
      end
      else
      begin
        if Length(ItemListField) > 0 then
        begin
          if Active then
            Self.Items.Text := Dataset.FieldByName(ItemListField).AsString
          else
            Self.Items.Clear;
        end;
      end;
    end
      // If not editing we try the supplied ItemListField first...
    else
    begin
      if Length(ItemListField) > 0 then
      begin
        if Active then
          Self.Items.Text := Dataset.FieldByName(ItemListField).AsString
        else
          Self.Items.Clear;
      end
      else
      begin
        if FItemListLink.Active then
          Self.Items.Text := FItemListLink.Field.AsString;
      end;
    end;
    // Now initialise the checkmarks...
    if FieldText <> '' then
    begin
      try
        Val := StrToInt(FieldText);
      except
        Val := 0;
      end;
      for i := 0 to (Items.Count - 1) do
      begin
        Self.State[i] := cbUnchecked;   // clear any cbGrayed values
        Self.Checked[i] := BitIsSet(Val, BitPosition + i);
      end;
    end
    else
    begin
      for i := 0 to (Items.Count - 1) do
        Self.State[i] := cbGrayed;
    end;
  end;
end;

procedure TIB_CustomBitCheckListEnh.SysUpdateData;
var
  i, Val: integer;
  AllGrayed: boolean;
begin
  with DataLink do
  begin
    AllGrayed := true;
    for i := 0 to (Items.Count - 1) do
    begin
      if Self.State[i] <> cbGrayed then
      begin
        AllGrayed := false;
        break;
      end;
    end;
    if AllGrayed then
    begin
      FieldText := '';
      if (State = dssSearch) or SearchAlways then
        AdvSearchData := '';
    end
    else
    begin
      Val := 0;
      for i := 0 to (Items.Count - 1) do
      begin
        if Self.State[i] = cbChecked then
          BitTurnOn(Val, BitPosition + i);
      end;
      FieldText := IntToStr(Val);
      if (State = dssSearch) or SearchAlways then
        AdvSearchData := FieldText; // for special prepare processing
    end;
  end;
end;

procedure TIB_CustomBitCheckListEnh.CMEnter(var Message: TCMEnter);
begin
  DataLink.SetFocus;
  inherited;
end;

procedure TIB_CustomBitCheckListEnh.CMExit(var Message: TCMExit);
begin
  try
    DataLink.UpdateRecord;
  except
    SetFocus;
    raise;
  end;
  inherited;
end;

procedure TIB_CustomBitCheckListEnh.SetBitPosition(AValue: TBitPosition);
begin
  if AValue <> FBitPosition then
  begin
    FBitPosition := AValue;
    SysDataChange(Field);
  end;
end;

procedure TIB_CustomBitCheckListEnh.PrepareBitsSQL(Sender: TIB_DataLink;
  ADataSource: TIB_DataSource);
var
  Hold: integer;
  tmpstr: string;
begin
  with DataLink do
  begin
    if (State = dssSearch) and (Length(FieldText) > 0) then
      tmpstr := FieldText
    else
      tmpstr := AdvSearchData;
    if Length(tmpstr) > 0 then
    begin
      try
        Hold := StrToInt(tmpstr);
      except
        Hold := 0;
      end;
      Dataset.SQLWhereItems.Add(GetBitsWhereSQL(FieldName, Hold, BitCompare));
    end;
  end;
end;

function TIB_CustomBitCheckListEnh.GetItemListDataField: string;
begin
  Result := FItemListLink.FieldName;
end;

procedure TIB_CustomBitCheckListEnh.SetItemListDataField(AValue: string);
begin
  FItemListLink.FieldName := AValue;
end;

function TIB_CustomBitCheckListEnh.GetItemListDataSource: TIB_DataSource;
begin
  Result := FItemListLink.DataSource;
end;

procedure TIB_CustomBitCheckListEnh.SetItemListDataSource(AValue:
  TIB_DataSource);
begin
  FItemListLink.DataSource := AValue;
end;

function TIB_CustomBitCheckListEnh.GetDisplayString(ForVal: integer): string;
var
  ItemsList: TStringList;
  i: integer;
begin
  Result := '';
  ItemsList := TStringList.Create;
  try
    if (FItemListLink.Active) then
    begin
      ItemsList.Text := FItemListLink.Field.AsString;
    end
    else
    begin
      if (Length(ItemListField) > 0) and DataLink.Active then
      begin
        ItemsList.Text := DataLink.Dataset.FieldByName(ItemListField).AsString;
      end
      else
      begin
        ItemsList.Text := Items.Text;
      end;
    end;
    for i := (ItemsList.Count - 1) downto 0 do
    begin
      if (ForVal and (1 shl i)) = 0 then
        ItemsList.Delete(i);
    end;
    Result := ItemsList.Text;
  finally
    ItemsList.Free;
  end;
end;

function TIB_CustomBitCheckListEnh.GetParamName: string;
begin
  Result := DataLink.ParamName;
end;

procedure TIB_CustomBitCheckListEnh.SetParamName(AVal: string);
begin
  DataLink.ParamName := AVal;
end;

end.