{ Telesis_IB_CheckBoxEnh.pas                      SearchTag: ~~EnhComponents~~

 Copyright 1999-2001 - Telesis Computing Pty Ltd.  Australia.
Contact: Geoff Worboys - geoff@telesiscomputing.com.au }

{: @summary IBO Aware variation of TCustomCheckBoxEnh.
<br>
@desc ...
<br>
<br>
<strong>History</strong><br>
<br>
16-Jan-2008 - Adjust AllowGrayed to allow IsParamEdit mode to turn off grayed
checkboxes when the parameters REQUIRED column-attribute is defined.<br>
<br>
29-May-2002 - Published some properties and events from base class.<br>
<br>
29-Apr-2002 - Addition of DoAfterValidateEdit override and rearranged
UpdateData processing. (Improvements to Parameter edit capability.)<br>
<br>
17-Dec-2001 - Published some properties and events from base class.<br>
<br>
23-Aug-2001 - Support to synchronise colors on drop box.  Additional
csDestroying checks.<br>
<br>
02-Aug-2001 - Documentation.<br>
<br>
<br>
<strong>Author: </strong>Geoff Worboys - Telesis Computing Pty Ltd.<br>
<br>}
unit Telesis_IB_CheckBoxEnh;

{$INCLUDE IB_Directives.inc}

interface

uses SysUtils,
  Messages,
  Windows,
  Classes,
  Graphics,
  Menus,
  Controls,
  Forms,
  StdCtrls,
  Mask,
  ExtCtrls,
  IB_Session,
  IB_Components,
  IB_Controls,
  IB_FieldDataLink,
  Telesis_CheckBoxEnh,
  IBC_Label;


type
  {$DEFINE IBC_CheckBoxEnh}
  {$DEFINE IB_UserClass}
  {$DEFINE IB_FieldControl}

  // I have to use IB_Control not IB_WinControl because the default IB_WinControl
  // stuff does some BorderStyle handling which is inappropriate.  Because of
  // this I need to include the bits of IB_WinControl that I do want...
  TIB_CustomCheckBoxEnh = class(TCustomCheckBoxEnh)
    {$DEFINE IB_AutoLabel}
    {$DEFINE IB_WinControl}
    {$I IB_Control.INT}
    {$UNDEF IB_WinControl}
    {$UNDEF IB_AutoLabel}
    // From IB_WinControl.INT
  private
    FTabStopAutomation: boolean;
    function StoreTabStop: boolean;
    procedure SetTabStopAutomation( AValue: boolean );
    procedure UpdateTabStop;
    function GetParamName: string;
    procedure SetParamName(AVal: string);
    procedure SetPreventEditing(AValue: Boolean);
    procedure SetPreventInserting(AValue: Boolean);
    procedure SetPreventSearching(AValue: Boolean);
    function GetPreventEditing: Boolean;
    function GetPreventInserting: Boolean;
    function GetPreventSearching: Boolean;
    procedure CMEnter(var Message: TCMEnter); message CM_ENTER;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure CMEnabledChanged(var Message: TMessage); message
      CM_ENABLEDCHANGED;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
  protected
    function GetReadOnly: boolean; virtual;
    procedure SetReadOnly(AValue: Boolean); virtual;
    procedure SysUpdateData(IB_Field: TIB_Column); virtual;
    procedure DoAfterValidateEdit(var ErrCode: integer); override;
    procedure UpdateData(Sender: TIB_DataLink;
      IB_DataSource: TIB_DataSource;
      IB_Field: TIB_Column);
  public
    function CanModify: Boolean; virtual;
  public
    property ReadOnly: boolean read GetReadOnly
    write SetReadOnly
      default false;
    property PreventEditing: boolean read GetPreventEditing
    write SetPreventEditing
      default false;
    property PreventInserting: boolean read GetPreventInserting
    write SetPreventInserting
      default false;
    property PreventSearching: boolean read GetPreventSearching
    write SetPreventSearching
      default false;
    property Ctl3D default true;
    property State stored false;
    property ParamName: string read GetParamName write SetParamName;
    property TabStopAutomation: boolean read FTabStopAutomation
                                        write SetTabStopAutomation
                                        default false;
    property TabStop stored StoreTabStop;

    // My Own Stuff - partly copied from IB_CheckBox...
  private
    FValueChecked: string;
    FValueUncheck: string;
    function GetFieldState: TCheckBoxState; virtual;

  protected
    procedure SetAllowGrayed(AValue: boolean); override;
    function IsMaskStored: boolean; override;
    function GetDBDisplayField: TObject; override;
    function IBG_GetAutoLabel: TObject; override;
    procedure Loaded; override;
    procedure Toggle; override;
    procedure KeyPress(var Key: Char); override;

    property Checked;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

  end;                                  // TIB_CustomCheckBoxEnh


  // I can publish IB_WinControl stuff, no special ListBox handling...
  TIB_CheckBoxEnh = class(TIB_CustomCheckBoxEnh)
    {$DEFINE IB_AutoLabel}
    {$I IB_WinControl.PBL}
    {$UNDEF IB_AutoLabel}
  public
  published
    property Ctl3DShallow;
    property ActiveWhenReadOnly;
    property AutoDropDown;
    property DropDownColor;
    property DropDownFont;
    property DropDownFontAuto;
    property GlyphAlign;
    property GlyphSize;
    property OnButton1;
    property OnButton2;
    property OnButton3;
    property OnDrawButton;
    property AfterValidateEdit;
    property OnValidateEdit;
    property OnValidateError;
    property OnGlyphClick;
    property OnDrawGlyph;

    // From IB_CheckBox
//    property Alignment;
//    property Caption;
//    property Checked;
//    property State;
//    property AllowGrayed;
//    property DisplayLabelAsCaption;

  end;                                  // TIB_CheckBoxEnh


implementation

uses Telesis_Locale, Telesis_Constants;

{$DEFINE IB_AutoLabel}
{$DEFINE IB_WinControl}
{$I IB_Control.IMP}
{$UNDEF IB_WinControl}
{$UNDEF IB_AutoLabel}
// From IB_WinControl.IMP

procedure TIB_CustomCheckBoxEnh.SetTabStopAutomation( AValue: boolean );
begin
  if FTabStopAutomation <> AValue then
  begin
    FTabStopAutomation := AValue;
    if FTabStopAutomation then
      UpdateTabStop
    else
      TabStop := true; // return to original default
  end;
end;

procedure TIB_CustomCheckBoxEnh.UpdateTabStop;
begin
  // Only perform update if automation is enabled
  if FTabStopAutomation then
  begin
    // We only play with TabStop when in a valid state.  Otherwise we
    // can end up setting all controls to readonly at form startup, which
    // may alter which control is selected.
    if DataLink.State in [dssSearch,dssBrowse,dssEdit,dssInsert,dssDelete] then
      TabStop := not DataLink.ReadOnly;
  end;
end;

function TIB_CustomCheckBoxEnh.StoreTabStop: boolean;
begin
  // Dont store TabStop when automating its value
  if FTabStopAutomation then
    Result := false
  // if no automation then store when not true (the usual default)
  else
    Result := (TabStop <> true);
end;

function TIB_CustomCheckBoxEnh.CanModify: Boolean;
begin
  Result := (DataLink.IsParamEdit or DataLink.CanModify);
end;

function TIB_CustomCheckBoxEnh.GetReadOnly: Boolean;
begin
  Result := DataLink.ControlIsReadOnly;
end;

procedure TIB_CustomCheckBoxEnh.SetReadOnly(AValue: Boolean);
begin
  DataLink.ControlIsReadOnly := AValue;
end;

function TIB_CustomCheckBoxEnh.GetPreventEditing: Boolean;
begin
  Result := DataLink.ControlPreventsEditing;
end;

procedure TIB_CustomCheckBoxEnh.SetPreventEditing(AValue: Boolean);
begin
  DataLink.ControlPreventsEditing := AValue;
end;

function TIB_CustomCheckBoxEnh.GetPreventInserting: Boolean;
begin
  Result := DataLink.ControlPreventsInserting;
end;

procedure TIB_CustomCheckBoxEnh.SetPreventInserting(AValue: Boolean);
begin
  DataLink.ControlPreventsInserting := AValue;
end;

function TIB_CustomCheckBoxEnh.GetPreventSearching: Boolean;
begin
  Result := DataLink.ControlPreventsSearching;
end;

procedure TIB_CustomCheckBoxEnh.SetPreventSearching(AValue: Boolean);
begin
  DataLink.ControlPreventsSearching := AValue;
end;

procedure TIB_CustomCheckBoxEnh.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  SysStateChanged;
end;

procedure TIB_CustomCheckBoxEnh.DoAfterValidateEdit(var ErrCode: integer);
begin
  inherited DoAfterValidateEdit(ErrCode);
  if (ErrCode = E_EDITENH_NOERROR) and DataLink.IsParamEdit then
  begin
    DataLink.UpdateData;
    SysDataChange(nil);
  end;
end;

procedure TIB_CustomCheckBoxEnh.UpdateData(Sender: TIB_DataLink;
  IB_DataSource: TIB_DataSource;
  IB_Field: TIB_Column);
begin
  SysUpdateData(IB_Field);
end;

procedure TIB_CustomCheckBoxEnh.WMSetFocus(var Message: TWMSetFocus);
begin
  inherited;
  DataLink.SetFocus;
end;


// My Own Stuff - partly copied from IB_CheckBox

constructor TIB_CustomCheckBoxEnh.Create(AOwner: TComponent);
begin
  FIB_FieldDataLink := TIB_FieldDataLink.Create(Self);
  FAutoLabel := TIB_AutoLabel.Create(Self);
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csReplicatable];
  DataLink.Control := Self;
  DataLink.OnStateChanged := StateChanged;
  DataLink.OnDataChange := DataChange;
  DataLink.OnUpdateData := Self.UpdateData;
  //FDisplayLabelAsCaption := false;
end;

destructor TIB_CustomCheckBoxEnh.Destroy;
begin
  FAutoLabel.Free;
  FAutoLabel := nil;
  DataSource := nil;
  DataLink.Control := nil;
  DataLink.OnStateChanged := nil;
  DataLink.OnDataChange := nil;
  DataLink.OnUpdateData := nil;
  inherited Destroy;
end;

procedure TIB_CustomCheckBoxEnh.Loaded;
begin
  inherited Loaded;
  SysStateChanged;
  SysDataChange(nil);
  RemakeAutoLabel;
end;

procedure TIB_CustomCheckBoxEnh.SysStateChanged;
begin
  // No point doing anything if we disappearing anyway.
  if csDestroying in ComponentState then
    Exit;
  with DataLink do
  begin
    if Assigned(Field) then
    begin
      FValueChecked := Field.BooleanTrue;
      FValueUncheck := Field.BooleanFalse;
    end
    else
    begin
      FValueChecked := ENHSTR_Bool_True[1];
      FValueUncheck := ENHSTR_Bool_False[1];
    end;
    if ColorScheme then
    begin
      inherited Color := Color;
    end;
    // Field AllowGrayed is managed by the IsNullable state
    // Param AllowGrayed can be managed by controlling the Required state
    inherited SetAllowGrayed( (not Assigned(Field)) or
      (IsParamEdit and not Field.Required) or
      ((not IsParamEdit) and ((DataLink.State = dssSearch) or Field.IsNullable))
      );
    UpdateTabStop;
  end;
  SyncDropBoxColors;
end;

procedure TIB_CustomCheckBoxEnh.SysDataChange;
begin
  // No point doing anything if we disappearing anyway.
  if csDestroying in ComponentState then
    Exit;
  State := GetFieldState;
end;

procedure TIB_CustomCheckBoxEnh.SysUpdateData;
begin
  with DataLink do
  begin
    if Self.State = cbGrayed then
    begin
      FieldText := '';
    end
    else
      if Checked then
      begin
        FieldText := FValueChecked;
      end
      else
      begin
        FieldText := FValueUncheck;
      end;
  end;
end;

function TIB_CustomCheckBoxEnh.GetFieldState: TCheckBoxState;
begin
  with DataLink do
  begin
    if not Assigned(Field) then
    begin
      Result := cbGrayed;
    end
    else
      if FValueChecked = FieldText then
      begin
        Result := cbChecked;
      end
      else
        if FValueUncheck = FieldText then
        begin
          Result := cbUnchecked;
        end
        else
        begin
          Result := cbGrayed;
        end;
  end;
end;

procedure TIB_CustomCheckBoxEnh.Toggle;
begin
  if DataLink.IsParamEdit or DataLink.Modify then
  begin
    inherited Toggle;
    DataLink.ControlIsModified := true;
  end;
end;

procedure TIB_CustomCheckBoxEnh.SetAllowGrayed(AValue: boolean);
begin
  if AllowGrayed <> AValue then
  begin
    //FAllowGrayed := AValue;
    SysStateChanged;
  end;
end;

function TIB_CustomCheckBoxEnh.IsMaskStored: boolean;
begin
  // IBO does not store masks with controls, only with the data definition
  Result := false;
end;

function TIB_CustomCheckBoxEnh.GetDBDisplayField: TObject;
begin
  Result := DataLink.Field;
end;

function TIB_CustomCheckBoxEnh.IBG_GetAutoLabel: TObject;
begin
  Result := FAutoLabel;
end;

procedure TIB_CustomCheckBoxEnh.KeyPress(var Key: Char);
begin
  inherited KeyPress(Key);
  case Key of
    #8, ' ':
      if DataLink.IsParamEdit or DataLink.Modify then
      begin
        DataLink.ControlIsModified := true;
      end;
    #27:
      SysDataChange(nil);
  end;
end;

procedure TIB_CustomCheckBoxEnh.CMEnter(var Message: TCMEnter);
begin
  inherited;
  SysDataChange(nil);
end;

procedure TIB_CustomCheckBoxEnh.CMExit(var Message: TCMExit);
begin
  inherited;   // base class does a ValidateEdit
  if not DataLink.IsParamEdit then
    DataLink.UpdateData;
  SysDataChange(nil);
end;

function TIB_CustomCheckBoxEnh.GetParamName: string;
begin
  Result := DataLink.ParamName;
end;

procedure TIB_CustomCheckBoxEnh.SetParamName(AVal: string);
begin
  DataLink.ParamName := AVal;
end;

end.