{ Telesis_IB_ComboBoxEnh.pas                      SearchTag: ~~EnhComponents~~

 Copyright 1999-2001 - Telesis Computing Pty Ltd.  Australia.
Contact: Geoff Worboys - geoff@telesiscomputing.com.au }

{: @summary IBO Aware variation of TCustomComboBox.
<br>
@desc ...
<br>
<br>
<strong>History</strong><br>
<br>
12-Aug-2005 - Changes for more consistent SelectAll.<br>
<br>
19-Jun-2005 - Trying to synchronise ValidateEdit and UpdateData calls so that
both get done but we minimise duplication and avoid recursion.<br>
<br>
22-Sep-2002 - Published ListOwnerDraw and OnListDrawItem properties.<br>
<br>
29-May-2002 - Published some properties and events from base class.<br>
<br>
29-Apr-2002 - Addition of DoAfterValidateEdit override and rearranged
UpdateData processing. (Improvements to Parameter edit capability.)<br>
<br>
17-Dec-2001 - Published some properties and events from base class.<br>
<br>
12-Sep-2001 - Prevent MaxLength being set when using DropDownList.<br>
<br>
23-Aug-2001 - Support to synchronise colors on drop box.  Additional
csDestroying checks.<br>
<br>
02-Aug-2001 - Documentation.<br>
<br>
<br>
<strong>Author: </strong>Geoff Worboys - Telesis Computing Pty Ltd.<br>
<br>}
unit Telesis_IB_ComboBoxEnh;

{$INCLUDE IB_Directives.inc}


interface

uses SysUtils,
  Messages,
  Windows,
  Classes,
  Graphics,
  Menus,
  Controls,
  Forms,
  StdCtrls,
  Mask,
  ExtCtrls,
  IB_Session,
  IB_Components,
  IB_Controls,
  IB_FieldDataLink,
  IB_Header,
  Telesis_MaskEnh,
  Telesis_EditEnh,
  Telesis_ComboEnh,
  IBC_Label;

type

  {$DEFINE IBC_ComboBoxEnh}
  {$DEFINE IB_UserClass}
  {$DEFINE IB_FieldControl}

{: @summary IBO-aware variation of TCustomComboBoxEnh.}
  TIB_CustomComboBoxEnh = class(TCustomComboboxEnh)
    {$DEFINE IB_CustomEdit}
    {$DEFINE IB_AutoLabel}
    {$I IB_WinControl.INT}
    {$UNDEF IB_AutoLabel}
    {$UNDEF IB_CustomEdit}
  private
    FItemsWereLoaded: boolean;
    FItemListAutoLoad: boolean;
    FDisplayCol: TIB_Column;
    FDisplayField: string;

  protected
    procedure LoadItemList; override;
    function AreItemsStored: boolean; override;
    function IsMaskStored: boolean; override;
    function GetDBDisplayField: TObject; override;
    function IBG_GetAutoLabel: TObject; override;
    procedure DoAfterValidateEdit(var ErrCode: integer); override;
    procedure Loaded; override;
    procedure Change; override;

    // from IB_CustomEdit.INT
    procedure EditingChanged(Sender: TIB_DataLink; IB_DataSource:
      TIB_DataSource); virtual;
    procedure PreparedChanged(Sender: TIB_DataLink; IB_DataSource:
      TIB_DataSource); virtual;

    // override the IIB_GridWinControl interface translation function to
    // intelligently return the correct value display value for rows
    // other than the current
    function IBG_TranslateText(AVal: string): string; override;

{: @summary  Is defined for use when dropping the control on IB_Grid.
<br>
@desc Set this to the field name of the column that this control should be
matched to in the grid.  If you are using the "display text=value" mode of
operation this would usually be a calculated field which displays the required
text based on the actual value column, otherwise it is usually the same as
the DataField value.<br>
<br>
The Telesis_ComboEnh.pas unit also defines a global function,
GetComboBoxEnhDisplayString, which can be particularly useful for setting
up calculated fields to support "display text=value" combo-boxes in IB_Grid.
Firstly define your calculated field, and then in the OnCalculateField event
for that dataset you define the value of the field as...<br>
<br>
<code>
    if AField.FieldName = "MYCALCFIELD" then
      AField.AsString := GetComboBoxEnhDisplayString( Combo1.Items,
        ARow.ByName( "MyValueField").AsString );
</code><br>
This means that you only need to define your "display text=value" definitions
in the combo box itself, using this function you can extract the definition
at runtime to define your calculated field.}
    property DisplayField: string read FDisplayField write FDisplayField;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure Reset; override;
    function EditCanModify: boolean; override;
    function EditBeginModify: boolean; override;

{: @summary  When this is true the control will use the global function
Load_IB_ComboBoxEnhItems (if it is defined) to assign the contents of the
Items property.}
    property ItemListAutoLoad: boolean read FItemListAutoLoad
    write FItemListAutoLoad default true;
    property Items;

  published

  end;                                  // TIB_CustomComboBoxEnh



{: @summary Published variation of TIB_CustomComboBoxEnh.}
  TIB_ComboBoxEnh = class(TIB_CustomComboBoxEnh)
    {$DEFINE IB_AutoLabel}
    {$I IB_WinControl.PBL}
    {$UNDEF IB_AutoLabel}
  public
    property IsMasked;
    property IsEmpty;
    property DisplayText;
    property ButtonState;
    property Text;
    property ItemIndex;
    property ItemValue;
    property ItemValues;

  published
    property ItemListAutoLoad;
    property DisplayField;

    // From TCustomComboBox
    property ListColumns;
    property ListRows;
    property ListWidth;
    property Items;
    property Sorted;
    property TabWidth;
    property ListOwnerDraw;
    property OnListDrawItem;

    // From TCustomComboEnh
    property Style;
    property ActiveWhenReadOnly;
    property AutoDropDown;
    property DropDownColor;
    property DropDownFont;
    property DropDownFontAuto;
    property OnCloseUp;
    property BeforeDropDown;
    property OnDropDown;

    // Stuff we want from CustomEdit.PBL
    property AutoSelect;
    property AutoSize;
    property HideSelection;
    property OEMConvert;
    property OnChange;

    // TCustomEditEnh publications...
    //property Alignment;      // only set via dataset/connection
    property Button;
    //property CharCase;       // only set via dataset/connection
    property Ctl3DShallow;
    //property EditMask;       // only set via dataset/connection
    //property EditMaskMode;   // only set via dataset/connection
    property GlyphAlign;
    property GlyphSize;
    property Options;
    //property Text;           // only set via dataset/connection

    property OnButton1;
    property OnButton2;
    property OnButton3;
    property OnDrawButton;
    property AfterValidateEdit;
    property OnValidateEdit;
    property OnValidateError;
    property OnGlyphClick;
    property OnDrawGlyph;

  end;                                  // TIB_ComboBoxEnh

{: @summary Type for Load_IB_ComboBoxEnhItems global function.
<br>
@desc Refer to <see Global=Load_IB_ComboBoxEnhItems> for details.}
  TLoad_IB_ComboBoxEnhItems = procedure(Sender: TIB_CustomComboBoxEnh;
    const AItems: TStrings; const DataField: TIB_Column; var WasLoaded:
    boolean);


var
{: @summary Used to automatically load values into IB_ComboBoxEnh.Items
properties.
<br>
@desc The global function Load_IB_ComboBoxEnhItems was provided so that
applications can setup a function that detects the domain of a field and
load the appropriate values into the Items automatically.  This makes it
possible for an application to maintain Items lists from a central point.}
  Load_IB_ComboBoxEnhItems: TLoad_IB_ComboBoxEnhItems;


implementation

uses IB_Parse, Telesis_Constants;

{$DEFINE IB_CustomEdit}
{$DEFINE IB_AutoLabel}
{$I IB_WinControl.IMP}
{$UNDEF IB_AutoLabel}
{$UNDEF IB_CustomEdit}


constructor TIB_CustomComboBoxEnh.Create(AOwner: TComponent);
begin
  FIB_FieldDataLink := TIB_FieldDataLink.Create(Self);
  FAutoLabel := TIB_AutoLabel.Create(Self);
  inherited Create(AOwner);
  FItemsWereLoaded := false;
  FItemListAutoLoad := true;
  DataLink.Control := Self;
  DataLink.OnStateChanged := StateChanged;
  DataLink.OnEditingChanged := EditingChanged;
  DataLink.OnPreparedChanged := PreparedChanged;
  DataLink.OnDataChange := DataChange;
  DataLink.OnUpdateData := UpdateData;
  // Somebody plays with BorderStyle in the generic units, default it
  // back to where it belongs...
  BorderStyle := bsSingle;
  FDisplayField := '';
  FDisplayCol := nil;
end;

destructor TIB_CustomComboBoxEnh.Destroy;
begin
  FAutoLabel.Free;
  FAutoLabel := nil;
  DataSource := nil;
  DataLink.Control := nil;
  DataLink.OnStateChanged := nil;
  DataLink.OnEditingChanged := nil;
  DataLink.OnPreparedChanged := nil;
  DataLink.OnDataChange := nil;
  DataLink.OnUpdateData := nil;
  inherited Destroy;
end;

procedure TIB_CustomComboBoxEnh.Loaded;
begin
  inherited Loaded;
  SysStateChanged;
  SysDataChange(nil);
  RemakeAutoLabel;
end;

// Copy of function from IB_CustomEdit.IMP

procedure TIB_CustomComboBoxEnh.EditingChanged(Sender: TIB_DataLink;
  IB_DataSource: TIB_DataSource);
var
  NewRO: boolean;
begin
  NewRO := DataLink.ReadOnly;
  if not NewRO then
  begin
    with DataLink do
    begin
      if not ((State in [dssInsert, dssEdit, dssSearch]) or IsParamEdit) then
      begin
        NewRO := not SearchAlways;
      end;
    end;
  end;
  inherited ReadOnly := NewRO;
end;


procedure TIB_CustomComboBoxEnh.PreparedChanged(Sender: TIB_DataLink;
  IB_DataSource: TIB_DataSource);
begin
  LoadItemList;
end;

procedure TIB_CustomComboBoxEnh.SysDataChange;
var
  tmpstr: string;
begin
  // No point doing anything if we disappearing anyway.
  if csDestroying in ComponentState then
    Exit;
  tmpstr := DataLink.FieldText;
  ItemValue := tmpstr;
  if (Style = csDropDownList) and (Length(tmpstr) > 0) and (ItemIndex < 0) then
  try
    Style := csDropDown;     // allow direct change to the text
    Text := tmpstr;          // assign even mismatched values
  finally
    Style := csDropDownList; // restore the style
  end;
  if Focused and AutoSelect and not TextMouseDown then
    SelectAll;
  Modified := false;
end;

procedure TIB_CustomComboBoxEnh.SysStateChanged;  { Handles LayoutChange event as well }
var
  tmpstr: string;
begin
  // No point doing anything if we disappearing anyway.
  if csDestroying in ComponentState then
    Exit;
  with DataLink do
  begin
    FDisplayCol := nil;
    if Dataset <> nil then
    begin
      if ColorScheme then
        inherited Color := Color;
        // Default the DisplayCol to the actual column...
      if IsParamEdit then
        FDisplayCol := DataLink.Field
      else
      begin
        if DisplayField = '' then
          FDisplayCol := DataLink.Field
        else
        begin
          if (not (csDesigning in ComponentState)) and
            (Dataset.Prepared) then
            FDisplayCol := Dataset.FieldByName(DisplayField);
        end;
      end;
    end;
  end;
  if FDisplayCol <> nil then
  begin
    case FDisplayCol.CharCase of
      ccUpper: CharCase := ecUpperCase;
      ccLower: CharCase := ecLowerCase;
      ccProper: CharCase := ecProperCase;
    else
      CharCase := ecNormal;
    end;
    if Alignment <> FDisplayCol.Alignment then
      Alignment := FDisplayCol.Alignment;
    tmpstr := FDisplayCol.EditMask;
    if Length(tmpstr) > 0 then
      Self.SetEditMaskEx(tmpstr)
    else
      Self.EditMaskMode := eemNone;
    //MaxLength := FDisplayCol.MaxLength;
    if (DataLink.State = dssSearch) or (Style = csDropDownList) then
      MaxLength := 0
    else
    begin
      if (FDisplayCol.SQLType = SQL_Varying) or
        (FDisplayCol.SQLType = SQL_Varying_) or
        (FDisplayCol.SQLType = SQL_Text) or
        (FDisplayCol.SQLType = SQL_Text_) then
        MaxLength := FDisplayCol.SQLLen
      else
        MaxLength := 0;
    end;
  end;
  EditingChanged(DataLink, DataSource);
  UpdateTabStop;
  SyncDropBoxColors;
end;

procedure TIB_CustomComboBoxEnh.SysUpdateData;
var
  tmpStr: string;
begin
  if Assigned(DataLink.Field) then
  begin
    if Modified then
    begin
      if not Validating then
        ValidateEdit;
      tmpStr := ItemValue;
      // If in search mode we use an explicit operator definition (=)
      // so that we avoid problems if the Item Value contains SQL
      // wildcard characters.
      if ((DataLink.State = dssSearch) or SearchAlways) and
        not DataLink.IsParamEdit then
      begin
        if tmpStr = '' then
        begin
          DataLink.SearchBuffer := '';
          DataLink.SetAdvSearchDisplay('');
        end
        else
        begin
          DataLink.SearchBuffer := '=' + tmpStr;
          DataLink.SetAdvSearchDisplay(tmpStr);
        end;
      end
      else
        DataLink.FieldText := tmpStr;
    end;
  end;
  Modified := false;
end;

procedure TIB_CustomComboBoxEnh.LoadItemList;
var
  WasLoaded: boolean;
begin
  // If no items specified then try to load from global function
  if DataLink.Prepared and (ItemListAutoLoad) and
    Assigned(Load_IB_ComboBoxEnhItems) then
  begin
    Load_IB_ComboBoxEnhItems(Self, Items, Field, WasLoaded);
    FItemsWereLoaded := WasLoaded;
  end
  else
    inherited;
end;

function TIB_CustomComboBoxEnh.AreItemsStored: boolean;
begin
  if ItemListAutoLoad and FItemsWereLoaded then
    Result := false
  else
    Result := true;
end;

function TIB_CustomComboBoxEnh.IsMaskStored: boolean;
begin
  // IBO does not store masks with controls, only with the data definition
  Result := false;
end;

function TIB_CustomComboBoxEnh.GetDBDisplayField: TObject;
begin
  Result := FDisplayCol;
end;

function TIB_CustomComboBoxEnh.IBG_GetAutoLabel: TObject;
begin
  Result := FAutoLabel;
end;

procedure TIB_CustomComboBoxEnh.DoAfterValidateEdit(var ErrCode: integer);
begin
  inherited DoAfterValidateEdit(ErrCode);
  if (ErrCode = E_EDITENH_NOERROR) and not DataLink.DataUpdating then
  begin
    DataLink.UpdateData;
    SysDataChange(nil);
  end;
end;

function TIB_CustomComboBoxEnh.IBG_TranslateText(AVal: string): string;
begin
  // If a special display column has been defined (different to the main data
  // field) then we expect the grid to be using the special display column,
  // and so the given value does not need any translation.
  // If the main data field and display column are the same then we may need
  // the value translated.  Only do the translation on DropDownList, because
  // we are not guaranteed to have a full list of translations in DropDown.
  if (Field = FDisplayCol) and (Style = csDropDownList) then
    Result := GetComboBoxEnhDisplayString(Items, AVal)
  else
    Result := AVal;
end;

function TIB_CustomComboBoxEnh.EditCanModify: boolean;
begin
  if DataLink.IsParamEdit then
    Result := true
  else
    Result := DataLink.CanModify and
      (DataLink.DataSource.Modifying or DataLink.DataSource.AutoEdit or
      (DataLink.DataSource.AutoInsert and
      (DataLink.Dataset.Fields.RowState = rsNone)));
end;

function TIB_CustomComboBoxEnh.EditBeginModify: boolean;
begin
  if DataLink.IsParamEdit then
    Result := true
  else
    Result := DataLink.Modify;
end;

procedure TIB_CustomComboBoxEnh.Reset;
begin
  SysDataChange(nil);
  SelectAll;
end;

procedure TIB_CustomComboBoxEnh.Change;
begin
  if not MaskChanging then
    with DataLink do
    begin
      if (DataChangingLevel = 0) and
        (NeedToPost or (State = dssSearch) or IsParamEdit) then
      begin
        ControlIsModified := true;
        Modified := true;
      end;
    end;
  inherited Change;
end;

procedure TIB_CustomComboBoxEnh.CMEnter(var Message: TCMEnter);
begin
  inherited;
  SysDataChange(nil);
end;

procedure TIB_CustomComboBoxEnh.CMExit(var Message: TCMExit);
begin
  inherited;   // base class does a ValidateEdit
end;

end.