{ Telesis_IB_DateTimeEnh.pas                      SearchTag: ~~EnhComponents~~

 Copyright 1999-2001 - Telesis Computing Pty Ltd.  Australia.
Contact: Geoff Worboys - geoff@telesiscomputing.com.au }

{: @summary IBO Aware variation of TCustomDateTimeEnh.
<br>
@desc ...
<br>
<br>
<strong>History</strong><br>
<br>
08-Mar-2008 - Dont use mask when not focused, allowing for a special display
format that may apply.<br>
<br>
23-May-2006 - Was not showing/hiding seconds field in popup box properly.<br>
<br>
21-Apr-2006 - Some attempts to fix synchronisation now removed.  No longer
taking any notice of the dtoShowTime, dto24HourTime, dtoShowSeconds options.
These to be automatically configured according to the format and data type.<br>
<br>
12-Aug-2005 - Rewrite SysDataChange, the old was too messy and unreliable.
And changes for consistent SelectAll.<br>
<br>
19-Jun-2005 - Trying to synchronise ValidateEdit and UpdateData calls so that
both get done but we minimise duplication and avoid recursion.<br>
<br>
29-May-2002 - Published some properties and events from base class.<br>
<br>
29-Apr-2002 - Addition of DoAfterValidateEdit override and rearranged
UpdateData processing. (Improvements to Parameter edit capability.)<br>
<br>
17-Dec-2001 - Published some properties and events from base class.<br>
<br>
23-Aug-2001 - Support to synchronise colors on drop box.  Additional
csDestroying checks.<br>
<br>
02-Aug-2001 - Documentation.<br>
<br>
<br>
<strong>Author: </strong>Geoff Worboys - Telesis Computing Pty Ltd.<br>
<br>}
unit Telesis_IB_DateTimeEnh;

{$INCLUDE IB_Directives.inc}

interface

uses SysUtils,
  Messages,
  Windows,
  Classes,
  Graphics,
  Menus,
  Controls,
  Forms,
  StdCtrls,
  Mask,
  ExtCtrls,
  IB_Session,
  IB_Components,
  IB_Controls,
  IB_FieldDataLink,
  Telesis_MaskEnh,
  Telesis_EditEnh,
  Telesis_DateTimeEnh,
  IBC_Label;

type

  {$DEFINE IBC_DateTimeEnh}
  {$DEFINE IB_UserClass}
  {$DEFINE IB_FieldControl}

  TIB_CustomDateTimeEnh = class(TCustomDateTimeEnh)
    {$DEFINE IB_CustomEdit}
    {$DEFINE IB_AutoLabel}
    {$I IB_WinControl.INT}
    {$UNDEF IB_AutoLabel}
    {$UNDEF IB_CustomEdit}
  private

  protected
    function IsMaskStored: boolean; override;
    function GetDBDisplayField: TObject; override;
    function IBG_GetAutoLabel: TObject; override;
    procedure DoValidateEdit(var ErrCode: integer); override;
    procedure DoAfterValidateEdit(var ErrCode: integer); override;
    procedure Loaded; override;
    procedure Change; override;

    // from IB_CustomEdit.INT
    procedure EditingChanged(Sender: TIB_DataLink; IB_DataSource:
      TIB_DataSource); virtual;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure Reset; override;
    function EditCanModify: boolean; override;
    function EditBeginModify: boolean; override;

  published

  end;                                  // TIB_CustomDateTimeEnh




  TIB_DateTimeEnh = class(TIB_CustomDateTimeEnh)
    {$DEFINE IB_AutoLabel}
    {$I IB_WinControl.PBL}
    {$UNDEF IB_AutoLabel}
  public
    property IsMasked;
    property IsEmpty;
    property DisplayText;
    property ButtonState;
    property Text;

  published
    property DateOptions;
    property DateStartDay;
    property DateTitleColor;
    property DateDateCellColor;
    //property DateFormat;
    property DateInputLayout;
    property ActiveWhenReadOnly;
    property AutoDropDown;
    property DropDownColor;
    property DropDownFont;
    property DropDownFontAuto;
    property OnCloseUp;
    property BeforeDropDown;
    property OnDropDown;

    // Stuff we want from CustomEdit.PBL
    property AutoSelect;
    property AutoSize;
    property HideSelection;
    property OEMConvert;
    property OnChange;

    // TCustomEditEnh publications...
    //property Alignment;      // only set via dataset/connection
    property Button;
    //property CharCase;       // only set via dataset/connection
    property Ctl3DShallow;
    //property EditMask;       // only set via dataset/connection
    //property EditMaskMode;   // only set via dataset/connection
    property GlyphAlign;
    property GlyphSize;
    property Options;
    //property Text;           // only set via dataset/connection

    property OnButton1;
    property OnButton2;
    property OnButton3;
    property OnDrawButton;
    property AfterValidateEdit;
    property OnValidateEdit;
    property OnValidateError;
    property OnGlyphClick;
    property OnDrawGlyph;

  end;                                  // TIB_DateTimeEnh


implementation

uses Telesis_Constants;

{$DEFINE IB_CustomEdit}
{$DEFINE IB_AutoLabel}
{$I IB_WinControl.IMP}
{$UNDEF IB_AutoLabel}
{$UNDEF IB_CustomEdit}


constructor TIB_CustomDateTimeEnh.Create(AOwner: TComponent);
begin
  FIB_FieldDataLink := TIB_FieldDataLink.Create(Self);
  FAutoLabel := TIB_AutoLabel.Create(Self);
  inherited Create(AOwner);
  DataLink.Control := Self;
  DataLink.OnStateChanged := StateChanged;
  DataLink.OnEditingChanged := EditingChanged;
  DataLink.OnDataChange := DataChange;
  DataLink.OnUpdateData := UpdateData;
  // Somebody plays with BorderStyle in the generic units, default it
  // back to where it belongs...
  BorderStyle := bsSingle;
end;

destructor TIB_CustomDateTimeEnh.Destroy;
begin
  FAutoLabel.Free;
  FAutoLabel := nil;
  DataSource := nil;
  DataLink.Control := nil;
  DataLink.OnStateChanged := nil;
  DataLink.OnEditingChanged := nil;
  DataLink.OnDataChange := nil;
  DataLink.OnUpdateData := nil;
  inherited Destroy;
end;

procedure TIB_CustomDateTimeEnh.Loaded;
begin
  inherited Loaded;
  SysStateChanged;
  SysDataChange(nil);
  RemakeAutoLabel;
end;

procedure TIB_CustomDateTimeEnh.EditingChanged(Sender: TIB_DataLink;
  IB_DataSource: TIB_DataSource);
// Copy of function from IB_CustomEdit.IMP
var
  NewRO: boolean;
begin
  NewRO := DataLink.ReadOnly;
  if not NewRO then
  begin
    with DataLink do
    begin
      if not ((State in [dssInsert, dssEdit, dssSearch]) or IsParamEdit) then
      begin
        NewRO := not SearchAlways;
      end;
    end;
  end;
  inherited ReadOnly := NewRO;
end;

procedure TIB_CustomDateTimeEnh.SysDataChange;
  procedure ClearAndSetDisplay(AStr: string);
  begin
    if IsMasked then
      Mask.Clear;
    DisplayText := AStr;
  end;
begin
  // No point doing anything if we disappearing anyway.
  if csDestroying in ComponentState then
    Exit;
  with DataLink do
  begin
    if not Assigned(Field) or not Prepared then
      ClearAndSetDisplay(InvalidText)
    else
    begin
      if (not IsParamEdit) and ((State = dssSearch) or SearchAlways) then
      begin
        // now do the text
        if SearchWithMask then
          Self.Text := FieldText
        else
          ClearAndSetDisplay(FieldText);
      end
      else if Self.IsMasked and Field.IsNotNull then
      begin
        if Field.IsDateTime or Field.IsDateOnly then
        begin
          if Field.AsDateTime = 0 then
            Self.DisplayText := ''
          else
            Self.DateTime := Field.AsDateTime;
        end
        else
          Self.Text := Field.AsString;
      end
      else
        ClearAndSetDisplay(FieldText);
    end;
  end;
  if Focused and AutoSelect and (not TextMouseDown) and HandleAllocated then
    SelectAll;
  Modified := false;
end;

procedure TIB_CustomDateTimeEnh.SysStateChanged;  { Handles LayoutChange event as well }
var
  tmpstr: string;
  pform: TCustomForm;
begin
  // No point doing anything if we disappearing anyway.
  if csDestroying in ComponentState then
    Exit;
  with DataLink do
  begin
    if ColorScheme then
    begin
      inherited Color := Color;
    end;
    case Datalink.FieldCharCase of
      ccUpper: CharCase := ecUpperCase;
      ccLower: CharCase := ecLowerCase;
      ccProper: CharCase := ecProperCase;
    else
      CharCase := ecNormal;
    end;
    if Alignment <> FieldAlignment then
    begin
      Alignment := FieldAlignment;
    end;
    if Assigned(Field) then
    begin
      Self.DateFormat := Field.DisplayFormat;
      if Field.IsDateOnly then
        Self.DateOptions := DateOptions - [dtoShowTime]
      else
      begin
        Self.DateOptions := DateOptions + [dtoShowTime];
        tmpstr := Self.DateFormat;
        if Length(tmpstr) = 0 then
          tmpstr := ShortTimeFormat;
        // if specified as tt (long time format) set tmpfmt directly
        if Pos('tt', tmpstr) > 0 then
          tmpstr := LongTimeFormat
        // if specified as t (short time format) set tmpfmt directly
        else if Pos('t', tmpstr) > 0 then
          tmpstr := ShortTimeFormat;
        // all of which allows us to test for specific 24 hour (upper case 'H')
        // in the format and so set the 24hourTime option directly
        if Pos('H', tmpstr) > 0 then
          Self.DateOptions := DateOptions + [dto24HourTime]
        else
          Self.DateOptions := DateOptions - [dto24HourTime];
        // and also for the presence of seconds in the format
        if (Pos('s', tmpstr) > 0) or (Pos('z', tmpstr) > 0) then
          Self.DateOptions := DateOptions + [dtoShowSeconds]
        else
          Self.DateOptions := DateOptions - [dtoShowSeconds];
      end;
    end;
    tmpstr := FieldEditMask;
    pform := GetParentForm(Self);
    //if Focused and (Length(tmpstr) > 0) then
    // We dont want just focused, we want "active"
    if Assigned(pform) and (pform.ActiveControl = Self) and (Length(tmpstr) > 0) then
    begin
      if (Self.Mask = nil) or (Self.Mask.MaskEx <> tmpstr) then
        Self.SetEditMaskEx(tmpstr);
    end
    else
      Self.EditMaskMode := eemNone;
    MaxLength := FieldMaxLength;
  end;
  EditingChanged(DataLink, DataSource);
  UpdateTabStop;
  SyncDropBoxColors;
end;

procedure TIB_CustomDateTimeEnh.SysUpdateData;
begin
  if Assigned(Field) then
  begin
    if Modified then
    begin
      if not Validating then
        ValidateEdit;
      if IsEmpty then
        DataLink.FieldText := ''
      else
      begin
        //DataLink.FieldText := Text;
        with DataLink do
        begin
          if IsParamEdit then
            Field.AsDateTime := Self.DateTime
          else
            if (State = dssSearch) or SearchAlways then
              SearchBuffer := Self.Text
            else
              case State of
                dssEdit,
                  dssInsert: Field.AsDateTime := Self.DateTime;
              end;
        end;
      end;
    end;
  end;
  Modified := false;
end;

function TIB_CustomDateTimeEnh.IsMaskStored: boolean;
begin
  // IBO does not store masks with controls, only with the data definition
  Result := false;
end;

function TIB_CustomDateTimeEnh.GetDBDisplayField: TObject;
begin
  Result := DataLink.Field;
end;

function TIB_CustomDateTimeEnh.IBG_GetAutoLabel: TObject;
begin
  Result := FAutoLabel;
end;

procedure TIB_CustomDateTimeEnh.DoValidateEdit(var ErrCode: integer);
begin
  // We need to turn validation off during search mode (unless searching
  // with a mask, in order to allow partial matching search (just date etc).
  if (DataLink.State = dssSearch) or SearchAlways then
  begin
    if SearchWithMask then
      inherited DoValidateEdit(ErrCode);
  end
  else
    inherited DoValidateEdit(ErrCode);
end;

procedure TIB_CustomDateTimeEnh.DoAfterValidateEdit(var ErrCode: integer);
begin
  inherited DoAfterValidateEdit(ErrCode);
  if (ErrCode = E_EDITENH_NOERROR) and not DataLink.DataUpdating then
  begin
    DataLink.UpdateData;
    SysDataChange(nil);
  end;
end;

function TIB_CustomDateTimeEnh.EditCanModify: boolean;
begin
  if DataLink.IsParamEdit then
    Result := true
  else
    Result := DataLink.CanModify and
      (DataLink.DataSource.Modifying or DataLink.DataSource.AutoEdit or
      (DataLink.DataSource.AutoInsert and
      (DataLink.Dataset.Fields.RowState = rsNone)));
end;

function TIB_CustomDateTimeEnh.EditBeginModify: boolean;
begin
  if DataLink.IsParamEdit then
  begin
    // Try to force the control to it's editing state so that the
    // mask etc is all in place for any subsequent assignment.
    DataLink.ControlIsModified := true;
    Modified := true;
    Result := true;
  end
  else
    Result := DataLink.Modify;
end;

procedure TIB_CustomDateTimeEnh.Reset;
begin
  SysDataChange(nil);
  SelectAll;
end;

procedure TIB_CustomDateTimeEnh.Change;
begin
  if not MaskChanging then
    with DataLink do
    begin
      if (DataChangingLevel = 0) and
        (NeedToPost or (State = dssSearch) or IsParamEdit) then
      begin
        ControlIsModified := true;
        Modified := true;
      end;
    end;
  inherited Change;
end;

procedure TIB_CustomDateTimeEnh.CMEnter(var Message: TCMEnter);
var
  tmpstr: string;
begin
  with DataLink do // assign any editmask
  begin
    tmpstr := FieldEditMask;
    if Length(tmpstr) > 0 then
      Self.SetEditMaskEx(tmpstr)
    else
      Self.EditMaskMode := eemNone;
  end;
  inherited;
  SysDataChange(nil);
end;

procedure TIB_CustomDateTimeEnh.CMExit(var Message: TCMExit);
begin
  inherited;           // base class does a ValidateEdit
  EditMaskMode := eemNone; // remove any edit-mask
  SysDataChange(nil);  // reset for null entries and display changes
end;

end.
