{ Telesis_IB_EditEnh.pas                          SearchTag: ~~EnhComponents~~

 Copyright 1999-2001 - Telesis Computing Pty Ltd.  Australia.
Contact: Geoff Worboys - geoff@telesiscomputing.com.au }

{: @summary IBO Aware variation of TCustomEditEnh.
<br>
@desc ...
<br>
<br>
<strong>History</strong><br>
<br>
12-Aug-2005 - Rewrite SysDataChange, the old was too messy and unreliable.
Also changes for SelectAll consistency.<br>
<br>
19-Jun-2005 - Trying to synchronise ValidateEdit and UpdateData calls so that
both get done but we minimise duplication and avoid recursion.<br>
<br>
13-May-2005 - Minor fix for IsParamEdit mode to show correct value while
focused.<br>
<br>
29-May-2002 - Published some properties and events from base class.<br>
<br>
29-Apr-2002 - Addition of DoAfterValidateEdit override and rearranged
UpdateData processing. (Improvements to Parameter edit capability.)<br>
<br>
17-Dec-2001 - Published some properties and events from base class.  Also
removed any attempt to support Delphi2.<br>
<br>
23-Aug-2001 - additional csDestroying checks.<br>
<br>
02-Aug-2001 - Documentation.<br>
<br>
<br>
<strong>Author: </strong>Geoff Worboys - Telesis Computing Pty Ltd.<br>
<br>}
unit Telesis_IB_EditEnh;

{$INCLUDE IB_Directives.inc}

interface

uses SysUtils,
  Messages,
  Windows,
  Classes,
  Graphics,
  Menus,
  Controls,
  Forms,
  StdCtrls,
  Mask,
  ExtCtrls,
  IB_Components,
  IB_FieldDataLink,
  Telesis_EditEnh,
  IBC_Label;


type
  {$DEFINE IBC_EditEnh}
  {$DEFINE IB_UserClass}
  {$DEFINE IB_FieldControl}

{: @summary The IBO-aware variation of TCustomEditEnh.
<br>
@desc Provides support for enhanced mask processing (see TMaskEnh) and
embedded buttons amongst other things.<br>
<br>
Refer to TMaskEnhString for details on how to specify masks.<br>
<br>
<strong>IMPORTANT:</strong> To have the control operate with PasswordChar
you MUST specify a mask.  The simplest option is to choose a mask-mode of
eemEditEnh and a mask value of "cccccccccc;0;_" (where you specify as
many "c" character positions as required/permitted for the password.}
  TIB_CustomEditEnh = class(TCustomEditEnh)
    {$DEFINE IB_AutoLabel}
    {$I IBC_CustomEdit.INT}
    {$UNDEF IB_AutoLabel}
  private
{!}
  private
    FBorderStyle: TBorderStyle;
    procedure SetBorderStyle(AValue: TBorderStyle);
  protected
    property BorderStyle: TBorderStyle read FBorderStyle
    write SetBorderStyle
      default bsSingle;
  protected
    function IsMaskStored: boolean; override;
    function GetDBDisplayField: TObject; override;
    function IBG_GetAutoLabel: TObject; override;
    procedure DoAfterValidateEdit(var ErrCode: integer); override;
    procedure Loaded; override;
    procedure Change; override;
  public
    property IsMasked;
    property IsEmpty;
    property DisplayText;
    property ButtonState;
    property Text;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Reset; override;
    function EditCanModify: boolean; override;
    function EditBeginModify: boolean; override;
  end;                                  // TIB_CustomEditEnh

{: @summary The published variation of TIB_CustomEditEnh.
<br>
@desc ...}
  TIB_EditEnh = class(TIB_CustomEditEnh)
    {$DEFINE IB_CustomEdit}
    {$DEFINE IB_AutoLabel}
    {$I IB_WinControl.PBL}
    {$UNDEF IB_AutoLabel}
    {$UNDEF IB_CustomEdit}
  published
    property AutoSelect;
    property AutoSize;
    property HideSelection;
    property OEMConvert;
    property OnChange;

    // TCustomEditEnh publications...
    //property Alignment;      // only set via dataset/connection
    property Button;
    //property CharCase;       // only set via dataset/connection
    property Ctl3DShallow;
    //property EditMask;       // only set via dataset/connection
    //property EditMaskMode;   // only set via dataset/connection
    property GlyphAlign;
    property GlyphSize;
    property Options;
    //property Text;           // only set via dataset/connection

    property OnButton1;
    property OnButton2;
    property OnButton3;
    property OnDrawButton;
    property AfterValidateEdit;
    property OnValidateEdit;
    property OnValidateError;
    property OnGlyphClick;
    property OnDrawGlyph;
  end;


implementation

uses Telesis_MaskEnh, Telesis_Constants;

{$DEFINE IB_AutoLabel}
{$I IB_WinControl.IMP}
{$UNDEF IB_AutoLabel}

constructor TIB_CustomEditEnh.Create(AOwner: TComponent);
begin
  FIB_FieldDataLink := TIB_FieldDataLink.Create(Self);
  FAutoLabel := TIB_AutoLabel.Create(Self);
  inherited Create(AOwner);
  DataLink.Control := Self;
  DataLink.OnStateChanged := StateChanged;
  DataLink.OnEditingChanged := EditingChanged;
  DataLink.OnDataChange := DataChange;
  DataLink.OnUpdateData := UpdateData;
  // Somebody plays with BorderStyle in the generic units, default it
  // back to where it belongs...
  BorderStyle := bsSingle;
end;

destructor TIB_CustomEditEnh.Destroy;
begin
  FAutoLabel.Free;
  FAutoLabel := nil;
  DataSource := nil;
  DataLink.Control := nil;
  DataLink.OnStateChanged := nil;
  DataLink.OnEditingChanged := nil;
  DataLink.OnDataChange := nil;
  DataLink.OnUpdateData := nil;
  inherited Destroy;
end;

procedure TIB_CustomEditEnh.Loaded;
begin
  inherited Loaded;
  SysStateChanged;
  SysDataChange(nil);
  RemakeAutoLabel;
end;

// Copy of function from IB_CustomEdit.IMP

procedure TIB_CustomEditEnh.EditingChanged(Sender: TIB_DataLink;
  IB_DataSource: TIB_DataSource);
var
  NewRO: boolean;
begin
  NewRO := DataLink.ReadOnly;
  if not NewRO then
  begin
    with DataLink do
    begin
      if not ((State in [dssInsert, dssEdit, dssSearch]) or IsParamEdit) then
      begin
        NewRO := not SearchAlways;
      end;
    end;
  end;
  inherited ReadOnly := NewRO;
end;

procedure TIB_CustomEditEnh.SysDataChange;
  procedure ClearAndSetDisplay(AStr: string);
  begin
    if IsMasked then
      Mask.Clear;
    DisplayText := AStr;
  end;
begin
  if csDestroying in ComponentState then
    Exit;
  with DataLink do
  begin
    if not Assigned(Field) or not Prepared then
      ClearAndSetDisplay(InvalidText)
    else
    begin
      if (not IsParamEdit) and ((State = dssSearch) or SearchAlways) then
      begin
        if SearchWithMask then
          Self.Text := FieldText
        else
          ClearAndSetDisplay(FieldText);
      end
      else if Self.IsMasked and Field.IsNotNull then
        Self.Text := Field.AsString
      else
        ClearAndSetDisplay(FieldText);
    end;
  end;
  if Focused and AutoSelect and (not TextMouseDown) and HandleAllocated then
    SelectAll;
  Modified := false;
end;

procedure TIB_CustomEditEnh.SysStateChanged;
var
  tmpstr: string;
begin
  // No point doing anything if we disappearing anyway.
  if csDestroying in ComponentState then
    Exit;
  with DataLink do
  begin
    if ColorScheme then
    begin
      inherited Color := Color;
    end;
    case FieldCharCase of
      ccUpper: CharCase := ecUpperCase;
      ccLower: CharCase := ecLowerCase;
      ccProper: CharCase := ecProperCase;
    else
      CharCase := ecNormal;
    end;
    if Alignment <> FieldAlignment then
    begin
      Alignment := FieldAlignment;
    end;
    tmpstr := FieldEditMask;
    if Length(tmpstr) > 0 then
      Self.SetEditMaskEx(tmpstr)
    else
      Self.EditMaskMode := eemNone;
    MaxLength := FieldMaxLength;
  end;
  EditingChanged(DataLink, DataSource);
  UpdateTabStop;
end;

procedure TIB_CustomEditEnh.SysUpdateData;
begin
  if Assigned(Field) then
  begin
    if Modified then
    begin
      if not Validating then
        ValidateEdit;
      if IsEmpty then
        DataLink.FieldText := ''
      else
        DataLink.FieldText := Text;
    end;
  end;
  Modified := false;
end;

function TIB_CustomEditEnh.IsMaskStored: boolean;
begin
  // IBO does not store masks with controls, only with the data definition
  Result := false;
end;

function TIB_CustomEditEnh.GetDBDisplayField: TObject;
begin
  Result := DataLink.Field;
end;

function TIB_CustomEditEnh.IBG_GetAutoLabel: TObject;
begin
  Result := FAutoLabel;
end;

procedure TIB_CustomEditEnh.DoAfterValidateEdit(var ErrCode: integer);
begin
  inherited DoAfterValidateEdit(ErrCode);
  if (ErrCode = E_EDITENH_NOERROR) and not DataLink.DataUpdating then
  begin
    DataLink.UpdateData;
    SysDataChange(nil);
  end;
end;

function TIB_CustomEditEnh.EditCanModify: boolean;
begin
  if DataLink.IsParamEdit then
    Result := true
  else
    Result := DataLink.CanModify and
      (DataLink.DataSource.Modifying or DataLink.DataSource.AutoEdit or
      (DataLink.DataSource.AutoInsert and
      (DataLink.Dataset.Fields.RowState = rsNone)));
end;

function TIB_CustomEditEnh.EditBeginModify: boolean;
begin
  if DataLink.IsParamEdit then
    Result := true
  else
    Result := DataLink.Modify;
end;

procedure TIB_CustomEditEnh.Reset;
begin
  SysDataChange(nil);
  SelectAll;
end;

procedure TIB_CustomEditEnh.Change;
begin
  if not MaskChanging then
    with DataLink do
    begin
      if (DataChangingLevel = 0) and
        (NeedToPost or (State = dssSearch) or SearchAlways or IsParamEdit) then
      begin
        ControlIsModified := true;
        Modified := true;
      end;
    end;
  inherited Change;
end;

procedure TIB_CustomEditEnh.CMEnter(var Message: TCMEnter);
begin
  inherited;
  SysDataChange(nil);
end;

procedure TIB_CustomEditEnh.CMExit(var Message: TCMExit);
begin
  inherited; // base class does a ValidateEdit if modified
end;

end.