{ Telesis_IB_LocateEditEnh.pas                    SearchTag: ~~EnhComponents~~

 Copyright 1999-2001 - Telesis Computing Pty Ltd.  Australia.
Contact: Geoff Worboys - geoff@telesiscomputing.com.au }

{: @summary Specialised TIB_EditEnh control used for use locate actions.
<br>
@desc ...
<br>
<br>
<strong>History</strong><br>
<br>
02-Aug-2001 - Documentation.<br>
<br>
<br>
<strong>Author: </strong>Geoff Worboys - Telesis Computing Pty Ltd.<br>
<br>}
unit
  Telesis_IB_LocateEditEnh;

{$INCLUDE IB_Directives.inc}

interface

uses Windows,
  Messages,
  SysUtils,
  Classes,
  Controls,
  StdCtrls,
  ExtCtrls,
  Mask,
  IB_Components,
  IB_Controls,
  Telesis_EditEnh,
  Telesis_IB_EditEnh;

type

  TIB_LocateEditEnh = class(TIB_EditEnh)
  private
    FLocateOptions: TIB_LocateOptions;
    FDisableEditing: boolean;

  protected
    procedure SetDisableEditing(AVal: boolean);
    procedure EditingChanged(Sender: TIB_DataLink;
      IB_DataSource: TIB_DataSource); override;
    procedure DoLocate; virtual;
    procedure KeyPress(var Key: Char); override;
    //procedure DoEditButtonClick; override;
    procedure SysStateChanged; override;
    //procedure UpdateMask( UnboundMask: string ); override;

  public
    function EditCanModify: Boolean; override;
    function EditBeginModify: Boolean; override;
    {: This is the text of the value that is sent to the Locate() method.}
    property Text;
    procedure Locate;

  published
    {: These are the options used when the internal call to Locate() is made.}
    property LocateOptions: TIB_LocateOptions read FLocateOptions
    write FLocateOptions
      default [];
    property DisableEditing: boolean read FDisableEditing write SetDisableEditing
      default false;
  end;



implementation

uses Telesis_MaskEnh;


function TIB_LocateEditEnh.EditBeginModify: Boolean;
begin
  if DataLink.State in [dssInsert, dssEdit, dssSearch] then
  begin
    if DisableEditing then
      Result := false
    else
      Result := inherited EditCanModify;
  end
  else
    Result := DataLink.Active;
end;

procedure TIB_LocateEditEnh.SetDisableEditing(AVal: boolean);
begin
  if AVal <> FDisableEditing then
  begin
    FDisableEditing := AVal;
    SysStateChanged;
  end;
end;

function TIB_LocateEditEnh.EditCanModify: Boolean;
begin
  if DataLink.State in [dssInsert, dssEdit, dssSearch] then
  begin
    if DisableEditing then
      Result := false
    else
      Result := inherited EditCanModify;
  end
  else
    Result := DataLink.Active;
end;

procedure TIB_LocateEditEnh.SysStateChanged;
var
  tmpstr: string;
begin
  with DataLink do
  begin
    if (State in [dssInsert, dssEdit, dssSearch]) and (not DisableEditing) then
    begin
      inherited SysStateChanged;
    end
    else
    begin
      if ColorScheme then
      begin
        if (State in [dssInsert, dssEdit, dssSearch]) and DisableEditing then
          inherited Color := IB_Session.ReadOnlyColor
        else
          inherited Color := IB_Session.SearchingColor;
      end;
      case FieldCharCase of
        ccUpper: CharCase := ecUpperCase;
        ccLower: CharCase := ecLowerCase;
        ccProper: CharCase := ecProperCase;
      else
        CharCase := ecNormal;
      end;
      if Alignment <> FieldAlignment then
      begin
        Alignment := FieldAlignment;
      end;
      tmpstr := FieldEditMask;
      if (Length(tmpstr) > 0) and SearchWithMask then
        Self.SetEditMaskEx(tmpstr)
      else
        Self.EditMaskMode := eemNone;
      MaxLength := FieldMaxLength;
    end;
  end;
end;

procedure TIB_LocateEditEnh.EditingChanged(Sender: TIB_DataLink;
  IB_DataSource: TIB_DataSource);
begin
  if (DataLink.State in [dssInsert, dssEdit, dssSearch]) then
  begin
    if DisableEditing then
      inherited ReadOnly := true
    else
      inherited EditingChanged(Sender, IB_DataSource);
  end;
end;

procedure TIB_LocateEditEnh.KeyPress(var Key: Char);
var
  HoldKey: Char;
begin
  HoldKey := Key;
  inherited KeyPress(Key);
  if (HoldKey = #13) and not (DataLink.State in [dssInsert, dssEdit, dssSearch])
    then
  begin
    DoLocate;
    Key := #0;
  end;
end;

{procedure TIB_LocateEditEnh.DoEditButtonClick;
begin
  if Assigned( OnButtonClick ) then
    inherited DoEditButtonClick
  else if not(DataLink.State in [dssInsert,dssEdit,dssSearch]) then
    DoLocate;
end;}

procedure TIB_LocateEditEnh.DoLocate;
var
  tmpFld: string;
begin
  with DataLink do
    if Active and (Dataset is TIB_BDataset) then
      with Dataset as TIB_BDataset do
      begin
        tmpFld := DataField;
        if tmpFld = '' then
          tmpFld := OrderingLink;
        if tmpFld = '' then
          MessageBeep(0)
        else
          Locate(tmpFld, Text, LocateOptions);
      end
    else
      MessageBeep(0);
end;

{procedure TIB_LocateEditEnh.UpdateMask( UnboundMask: string );
begin
  if (DataLink.State in [dssInsert,dssEdit,dssSearch]) and
    (not DisableEditing) then
    inherited UpdateMask( UnboundMask );
// Prevent mask from being applied - in locate mode
end;}

procedure TIB_LocateEditEnh.Locate;
begin
  DoLocate;
end;

end.
