{ Telesis_IB_LookupEnh.pas                        SearchTag: ~~EnhComponents~~

 Copyright 1999-2001 - Telesis Computing Pty Ltd.  Australia.
Contact: Geoff Worboys - geoff@telesiscomputing.com.au }

{: @summary Specialised variation of TCustomComboEnh used as dataset lookup.
<br>
@desc ...
<br>
<br>
<strong>History</strong><br>
<br>
03-Jun-2008 - A fix to ActivateLookup thanks to Craig Leonardi that ensures
the lookup remains active if the dataset returns to prepared (after posting
an insert).  Some additional adjustments to try and keep the Modified flag
properly synchronised - to avoid excessive processing and also to ensure that
changes are detected when dataset update-data methods are called.<br>
<br>
12-Feb-2008 - MouseWheel support.<br>
<br>
06-Jun-2006 - incremental searching needs to be char by char.<br>
<br>
12-Aug-2005 - changes for SelectAll consistency (not during button down).<br>
<br>
13-May-2005 - SearchAlways used to allow param editing to remain active
even when dataset is not in prepared or search state.
<br>
28-Jan-2005 - DoAfterValidateEdit needed to check that its datalink was
prepared.  Calling EditEndModify during SysUpdateData in order to ensure
that search criteria is saved correctly.  Improved behaviour when used
for parameter editing - if you still have trouble try turning off the
AutoFetchFirst option of the lookup dataset (as a transaction refresh can
have it refresh and select the first row automatically).<br>
<br>
13-Jan-2005 - Fixed GetDBDisplayField to check if field exists, otherwise
apps could occassionally get into exception catch loops.  Added a FilterMode
property so it is possible to setup filter change capability only in search
mode (useful to stop users from editing or creating records with unwanted
values, but still allowing search).<br>
<br>
08-Nov-2002 - Adjusted the access of DisplayCol so that it is possible to
use the control without a DisplayField (no calculated field on the main
dataset.  For this to work the lookup dataset must not be shared and the
LookupDeactivation property must be false.<br>
<br>
29-May-2002 - Published some properties and events from base class.<br>
<br>
29-Apr-2002 - Addition of DoAfterValidateEdit override and rearranged
UpdateData processing. (Improvements to Parameter edit capability.)<br>
<br>
05-Apr-2002 - Altered so Reset resets the modified state.<br>
<br>
25-Dec-2001 - Published some more events and properties from base classes.
Added some Locale strings.  Improved some code to prevent AVs if the control
loses focus during mouse processing - does not really solve all problems
but at least the AVs dont happen.<br>
<br>
11-Nov-2001 - Preventing OnChange from being called multiple times during
the one search event - I suspect something altered slightly in IBO and an
earlier work around was no longer totally appropriate.<br>
<br>
16-Sep-2001 - When IndicateTitles was true the record selected when
the popup first displayed was not showing (hidden under title).
Now fixed I think.<br>
<br>
13-Sep-2001 - Setup so that the control will disable callbacks during
incremental searching and activation processing.  With the IBO4 horizontal
refinement capabilities the visual feedback supported with callbacks during
lookup searching appears to be redundant, and preventing callbacks allows
the typing into the control to be much smoother.<br>
<br>
Adjusted BruteSearch to use IB_Connection.CustomCompareText if
it is defined.<br>
<br>
Removed assignment of MaxLength as this was causing problems when existing
values filled the control - prevent insert of a new character to search for
something different.<br>
<br>
23-Aug-2001 - Made is that button is invalidated on state change. Reset should
attempt to restore previous display value - and so reselect any prior record
selection.  Stopped an AV from occuring when the datatype of the LookupField
did not match the datatype of the DisplayField - this was causing the the
Alignment to change, resulting in RecreateWnd call with unpredictable
results.  Some csDestroying checks to prevent AVs at destruction.<br>
<br>
02-Aug-2001 - Documentation.<br>
<br>
<br>
<strong>Author: </strong>Geoff Worboys - Telesis Computing Pty Ltd.<br>
<br>}
unit Telesis_IB_LookupEnh;

{$INCLUDE IB_Directives.inc}




interface

uses SysUtils,
  Messages,
  Windows,
  Classes,
  Graphics,
  Menus,
  Controls,
  Forms,
  StdCtrls,
  Mask,
  ExtCtrls,
  Grids,
  IB_Session,
  IB_Components,
  IB_Controls,
  IB_FieldDataLink,
  IB_Grid,
  IB_Header,
  Telesis_MaskEnh,
  Telesis_EditEnh,
  Telesis_ComboEnh,
  IBC_Label;

type

{: @summary Dropbox class used by TIB_CustomLookupEnh.
<br>
@desc Used to display a customised TIB_LookupList.}
  TIB_ComboLookupListEnh = class(TCustomComboControlEnh)
  private
    FList: TIB_LookupList;
    FFilterPnl: TPanel;
    FFilterCB: TCheckBox;
    FFilterInternal: boolean;
    FWorkPnl: TPanel;

    procedure OnFilterCBChange(Sender: TObject);

  protected
    procedure InitFilterCB;
    procedure ToggleFilterCB;

    function DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): boolean; override;
    function DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): boolean; override;
    function DoMouseWheel(Shift: TShiftState; WheelDelta: integer; MousePos: TPoint): boolean; override;
    procedure SizeControl; override;
    procedure SyncToParentValue; override;
    procedure SyncParentToValue; override;
    procedure SyncColors; override;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure CloseUp; override;
    procedure SetWorking(DspStr: string);
    procedure ClrWorking;

  end;

  TIB_LookupFilterMode = (
    lfmDisabled,
    lfmAlways,
    lfmSearchOnly );

  {$DEFINE IBC_LookupEnh}
  {$DEFINE IB_UserClass}
  {$DEFINE IB_FieldControl}

{: @summary Specialised dataset lookup control.
<br>
@desc Very similar to TIB_LookupCombo that comes with IBObjects but it is
setup a little differently and comes with a few new features.<br>
<br>
<strong>Example:</strong><br>
<br>
Its probably best if I start with an example, which will give you familiarity
with the terms used in the remainder of the description.<br>
<br>
Given a two tables defined as...<br>
<code>
  CREATE TABLE LOOKUP_TABLE (
    PKEY INTEGER NOT NULL PRIMARY KEY,
    NAME VARCHAR(30) NOT NULL );

  CREATE TABLE MAIN_TABLE (
    PKEY INTEGER NOT NULL PRIMARY KEY,
    AFIELD VARCHAR(30) NOT NULL,
    LOOKUP_KEY INTEGER,
    FOREIGN KEY (LOOKUP_KEY) REFERENCES LOOKUP_TABLE (PKEY) );

  INSERT INTO LOOKUP_TABLE( PKEY, NAME ) VALUES( 1, 'A NAME' );
  INSERT INTO LOOKUP_TABLE( PKEY, NAME ) VALUES( 2, 'B NAME' );
  INSERT INTO LOOKUP_TABLE( PKEY, NAME ) VALUES( 3, 'C NAME' );
</code><br>
Where LOOKUP_TABLE is a set of valid options that can be selected into the
foreign key field (FKEY) of the MAIN_TABLE.<br>
<br>
The form DFM to use these tables would come out something like the following
(showing only the properties of relevance).<br>
<code>
  object MainQuery: TIB_Query
    SQL.Strings = (
      'SELECT PKEY, AFIELD, LOOKUP_KEY,'
      '  (SELECT NAME FROM LOOKUP_TABLE'
      '   WHERE LOOKUP_TABLE.PKEY=MAIN_TABLE.LOOKUP_KEY)'
      '   AS LOOKUP_NAME'
      'FROM MAIN_TABLE'
      'FOR UPDATE')
    KeyLinks.Strings = (
      'PKEY')
  end
  object MainSource: TIB_DataSource
    Dataset = MainQuery
  end
  object LookupQuery: TIB_Query
    FieldsVisible.Strings = (
      'PKEY=FALSE')
    SQL.Strings = (
      'SELECT PKEY, NAME FROM LOOKUP_TABLE')
    KeyLinks.Strings = (
      'PKEY=LOOKUP_KEY')
    KeyDescLinks.Strings = (
      'NAME=LOOKUP_NAME')
    KeySource = MainSource
    OrderingItemNo = 1
    OrderingItems.Strings = (
      'By Name=NAME')
    OrderingLinks.Strings = (
      'NAME=1')
  end
  object LookupSource: TIB_DataSource
    Dataset = LookupQuery
  end
</code><br>
The TIB_LookupEnh would be setup as...<br>
<code>
  object IB_LookupEnh1: TIB_LookupEnh
    DataField = 'LOOKUP_KEY'
    DataSource = MainSource
    LookupField = 'NAME'
    LookupSource = LookupSource
    DisplayField = 'LOOKUP_NAME'
  end
</code><br>
You can see that the main and lookup queries are setup in the same way
that you would setup to use the IBObjects TIB_LookupCombo control.  You should
note that the use of the KeyDescLinks property and the embedded select (or
join to get equivalent descriptive field) would usually be defined to use
TIB_LookupEnh (as opposed to TIB_LookupCombo where those properties are
only required when working inside grids).  See the notes on automatically
activated lookups below.<br>
<br>
Notice that TIB_LookupEnh has a specific connection to the MainSource
and the foreign key field (where TIB_LookupCombo does not).  If your
foreign key is a compound key you should select the last field of the
compound key.<br>
<br>
The DisplayField property should be setup to connect to the descriptive
field representing the foreign key link - in the example above this is
the LOOKUP_NAME field obtained via the embedded select in the main query.<br>
<br>
The LookupField property should select the field from the lookup dataset
that is used in the selected Ordering* properties of the dataset.  To operate
properly the contents of this field should match the descriptive field in
the main dataset (since the LookupQuery will directly alter that descriptive
field during edit/insert via the KeyDescLinks definition).<br>
<br>
<br>
<strong>Features:</strong><br>
<br>
Using the example given above you can see that this control is very similar
to TIB_LookupCombo, but it does offer some additional features that are
not possible with TIB_LookupCombo.<br>
<br>
<br>
Automatic Lookup Dataset Activation and Deactivation.<br>
<br>
By default this control takes control over the lookup dataset
activation/deactivation.  That is; When the main dataset is in a browse state
then the lookup dataset will have been closed by this control.  Since the
lookup is not active the control displays the value of the DisplayField
property.<br>
<br>
When the main dataset is editing (or the user performs a lookup and the
dataset is in auto-edit) then the control will activate the lookup dataset
(which will automatically synchronise itself).   The lookup dataset will
remain active until the main dataset leaves its editing state.  While it is
active the control displays the value retrieved from the LookupField.<br>
<br>
What all this means is that, when a user is scrolling through a list of
records, they do not have to wait for all the lookup datasets to synchronise
before scrolling to the next record - because with this control the lookups
are not active.  It also means that lookup datasets do not have to be
explicitly activated in your form code and will not add to the delays in
opening a form.<br>
<br>
This can make a significant difference if the lookup datasets are accessing
large tables.<br>
<br>
This feature can be controlled via the LookupActivation and LookupDeactivation
properties of the control.<br>
<br>
Note that you can avoid needing the embedded lookup and KeyDescLinks if you
manually activate the lookup dataset (as you would when using TIB_LookupCombo)
and set the LookupDeactivation property to false.  If the lookup dataset is
always active then the display value can be retrieved from the lookup dataset
instead of the descriptive field in the main dataset.<br>
<br>
<br>
Shared Lookup Datasets.<br>
<br>
The TIB_LookupEnh control can be setup so that multiple controls can share
the same lookup dataset.  It is IMPORTANT that all TIB_LookupEnh controls
that use a shared lookup dataset have the LookupShared property set to true.
Results are unpredictable if one or more controls thinks it has exclusive
use of the lookup dataset.<br>
<br>
To use this facility you use the following properties:<br>
<code>
    property LookupShared: boolean;
    property LookupKeyLinks: TIB_StringList;
    property LookupKeyDescLinks: TIB_StringList;
</code><br>
Set LookupShared to true and set the LookupKeyLinks and LookupKeyDescLinks to
the values that you want the lookup dataset to have (for its KeyLinks and
KeyDescLinks properties) when this control has possession of the lookup
dataset.  LookupKeyLinks and LookupKeyDescLinks are only used when
LookupShared is true - they are ignored when not sharing a lookup dataset
so an unshared dataset must have its own values defined as usual.<br>
<br>
You HAVE TO specify appropriate LookupKeyDescLinks values for the control
to function correctly when sharing a lookup dataset.  This is because the
display field value MUST be updated while the control is using the dataset
and this requires the KeyDescLinks defined to say what fields should be
updated.<br>
<br>
When LookupShared is true the LookupActivation and LookupDeactivation settings
are ignored.  Instead the control will always automatically activate and
deactivate the lookup dataset as required.  The lookup dataset is only
activated when the control has focus AND input specifies that a lookup is
to be performed (a keystroke, or dropdown of the lookup box).  The lookup
dataset is deactivated as soon as the control loses focus (so you will need
to pay attention to the focus issues described below).<br>
<br>
When TIB_LookupEnh activates a shared lookup dataset it resets the datasets
IB_Connection, IB_Transaction, KeySource, KeyLinks and KeyDescLinks
properties.  This allows a lookup dataset to be shared between multiple
forms and even multiple connections/transactions if required.<br>
<br>
Because shared lookups will be activated/deactivated by the TIB_LookupEnh
controls, these datasets should not be used for other purposes - such as
displaying other field values in other sorts of controls.<br>
<br>
Note: TIB_LookupEnh takes advantage of IBObjects AdvSearch* datalink
properties to support search mode use via shared lookup datasets as well
as normal browse/edit modes.<br>
<br>
<br>
Lost Focus Issues:<br>
<br>
If the lookup dataset is large enough to result in a CallBack then the
dialog which is shown (whether the IBO default or one of your own) takes
the focus away from the LookupEnh control which results in the dropbox
being closed.  This can be rather annoying.<br>
<br>
To avoid this problem you override the OnCallback event for the lookup
dataset and when Status is not csFinal call the TIB_LookupEnh method
SetWorking with a display string value of your choice - the control will
display this string in a small panel at the bottom of the dropbox.  When
Status is csFinal call the TIB_LookupEnh method ClrWorking which will
hide the panel from the bottom of the dropbox.<br>
<br>
<br>
Dynamically Changing Lookup Dataset Filtering.<br>
<br>
IBObjects TIB_Query supports VCL compatible filter definitions.  (See the
VCL online help for details.)  It is possible to setup TIB_LookupEnh so that
the user can dynamically activate/deactivate a defined filter while
performing a lookup.   This lets you setup the interface so that the user
will normally see only a small set of commonly used records, but still
lets them expand the lookup to show all records if/when this is required.<br>
<code>
    property FilterCaption: string;
    property FilterDefault: boolean;
    property FilterHotKey: string;
    property FilterMode: TIB_LookupFilterMode;
    property FilterInvertCB: boolean;
</code><br>
Set the Caption to a string that will be displayed at the bottom of the lookup
controls dropbox.  If not defined the caption will default to "Filtered" or
"Show All", depending on the FilterInvertCB.  For backwards compatibility the
FilterMode will be set to lfmAlways if a FilterCaption is set and the mode
is current at lfmDisabled.<br>
<br>
FilterDefault controls whether the lookup dataset is filtered when first
displayed or not filtered (defaults to true).<br>
<br>
FilterHotKey allows you to define the keystroke that the user can press to
toggle the filter status (defaults to F8).<br>
<br>
FilterMode to is used to control how the filter can be used.  When set to
"lfmDisabled" (the default) the Filter option is not shown and so the
filter cannot be access.  lfmAlways shows the filter option in all modes,
lfhSearchOnly restricts filter changes to search mode.<br>
<br>
FilterInvertCB allows you to reverse the logic of the checkbox displayed
in the lookup control dropbox.  If FilterInvertCB is false (the default)
then the checkbox is checked when the lookup dataset is filtered and
cleared if the lookup dataset is not filtered - in this case an
appropriate FilterCaption might be "Filtered".   If FilterInvertCB is true
then when the checkbox is checked the lookup dataset is not filtered - in
this case an appropriate FilterCaption might be "Show All".<br>
<br>
Of course all of this only works if there is a filter defined to the
lookup dataset.<br>
<br>
Note:  The user may need to be careful when editing records with this
capability enabled - since it is possible that the lookup record currently
in use by the keydataset may not be available in the lookup list.  To
minimise such problems you should have the LookupActivation and
LookupDeactivation properties set to true - or be using a shared lookups.<br>
<br>
<br>
Calculated/Compound Search Fields.<br>
<br>
I use a lookup dataset to set the suburb/zip-code information for an address,
that is; the user looks up a location by name and the address automatically
shows the required zip-code and state details.  This requires that the
lookup control actually shows a calcuated field formed as:<br>
<code>
   location_name || ', ' || zip_code || ' ' ||  state_abbrev
</code><br>
(I use a computed field in the location table to form this string for
easy access, and it is this computed field that I retrieve into the key
datasets for display in grids etc.)<br>
<br>
However during a lookup it is the location name field that is actually
indexed and used for incremental searching (and so it is the location
name field that appears in the Ordering details for the lookup dataset).
This presented a problem where the text that I wanted to display was not
identical to the text used for incremental searching.  To provide a
work-around for this problem I created the following property...<br>
<code>
    property SearchDelimiter: char;
</code><br>
This property defaults to #0 (null) and with this setting the lookup
control expects that the DisplayField, LookupField and first OrderingItem
field will all correspond to the same text result.  In this situation
incremental search takes place as expected on the full field value.<br>
<br>
When the property is set to some other value, the display text is
considered to be "delimited" by that value.  The text before the first
occurance of the delimiter is used for the incremental search (which uses
the first OrderingItem field for its match).  When the incremental search
is done the full text is then used in a brute force search (record by
record matching) against the actual LookupField value in the lookup
dataset - from the point where the incremental search finished.<br>
<br>
In the example above, the LookupField displays the computed value
"location_name, zip_code state_abbrev", whereas the first ordering
field is the location_name.  By setting SearchDelimiter to "," (comma)
the incremental search will only use the LookupField text upto the first
comma for the incremental search against the location_name field.  It
will then do a brute force search from there against the actual
LookupField (the computed value) to find the match required.  This
means that the lookup will therefore always land on the required
record, even if multiple records exist with the same location_name value.<br>
<br>
This feature needs to be used carefully, if there are too many records
with duplicated values in the first OrderingItem field then the brute
force search may take an unreasonably long time.  However in many
situations the additional fields are only there to support unusual or
uncommon situations and there may only ever be a small number of such
duplicated entries.<br>
}
  TIB_CustomLookupEnh = class(TCustomComboEnh)
    {$DEFINE IB_CustomEdit}
    {$DEFINE IB_AutoLabel}
    {$I IB_WinControl.INT}
    {$UNDEF IB_AutoLabel}
    {$UNDEF IB_CustomEdit}
  private
    FDropDownCount: integer;
    FDropDownRows: integer;
    FDropDownWidth: integer;
    FRefreshHotKey: TShortCut;
    FFilterCaption: string;             // default '' = dont show filter CheckBox
    FFilterDefault: boolean;            // default true = filter if available
    FFilterHotKey: TShortCut;           // default 'Ctrl+F'
    FFilterInvertCB: boolean;           // default false = set true to invert CB logic
    FFilterMode: TIB_LookupFilterMode;  // default lfmDisabled - no filter access
    FSkipChar: boolean;
    FColLines: boolean;                 // false
    FDefaultRowHeight: integer;         // 17
    FFixedColLines: boolean;            // true
    FFixedRowLines: boolean;            // true
    FGridLinks: TIB_StringList;         // null
    FIndicateOrdering: boolean;         // true
    FIndicateRow: boolean;              // false
    FIndicateSelected: boolean;         // false
    FIndicateTitles: boolean;           // false
    FRowLines: boolean;                 // false
    FSearchKeyByKey: boolean;           // true
    FSeekNearest: boolean;              // false
    FLookupActivation: boolean;         // true
    FLookupDeactivation: boolean;       // true
    FLookupLink: TIB_FieldDataLink;     //
    FLookupShared: boolean;             // false
    FLookupKeyLinks: TIB_StringList;
    FLookupKeyDescLinks: TIB_StringList;
    FLookupPossession: boolean;
    FSearchDelimiter: char;             // #0
    FDisplayCol: TIB_Column;            //
    FDisplayField: string;
    FInternalRowChg: integer;
    FDataChangeDone: boolean;           // incremental search change flag
    FUserInputFlag: integer;
    FSavedLookupText: string;
    FCallbackAbort: boolean;            // set true to abort callback proc

    function GetLookupField: string;
    procedure SetLookupField(AValue: string);
    function GetLookupSource: TIB_DataSource;
    procedure SetLookupSource(AValue: TIB_DataSource);
    procedure SetDisplayCol;
    procedure LookupDataChange(Sender: TIB_DataLink; IB_DataSource:
      TIB_DataSource; IB_Field: TIB_Column);
    procedure SetGridLinks(AValue: TIB_StringList);
    procedure SetLookupKeyLinks(AValue: TIB_StringList);
    procedure SetLookupKeyDescLinks(AValue: TIB_StringList);
    function GetInUserInput: boolean;

    procedure SetFilterCaption(AValue: string);
    function GetFilterHotKey: string;
    procedure SetFilterHotKey(AValue: string);
    function IsFilterHotKeyStored: boolean;
    function GetRefreshHotKey: string;
    procedure SetRefreshHotKey(AValue: string);
    function IsRefreshHotKeyStored: boolean;
    procedure SetFilterInvertCB(AValue: boolean);
    procedure SetFilterMode(AValue: TIB_LookupFilterMode);

    function IsFilterOptAvailable: boolean;

  protected
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: char); override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    // Must use SETFOCUS/KILLFOCUS messages to be sure of being told when
    // focus is leaving/returning - even when forms change (which CM_ENTER
    // and CM_EXIT do not do).
    //procedure WMSetFocus(var Message: TMessage); message WM_SETFOCUS;
    procedure WMKillFocus(var Message: TMessage); message WM_KILLFOCUS;

    function IncrementalSearch(TxtVal: string; ChkLen: integer): integer;
      override;
    function BruteSearch(SearchVal: string; MatchNear, BufOnly: boolean):
      boolean;

    function CreateComboControl: TCustomComboControlEnh; override;
    procedure DestroyComboControl(CurrentControl: TCustomComboControlEnh);
      override;

  protected
    function GetLookupKeyText: string;
    function IsMaskStored: boolean; override;
    function GetDBDisplayField: TObject; override;
    function IBG_GetAutoLabel: TObject; override;
    procedure DoAfterValidateEdit(var ErrCode: integer); override;
    procedure Loaded; override;
    procedure Change; override;

    procedure ActivateLookup;
    procedure DeactivateLookup;
    procedure SaveSearchInfo(Deactivating: boolean);
    procedure RestoreSearchInfo;

    function GetFilterHotKeyDefault: TShortCut; virtual;
    function GetRefreshHotKeyDefault: TShortCut; virtual;

    // from IB_CustomEdit.INT
    procedure EditingChanged(Sender: TIB_DataLink; IB_DataSource:
      TIB_DataSource); virtual;

    property DropDownCount: integer read FDropDownCount write FDropDownCount
    default 8;
    property DropDownWidth: integer read FDropDownWidth write FDropDownWidth
    default 0;
    property ColLines: boolean read FColLines write FColLines default false;
    property DefaultRowHeight: integer read FDefaultRowHeight write
    FDefaultRowHeight default 17;
    property FilterCaption: string read FFilterCaption write SetFilterCaption;
    property FilterDefault: boolean read FFilterDefault write FFilterDefault
    default true;
    property FilterHotKey: string read GetFilterHotKey write SetFilterHotKey
    stored IsFilterHotKeyStored;
    property FilterInvertCB: boolean read FFilterInvertCB write SetFilterInvertCB
    default false;
    property FilterMode: TIB_LookupFilterMode read FFilterMode write SetFilterMode
    default lfmDisabled;
    property FixedColLines: boolean read FFixedColLines write FFixedColLines
    default true;
    property FixedRowLines: boolean read FFixedRowLines write FFixedRowLines
    default true;
    property GridLinks: TIB_StringList read FGridLinks write SetGridLinks;
    property IndicateOrdering: boolean read FIndicateOrdering write
    FIndicateOrdering default true;
    property IndicateRow: boolean read FIndicateRow write FIndicateRow default
    false;
    property IndicateSelected: boolean read FIndicateSelected write
    FIndicateSelected default false;
    property IndicateTitles: boolean read FIndicateTitles write FIndicateTitles
    default false;
    property RefreshHotKey: string read GetRefreshHotKey write SetRefreshHotKey
    stored IsRefreshHotKeyStored;
    property RowLines: boolean read FRowLines write FRowLines default false;
    property SearchKeyByKey: boolean read FSearchKeyByKey write FSearchKeyByKey
    default true;
    property SearchDelimiter: char read FSearchDelimiter write FSearchDelimiter
    default #0;
    property SeekNearest: boolean read FSeekNearest write FSeekNearest default
    false;
    property LookupActivation: boolean read FLookupActivation write
    FLookupActivation default true;
    property LookupDeactivation: boolean read FLookupDeactivation write
    FLookupDeactivation default true;
    property LookupShared: boolean read FLookupShared write FLookupShared default
    false;
    property LookupKeyLinks: TIB_StringList read FLookupKeyLinks write
    SetLookupKeyLinks;
    property LookupKeyDescLinks: TIB_StringList read FLookupKeyDescLinks write
    SetLookupKeyDescLinks;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure Reset; override;
    function EditCanModify: boolean; override;
    function EditBeginModify: boolean; override;
    function EditEndModify: boolean; virtual;

    procedure SetWorking(DspStr: string);
    procedure ClrWorking;
    procedure WorkingCallback(Sender: TIB_Dataset; Status: TIB_CallbackStatus;
      CursorRowNum: integer; var Abort: boolean);

    // Actual number of rows in dropdown
    property DropDownRows: integer read FDropDownRows;
{: @summary Determines if current processing is the result of a mouse or
keyboard input.
<br>
@desc Useful in the OnChange event to determine if the change was due to
user input or some other processing.}
    property InUserInput: boolean read GetInUserInput;

    // need to be public for registration in separate unit...
    property LookupLink: TIB_FieldDataLink read FLookupLink; //
    property LookupSource: TIB_DataSource read GetLookupSource write
    SetLookupSource;
    property LookupField: string read GetLookupField write SetLookupField;
    property DisplayField: string read FDisplayField write FDisplayField;

  published

  end;                                  // TIB_CustomComboEnh

{: @summary Published variation of TIB_CustomLookupEnh.
<br>
@desc For details and example please <see Class=TIB_CustomLookupEnh>}
  TIB_LookupEnh = class(TIB_CustomLookupEnh)
    {$DEFINE IB_AutoLabel}
    {$I IB_WinControl.PBL}
    {$UNDEF IB_AutoLabel}
  public
    property IsMasked;
    property IsEmpty;
    property DisplayText;
    property ButtonState;

  published
    property DropDownCount;
    property DropDownWidth;
    property ColLines;
    property DefaultRowHeight;
    property FilterCaption;
    property FilterDefault;
    property FilterHotKey;
    property FilterInvertCB;
    property FilterMode;
    property FixedColLines;
    property FixedRowLines;
    property GridLinks;
    property IndicateOrdering;
    property IndicateRow;
    property IndicateSelected;
    property IndicateTitles;
    property RefreshHotKey;
    property RowLines;
    property SearchKeyByKey;
    property SearchDelimiter;
    property SeekNearest;
    property LookupActivation;
    property LookupDeactivation;
    property LookupField;
    property LookupSource;
    property LookupShared;
    property LookupKeyLinks;
    property LookupKeyDescLinks;
    property DisplayField;

    // From TCustomComboEnh
    property ActiveWhenReadOnly;
    property AutoDropDown;
    property DropDownColor;
    property DropDownFont;
    property DropDownFontAuto;
    property OnCloseUp;
    property BeforeDropDown;
    property OnDropDown;

    // Stuff we want from CustomEdit.PBL
    property AutoSelect;
    property AutoSize;
    property HideSelection;
    property OEMConvert;
    property OnChange;

    // TCustomEditEnh publications...
    //property Alignment;      // only set via dataset/connection
    property Button;
    //property CharCase;       // only set via dataset/connection
    property Ctl3DShallow;
    //property EditMask;       // only set via dataset/connection
    //property EditMaskMode;   // only set via dataset/connection
    property GlyphAlign;
    property GlyphSize;
    property Options;
    //property Text;           // only set via dataset/connection

    property OnButton1;
    property OnButton2;
    property OnButton3;
    property OnDrawButton;
    property AfterValidateEdit;
    property OnValidateEdit;
    property OnValidateError;
    property OnGlyphClick;
    property OnDrawGlyph;

  end;                                  // TIB_LookupEnh

var
  IB_ComboLookupEnh_ListBox: TIB_ComboLookupListEnh;
  IB_ComboLookupEnh_ListBoxRefCount: integer;


implementation

uses Telesis_Utils, Telesis_Locale, Telesis_Constants;


type


  TIB_InternalLookupListEnh = class(TIB_LookupList)
  private
    FCellColor: TColor;
    procedure CMShowingChanged(var Message: TMessage); message CM_SHOWINGCHANGED;
  protected
    function DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): boolean; override;
    function DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): boolean; override;
    function DoMouseWheel(Shift: TShiftState; WheelDelta: integer; MousePos: TPoint): boolean; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure WMVScroll(var Msg: TWMVScroll); message WM_VSCROLL;
    procedure WMHScroll(var Msg: TWMHScroll); message WM_HSCROLL;

  public
    procedure DefaultDrawCell(ACol, ARow: Longint;
      ARect: TRect;
      AState: TGridDrawState;
      CellDisplayText: string;
      CellAlignment: TAlignment); override;
    procedure GetCellProps(ACol, ARow: Longint;
      AState: TGridDrawState;
      var AColor: TColor;
      AFont: TFont); override;

  public
    constructor Create(AOwner: TComponent); override;

    property CellColor: TColor read FCellColor write FCellColor default clWindow;
    property ColCount;
    property FixedCols;
    property ColWidths;
    property RowHeights;

  end;

  // need to override WndProc for TCheckBox so to remove the VCL dependance
  // on TCheckBox having focus before accepting mouse messages (part of the
  // TButtonControl base to TCheckBox).
  TCheckBoxNoFocus = class(TCheckBox)
  protected
    procedure WndProc(var Message: TMessage); override;
    procedure WMLButtonDown(var Message: TWMLButtonDown); message
      WM_LBUTTONDOWN;
    procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
  end;


procedure TCheckBoxNoFocus.WndProc(var Message: TMessage);
begin
  if Message.Msg = WM_LBUTTONDOWN then
  begin
    WMLButtonDown(TWMLButtonDown(Message));
  end
  else
    if Message.Msg = WM_LBUTTONUP then
    begin
      WMLButtonUp(TWMLButtonUp(Message));
    end
    else
      inherited;
end;

procedure TCheckBoxNoFocus.WMLButtonDown(var Message: TWMLButtonDown);
begin
  inherited;
end;

procedure TCheckBoxNoFocus.WMLButtonUp(var Message: TWMLButtonUp);
begin
  inherited;
end;



constructor TIB_InternalLookupListEnh.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  inherited ScrollBars := ssVertical;
  // It is replicatable because there is only one!
  ControlStyle := ControlStyle + [csReplicatable];
  FIgnoreFocus := true;
  ReadOnly := true;
  // ??? Not sure about this - different to lookup list, but with new
  // grid work for IBO4 it appears to be necessary in my lookup.
  TrackGridRow := false;
end;

function TIB_InternalLookupListEnh.DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): boolean;
begin
  Result := inherited DoMouseWheelDown(Shift, MousePos);
end;

function TIB_InternalLookupListEnh.DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): boolean;
begin
  Result := inherited DoMouseWheelUp(Shift, MousePos);
end;

function TIB_InternalLookupListEnh.DoMouseWheel(Shift: TShiftState; WheelDelta: integer; MousePos: TPoint): boolean;
begin
  Result := inherited DoMouseWheel(Shift, WheelDelta, MousePos);
end;

procedure TIB_InternalLookupListEnh.MouseDown(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  tmpController: TIB_CustomLookupEnh;
begin
  tmpController := TIB_CustomLookupEnh(TIB_ComboLookupListEnh(Parent).Controller);
  Inc(tmpController.FUserInputFlag);
  try
    inherited;
  finally
    Dec(tmpController.FUserInputFlag);
  end;
end;

procedure TIB_InternalLookupListEnh.MouseMove(Shift: TShiftState; X, Y:
  Integer);
//var
//    i: integer;
//    Pt: TPoint;
begin
  inherited;
  // In ComboBox we dynamically move the pointer - but we can do that without
  // actually impacting the control value.  This is not possible here, unless
  // additional work is done on the grid.  I'll review this when the rest of
  // the control is working.
  //Pt.X := X;
  //Pt.Y := Y;
  //i := ItemAtPos( Pt, false );
  //if i >= 0 then begin
  //    ItemIndex := i;
  //    end
  //else ItemIndex := -1;
end;

procedure TIB_InternalLookupListEnh.MouseUp(Button: TMouseButton; Shift:
  TShiftState; X, Y: Integer);
var
  Cell: TGridCoord;
  DRow: integer;
  tmpController: TIB_CustomLookupEnh;
  //i: integer;
  //Pt: TPoint;
begin
  tmpController := TIB_CustomLookupEnh(TIB_ComboLookupListEnh(Parent).Controller);
  Inc(tmpController.FUserInputFlag);
  try
    inherited;
    Cell := MouseCoord(X, Y);
    // Only process the mouse event if the cell corresponds to an actual
    // data-cell (as opposed to a fixed/border cell)
    if (Cell.X >= FixedCols) and (Cell.Y >= FixedRows) then
    begin
      DRow := DataRow[Cell.Y];
      if (DataSource <> nil) and (DataSource.Dataset <> nil) and
        (DataSource.Dataset.RowNum <> DRow) then
      begin
        DataSource.Dataset.RowNum := DRow;
        TIB_ComboLookupListEnh(Parent).Change;
      end;
      TIB_ComboLookupListEnh(Parent).CloseUp;
    end;
  finally
    Dec(tmpController.FUserInputFlag);
  end;
end;

procedure TIB_InternalLookupListEnh.CMShowingChanged(var Message: TMessage);
begin
  inherited;
  // to fix what inherited will have done
  if Showing then UpdateScrollBar;
end;

procedure TIB_InternalLookupListEnh.WMVScroll(var Msg: TWMVScroll);
var
  tmpController: TIB_CustomLookupEnh;
begin
  tmpController := TIB_CustomLookupEnh(TIB_ComboLookupListEnh(Parent).Controller);
  Inc(tmpController.FUserInputFlag);
  try
    inherited;
  finally
    Dec(tmpController.FUserInputFlag);
  end;
end;

procedure TIB_InternalLookupListEnh.WMHScroll(var Msg: TWMHScroll);
var
  tmpController: TIB_CustomLookupEnh;
begin
  tmpController := TIB_CustomLookupEnh(TIB_ComboLookupListEnh(Parent).Controller);
  Inc(tmpController.FUserInputFlag);
  try
    inherited;
  finally
    Dec(tmpController.FUserInputFlag);
  end;
end;

procedure TIB_InternalLookupListEnh.DefaultDrawCell(ACol, ARow: Longint;
  ARect: TRect;
  AState: TGridDrawState;
  CellDisplayText: string;
  CellAlignment: TAlignment);
var
  AColumn: TIB_Column;
  TxtCol: TColor;
  i, j: integer;
begin
  inherited;
  AColumn := nil;
  if DataLink.Prepared then
  begin
    if (DataLink.ActiveRecord = DataRow[ARow]) or
      (DataLink.BufferRowCount = 0) then
      AColumn := GridFields[DataCol[ACol]]
    else
    begin
      DataLink.BufferRecord := DataRow[ARow];
      if DataLink.Dataset.BufferFields.RowState <> rsNone then
        AColumn := BuffFields[DataCol[ACol]];
    end;
  end;
  if IndicateBooleans and
    (ARow >= FixedRows) and
    Assigned(AColumn) and AColumn.IsBoolean and
    (gdSelected in AState) then
  begin
    // Need to update the boolean glyph color so it can
    // be seen when highlighted.
    with Canvas do
    begin
      TxtCol := TextColorFromColor(Pixels[ARect.Left + 1, ARect.Top + 1]);
      for i := ARect.Left to ARect.Right do
      begin
        for j := ARect.Top to ARect.Bottom do
        begin
          if Pixels[i, j] = clBlack then
            Pixels[i, j] := TxtCol
          else
            if Pixels[i, j] = clGray then
              Pixels[i, j] := clBtnShadow
            else
              if Pixels[i, j] = clSilver then
                Pixels[i, j] := clBtnFace;
        end;
      end;
    end;
  end;
end;

procedure TIB_InternalLookupListEnh.GetCellProps(ACol, ARow: Longint;
  AState: TGridDrawState;
  var AColor: TColor;
  AFont: TFont);
begin
  AColor := FCellColor;
  inherited;
  if (gdSelected in AState) and DataLink.Prepared and
    (DataLink.ActiveRecord = DataRow[ARow]) then
  begin
    AColor := clHighlight;
    if AFont <> nil then
      AFont.Color := clHighlightText;
  end;
end;


constructor TIB_ComboLookupListEnh.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FWorkPnl := TPanel.Create(Self);
  FWorkPnl.Parent := Self;
  FWorkPnl.Visible := false;
  FWorkPnl.BevelInner := bvLowered;
  FWorkPnl.BevelOuter := bvNone;
  FWorkPnl.BevelWidth := 2;
  FWorkPnl.Caption := '';
  //FWorkPnl.Align := alBottom; We dont want the appearance of the working
  // panel to change the size of the grid - this upsets various things
  // as the grid is forced to update.  So we will size and place this
  // as necessary.
  FWorkPnl.Alignment := taLeftJustify;
  FWorkPnl.Color := clHighlight;

  FFilterPnl := TPanel.Create(Self);
  FFilterPnl.Parent := Self;
  FFilterPnl.Visible := false;
  FFilterPnl.BevelInner := bvLowered;
  FFilterPnl.BevelOuter := bvNone;
  FFilterPnl.BevelWidth := 1;
  FFilterPnl.BorderWidth := 2;
  FFilterPnl.Caption := '';
  FFilterPnl.Align := alBottom;
  FFilterPnl.Alignment := taLeftJustify;
  //??FFilterPnl.Color := clHighlight;
  FFilterPnl.ParentColor := true;

  FFilterCB := TCheckBoxNoFocus.Create(FFilterPnl);
  FFilterCB.Parent := FFilterPnl;
  FFilterCB.Visible := true;
  FFilterCB.Caption := '';
  FFilterCB.Checked := false;
  FFilterCB.OnClick := OnFilterCBChange;
  //??FFilterPnl.Alignment := taRightJustify; default

  FList := TIB_InternalLookupListEnh.Create(Self);
  ActiveWhenInvisible := false; // We use a shared listbox so cannot be active
  // until visible
  FList.Parent := Self;
  FList.Align := alClient;
  FList.BorderStyle := bsNone;
  FList.Visible := true;
  //FList.ExtendedSelect := false;
  //FList.MultiSelect := false;
end;

destructor TIB_ComboLookupListEnh.Destroy;
begin
  if FList <> nil then
  begin
    FList.Free;
    FList := nil;
  end;
  inherited Destroy;
end;

procedure TIB_ComboLookupListEnh.OnFilterCBChange(Sender: TObject);
begin
  ToggleFilterCB;
end;

procedure TIB_ComboLookupListEnh.InitFilterCB;
begin
  if not FFilterInternal then
  try
    FFilterInternal := true;
    if not ((FList.DataSource = nil) or (FList.DataSource.Dataset = nil)) then
    begin
      with Controller as TIB_CustomLookupEnh do
      begin
        if FFilterInvertCB then
          FFilterCB.Checked := not
            TIB_BDataset(FList.DataSource.Dataset).Filtered
        else
          FFilterCB.Checked := TIB_BDataset(FList.DataSource.Dataset).Filtered;
        if (FFilterCaption = '') then
        begin
          if FFilterInvertCB then
            FFilterCB.Caption := 'Show All'
          else
            FFilterCB.Caption := 'Filtered';
        end
        else
          FFilterCB.Caption := FFilterCaption;
      end;
    end;
  finally
    FFilterInternal := false;
  end;
end;

procedure TIB_ComboLookupListEnh.ToggleFilterCB;
begin
  if not FFilterInternal then
  try
    FFilterInternal := true;
    // check availability just to be sure
    if (Controller as TIB_CustomLookupEnh).IsFilterOptAvailable then
    begin
      TIB_BDataset(FList.DataSource.Dataset).Filtered :=
        not TIB_BDataset(FList.DataSource.Dataset).Filtered;
      with Controller as TIB_CustomLookupEnh do
      begin
        if FFilterInvertCB then
          FFilterCB.Checked := not
            TIB_BDataset(FList.DataSource.Dataset).Filtered
        else
          FFilterCB.Checked := TIB_BDataset(FList.DataSource.Dataset).Filtered;
      end;
      SyncToParentValue;
    end;
  finally
    FFilterInternal := false;
  end;
end;

function TIB_ComboLookupListEnh.DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): boolean;
begin
  Result := TIB_InternalLookupListEnh(FList).DoMouseWheelDown(Shift, MousePos);
end;

function TIB_ComboLookupListEnh.DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): boolean;
begin
  Result := TIB_InternalLookupListEnh(FList).DoMouseWheelUp(Shift, MousePos);
end;

function TIB_ComboLookupListEnh.DoMouseWheel(Shift: TShiftState; WheelDelta: integer; MousePos: TPoint): boolean;
begin
  Result := TIB_InternalLookupListEnh(FList).DoMouseWheel(Shift, WheelDelta, MousePos);
end;

procedure TIB_ComboLookupListEnh.SizeControl;
begin
  // I all sizing in the SyncToParent value since the final sizes may not
  // be available until we have been synchronised with the controller properties.
end;

procedure TIB_ComboLookupListEnh.SyncToParentValue;
var
  i, tmpColW, tmpRowH, tmpColCount: integer;
  CanFilter: boolean;
begin
  CanFilter := false;
  if FInternalChange = 0 then
  begin
    // We are using a shared listbox, so clear and re-initialise the listbox
    // contents first...
    ClrWorking;
    with Controller as TIB_CustomLookupEnh do
    begin
      if LookupLink.Dataset <> nil then
        LookupLink.Dataset.BeginCallbackFreeze;
      try
        FList.DataSource := nil;          // disable while we do this
        ActivateLookup;
        Self.SyncColors;
        FList.ColLines := ColLines;
        FList.DefaultRowHeight := DefaultRowHeight;
        FList.FixedColLines := FixedColLines;
        FList.FixedRowLines := FixedRowLines;
        FList.GridLinks := GridLinks;
        FList.IndicateOrdering := IndicateOrdering;
        FList.IndicateRow := IndicateRow;
        FList.IndicateSelected := IndicateSelected;
        FList.IndicateTitles := IndicateTitles;
        FList.RowLines := RowLines;
        FList.SearchKeyByKey := SearchKeyByKey;
        FList.SeekNearest := SeekNearest;
        FList.ListBoxStyle := false;
        FList.DataSource := LookupLink.DataSource; // re-enable
        if (LookupLink.Dataset <> nil) and LookupLink.Dataset.Active then
        begin
          LookupLink.Dataset.ValidateRows(
            LookupLink.Dataset.RowNum - DropDownCount,
            LookupLink.Dataset.RowNum + DropDownCount );
          if (LookupLink.Dataset.RowNum <> LookupLink.Dataset.BofRowNum) and
            (LookupLink.Dataset.RowNum <> LookupLink.Dataset.EofRowNum) then
            FList.TopRowNum := LookupLink.Dataset.RowNum; //FList.GridRow[];
        end;
        FFilterPnl.Height := DefaultRowHeight + (4 + 4);
        // Set height only to determine if scrollbars are necessary
        // Change the dropdown height according to available rows...
        if (LookupLink.Dataset = nil) or (not LookupLink.Dataset.Active) then
        begin
          FDropDownRows := 1;
        end
        else
        begin
          with LookupLink.Dataset as TIB_BDataset do
          begin
            // force at least DropDownCount rows to be loaeded...
            if not BufferHasEof then
            begin
              BufferRowNum := DropDownCount;
            end;
            if BufferHasEof then
            begin
              if BufferRowCount > DropDownCount then
              begin
                FDropDownRows := DropDownCount;
              end
              else
              begin
                if BufferRowCount > 0 then
                begin
                  FDropDownRows := BufferRowCount;
                end
                else
                begin
                  FDropDownRows := 1;
                end;
              end;
            end
              // If we dont know how many make full height
            else
              FDropDownRows := DropDownCount;
            // while we are here, check the filter requirements...
            if IsFilterOptAvailable then
            begin
              CanFilter := true;
              InitFilterCB;
            end;
          end;
        end;
        if CanFilter then
          FFilterPnl.Visible := true
        else
          FFilterPnl.Visible := false;
        // not reliable
        //tmpRowH := FList.HeightForGridRows(DropDownRows);
        // so use:
        with FList as TIB_InternalLookupListEnh do
        begin
          tmpRowH := 0;
          for i := 0 to FixedRows - 1 do
            Inc( tmpRowH, RowHeights[i] );
          if RowLines or FixedRowLines then
            Inc( tmpRowH, GridLineWidth * ( FixedRows + DropDownRows + 1 ));
          Inc( tmpRowH, DefaultRowHeight * DropDownRows );
        end;
        //if HandleAllocated then  - no border so no need for this
        //  Inc( Result, Height - ClientHeight )
        if FFilterPnl.Visible then
          tmpRowH := tmpRowH + FFilterPnl.Height;
        Self.Height := (Self.Height - Self.ClientHeight) +
          (2 * (Self.BevelWidth + Self.BorderWidth)) + tmpRowH;
        // Set drop down width
        if DropDownWidth <= 0 then
        begin
          if (LookupLink.Dataset <> nil) and LookupLink.Dataset.Active then
          begin
            tmpColW := 0;
            // cannot use WidthForAllCols due to ListBox style
            with FList as TIB_InternalLookupListEnh do
            begin
              // cant seem to rely on ColCount in first call, so we calculate
              // the effective count for ourselves.
              TmpColCount := FixedCols - FixedDataCols + GridFieldCount;
              if TmpColCount <= FixedCols then
                TmpColCount := FixedCols + 1;
              for i := 0 to TmpColCount - 1 do
              begin
                if (i >= FixedCols) and (i < GridFieldCount) then //(i > 0) and ??
                  Inc(tmpColW, GridFields[i - FixedCols].DisplayWidth)
                else
                  Inc(tmpColW, ColWidths[i]);
              end;
              if ColLines or FixedColLines then
              begin
                Inc(tmpColW, GridLineWidth * (ColCount + 1));
              end;
              // This gives inconsistent results
              //?? Inc(tmpColW, Width - ClientWidth);
              // so always allow for vertical scroll 
              Inc(tmpColW, GetSystemMetrics( SM_CXVSCROLL ));
            end;
            Self.Width := (Self.Width - Self.ClientWidth) +
              (2 * (Self.BevelWidth + Self.BorderWidth)) + tmpColW;
          end
          else
            Self.Width := Controller.Width;
        end
        else
        begin
          Self.Width := DropDownWidth;
          FList.ListBoxStyle := true;
        end;
        FFilterCB.Left := 6;
        FFilterCB.Top := 4;
        FFilterCB.Width := FFilterPnl.ClientWidth - (6 + 4);
        // and now we set the height again
        // not reliable
        //tmpRowH := FList.HeightForGridRows(DropDownRows);
        // so use:
        with FList as TIB_InternalLookupListEnh do
        begin
          tmpRowH := 0;
          for i := 0 to FixedRows - 1 do
            Inc( tmpRowH, RowHeights[i] );
          if RowLines or FixedRowLines then
            Inc( tmpRowH, GridLineWidth * ( FixedRows + DropDownRows + 1 ));
          Inc( tmpRowH, DefaultRowHeight * DropDownRows );
        end;
        if FFilterPnl.Visible then
          tmpRowH := tmpRowH + FFilterPnl.Height;
        Self.Height := (Self.Height - Self.ClientHeight) +
          (2 * (Self.BevelWidth + Self.BorderWidth)) + tmpRowH;
        FWorkPnl.Height := DefaultRowHeight + 4;
        FWorkPnl.Top := Self.ClientHeight -
          (FWorkPnl.Height + Self.BevelWidth + Self.BorderWidth);
        FWorkPnl.Left := (Self.BevelWidth + Self.BorderWidth);
        FWorkPnl.Width := Self.ClientWidth - (2 * (Self.BevelWidth +
          Self.BorderWidth));
        FWorkPnl.ParentFont := true;
        FWorkPnl.Font.Style := FWorkPnl.Font.Style + [fsBold];
        FWorkPnl.Font.Color := clHighlightText;
      finally
        if LookupLink.Dataset <> nil then
          LookupLink.Dataset.EndCallbackFreeze;
      end;
    end;
  end
  else
  begin
    // Actual synchronisation process takes place with the datasets and
    // so there is nothing we need do here.
  end;
end;

procedure TIB_ComboLookupListEnh.SyncParentToValue;
begin
  // Actual synchronisation process takes place with the datasets and
  // so there is nothing we need do here.
end;

procedure TIB_ComboLookupListEnh.SyncColors;
begin
  with Controller as TCustomComboEnh do
  begin
    if DropDownColor = clNone then
      Self.FList.Color := Color
    else
      Self.FList.Color := DropDownColor;
  end;
  TIB_InternalLookupListEnh(FList).CellColor := FList.Color;
  TIB_InternalLookupListEnh(FList).OrderingColor := FList.Color;
end;

procedure TIB_ComboLookupListEnh.CloseUp;
begin
  inherited;
  // if a shared lookup and we no longer have possession then clear
  // the list dataset to prevent receipt of inappropriate messages
  if Assigned(Controller) then
    with Controller as TIB_CustomLookupEnh do
    begin
      if FLookupPossession and          //LookupShared and
        (FList.DataSource = LookupLink.DataSource) then
        if FList.DataSource.Active and FList.DataSource.Dataset.Fetching then
          FList.DataSource.Dataset.AbortFetching;
      FList.DataSource := nil;
    end;
end;

procedure TIB_ComboLookupListEnh.SetWorking(DspStr: string);
begin
  if DspStr = '' then
    DspStr := ENHSTR_Lookup_Searching;
  FWorkPnl.Caption := DspStr;
  FWorkPnl.Visible := true;
  FWorkPnl.BringToFront;
end;

procedure TIB_ComboLookupListEnh.ClrWorking;
begin
  FWorkPnl.Caption := '';
  FWorkPnl.Visible := false;
end;


{$DEFINE IB_CustomEdit}
{$DEFINE IB_AutoLabel}
{$I IB_WinControl.IMP}
{$UNDEF IB_AutoLabel}
{$UNDEF IB_CustomEdit}


constructor TIB_CustomLookupEnh.Create(AOwner: TComponent);
begin
  FCallbackAbort := false;
  FIB_FieldDataLink := TIB_FieldDataLink.Create(Self);
  FLookupLink := TIB_FieldDataLink.Create(Self);
  FAutoLabel := TIB_AutoLabel.Create(Self);
  FDisplayCol := nil;
  FLookupShared := false;
  FLookupKeyLinks := TIB_StringList.Create;
  FLookupKeyDescLinks := TIB_StringList.Create;
  inherited Create(AOwner);
  FDropDownCount := 8;
  FDropDownWidth := 0;
  FColLines := false;
  FDefaultRowHeight := 17;
  FFilterCaption := '';
  FFilterDefault := true;
  FFilterHotKey := GetFilterHotKeyDefault;
  FFilterInvertCB := false;
  FFilterMode := lfmDisabled;
  FFixedColLines := true;
  FFixedRowLines := true;
  FGridLinks := TIB_StringList.Create;
  FIndicateOrdering := true;
  FIndicateRow := false;
  FIndicateSelected := false;
  FIndicateTitles := false;
  FRefreshHotKey := GetRefreshHotKeyDefault;
  FRowLines := false;
  FSearchKeyByKey := true;
  FSearchDelimiter := #0;
  FSeekNearest := false;
  FLookupActivation := true;
  FLookupDeactivation := true;
  FLookupPossession := false;
  FDisplayField := '';
  FInternalRowChg := 0;
  FDataChangeDone := false;
  FUserInputFlag := 0;
  FSavedLookupText := '';
  DataLink.Control := Self;
  DataLink.OnStateChanged := StateChanged;
  DataLink.OnEditingChanged := EditingChanged;
  DataLink.OnDataChange := DataChange;
  DataLink.OnUpdateData := UpdateData;
  FLookupLink.OnDataChange := LookupDataChange;
  // Somebody plays with BorderStyle in the generic units, default it
  // back to where it belongs...
  BorderStyle := bsSingle;
end;

destructor TIB_CustomLookupEnh.Destroy;
begin
  //??if (ComboControl <> nil) then
  //??    ComboControl.CloseUp;
  FAutoLabel.Free;
  FAutoLabel := nil;
  FDisplayCol := nil;
  DataSource := nil;
  DataLink.Control := nil;
  DataLink.OnStateChanged := nil;
  DataLink.OnEditingChanged := nil;
  DataLink.OnDataChange := nil;
  DataLink.OnUpdateData := nil;
  FLookupLink.DataSource := nil;
  FLookupLink.Control := nil;
  FLookupLink.OnDataChange := nil;
  if FGridLinks <> nil then
  begin
    FGridLinks.Free;
    FGridLinks := nil;
  end;
  if FLookupKeyLinks <> nil then
  begin
    FLookupKeyLinks.Free;
    FLookupKeyLinks := nil;
  end;
  if FLookupKeyDescLinks <> nil then
  begin
    FLookupKeyDescLinks.Free;
    FLookupKeyDescLinks := nil;
  end;
  inherited Destroy;
end;

function TIB_CustomLookupEnh.GetLookupField: string;
begin
  Result := FLookupLink.FieldName;
end;

procedure TIB_CustomLookupEnh.SetLookupField(AValue: string);
begin
  FLookupLink.FieldName := AValue;
end;

function TIB_CustomLookupEnh.GetLookupSource: TIB_DataSource;
begin
  Result := FLookupLink.DataSource;
end;

procedure TIB_CustomLookupEnh.SetLookupSource(AValue: TIB_DataSource);
begin
  FLookupLink.DataSource := AValue;
end;

procedure TIB_CustomLookupEnh.SetGridLinks(AValue: TIB_StringList);
begin
  FGridLinks.Text := AValue.Text;
end;

procedure TIB_CustomLookupEnh.SetLookupKeyLinks(AValue: TIB_StringList);
begin
  FLookupKeyLinks.Text := AValue.Text;
end;

procedure TIB_CustomLookupEnh.SetLookupKeyDescLinks(AValue: TIB_StringList);
begin
  FLookupKeyDescLinks.Text := AValue.Text;
end;

function TIB_CustomLookupEnh.GetInUserInput: boolean;
begin
  Result := (FUserInputFlag <> 0);
end;

procedure TIB_CustomLookupEnh.SetFilterCaption(AValue: string);
begin
  if (AValue <> FFilterCaption) then
  begin
    // for backward compatibility (and convenience) set filtermode
    // automatically if initialising the filter caption.
    if (FFilterCaption = '') and (FFilterMode = lfmDisabled) then
      FFilterMode := lfmAlways;
    FFilterCaption := AValue;
    if DroppedDown then
      TIB_ComboLookupListEnh(ComboControl).FFilterCB.Caption := AValue;
  end;
end;

function TIB_CustomLookupEnh.GetFilterHotKey: string;
begin
  Result := ShortCutToText(FFilterHotKey);
end;

procedure TIB_CustomLookupEnh.SetFilterHotKey(AValue: string);
begin
  FFilterHotKey := TextToShortCut(AValue);
end;

function TIB_CustomLookupEnh.IsFilterHotKeyStored: boolean;
begin
  Result := (FFilterHotKey <> GetFilterHotKeyDefault);
end;

function TIB_CustomLookupEnh.GetRefreshHotKey: string;
begin
  Result := ShortCutToText(FRefreshHotKey);
end;

procedure TIB_CustomLookupEnh.SetRefreshHotKey(AValue: string);
begin
  FRefreshHotKey := TextToShortCut(AValue);
end;

function TIB_CustomLookupEnh.IsRefreshHotKeyStored: boolean;
begin
  Result := (FRefreshHotKey <> GetRefreshHotKeyDefault);
end;

procedure TIB_CustomLookupEnh.SetFilterInvertCB(AValue: boolean);
begin
  if AValue <> FFilterInvertCB then
  begin
    FFilterInvertCB := AValue;
    TIB_ComboLookupListEnh(ComboControl).InitFilterCB;
  end;
end;

procedure TIB_CustomLookupEnh.SetFilterMode(AValue: TIB_LookupFilterMode);
begin
  if AValue <> FFilterMode then
  begin
    FFilterMode := AValue;
    TIB_ComboLookupListEnh(ComboControl).InitFilterCB;
  end;
end;

function TIB_CustomLookupEnh.IsFilterOptAvailable: boolean;
begin
  if (LookupLink = nil) or (LookupLink.Dataset = nil) or
    not (LookupLink.Dataset is TIB_BDataset) or
    ((LookupLink.Dataset as TIB_BDataset).Filter = '') then
    Result := false
  else
  begin
    case FilterMode of
    lfmDisabled:   Result := false;
    lfmAlways:     Result := true;
    lfmSearchOnly: Result :=
      ((DataLink <> nil) and (DataLink.State in [dssSearch]));
    else
      Result := false;
    end;
  end;
end;


procedure TIB_CustomLookupEnh.Loaded;
begin
  inherited Loaded;
  SysStateChanged;
  SysDataChange(nil);
  RemakeAutoLabel;
end;

procedure TIB_CustomLookupEnh.ActivateLookup;

  procedure CheckMasterActive(DSet: TIB_Dataset);
  begin
    if assigned(DSet.MasterSource) then
    begin
      DSet := DSet.MasterSource.Dataset;
      if assigned(DSet) and (not DSet.Active) then
      begin
        CheckMasterActive(DSet);
        DSet.Open;
      end;
    end;
  end;
begin
  if not (FLookupPossession and LookupLink.Active) then
  begin
    // First check whether the we really want the lookup active right now...
    if ( DataLink.Active or
         (DataLink.State in [dssInsert, dssPrepared, dssEdit, dssSearch]) or
         SearchAlways or
         DataLink.IsParamEdit
       ) and
      (LookupLink.Dataset <> nil) and
      (LookupShared or LookupActivation) then
    begin
      LookupLink.Dataset.BeginCallbackFreeze;
      try
        // Need to make sure that any "default" connection/transaction
        // settings have been initialised.
        if DataLink.Dataset.IB_Connection = nil then
          DataLink.Dataset.CheckConnection(false);
        if DataLink.Dataset.IB_Transaction = nil then
          DataLink.Dataset.CheckTransaction(false);
        if LookupLink.Dataset.IB_Connection = nil then
          LookupLink.Dataset.CheckConnection(false);
        if LookupLink.Dataset.IB_Transaction = nil then
          LookupLink.Dataset.CheckTransaction(false);
        // Only change the lookup parameters if necessary (to avoid a full
        // reprepare if it is not really necessary...
        if LookupShared and (
          (LookupLink.Dataset.KeyDataset <> DataLink.Dataset) or
          (LookupLink.Dataset.KeyLinks.Text <> FLookupKeyLinks.Text) or
          (LookupLink.Dataset.KeyDescLinks.Text <> FLookupKeyDescLinks.Text)
          ) then
        begin
          if LookupLink.Dataset.Fetching then
            LookupLink.Dataset.AbortFetching;
          LookupLink.Dataset.Close;
          if DataLink.IsParamEdit then
            LookupLink.Dataset.KeySource := nil
          else
            LookupLink.Dataset.KeySource := DataLink.DataSource;
          LookupLink.Dataset.IB_Connection := DataLink.Dataset.IB_Connection;
          LookupLink.Dataset.IB_Transaction := DataLink.Dataset.IB_Transaction;
          LookupLink.Dataset.KeyLinks := FLookupKeyLinks;
          LookupLink.Dataset.KeyDescLinks := FLookupKeyDescLinks;
        end;
        // If we are not already active then...
        if not LookupLink.Active then
        begin
          // Only activate if the connection is available...
          if (LookupLink.Dataset.IB_Connection <> nil) and
            (LookupLink.Dataset.IB_Transaction <> nil) and
            (LookupLink.Dataset.IB_Connection.Connected) then
          begin
            // ?? 15/3/2006 This was clearing any search, but seems not to be
            // now now sure what fixed it, but the line does need to be here
            // otherwise the lookups dont sync on first activate.
            LookupLink.Dataset.KeySource := nil;
            // So try just disabling controls while we play
            LookupLink.Dataset.DisableControls;
            try
              // If we are activating the lookup then we control the
              // initial value of the filtered setting
              if not( (FilterMode = lfmDisabled) or
                (TIB_BDataset(LookupLink.Dataset).Filter = '') ) then
              begin
                TIB_BDataset(LookupLink.Dataset).Filtered := FFilterDefault;
                TIB_ComboLookupListEnh(ComboControl).InitFilterCB;
              end;
              CheckMasterActive(LookupLink.Dataset);
              LookupLink.Dataset.Open;
              //LookupLink.Dataset.RowNum := LookupLink.Dataset.BofRowNum;
              (LookupLink.Dataset as TIB_BDataset).GotoNull;
              if DataLink.IsParamEdit then
                LookupLink.Dataset.KeySource := nil
              else
                LookupLink.Dataset.KeySource := DataLink.DataSource;
            finally
              LookupLink.Dataset.EnableControls;
            end;
          end;
        end;
        if (DataLink.State in [dssInsert, dssEdit]) and
          (FilterMode = lfmSearchOnly) and
          (LookupLink.Dataset is TIB_BDataset) and
          (TIB_BDataset(LookupLink.Dataset).Filter <> '') and
          (TIB_BDataset(LookupLink.Dataset).Filtered <> FFilterDefault) then
        begin
          TIB_BDataset(LookupLink.Dataset).Filtered := FFilterDefault;
          if (ComboControl <> nil) and ComboControl.Active and
            (not TIB_ComboLookupListEnh(ComboControl).FFilterInternal) then
            TIB_ComboLookupListEnh(ComboControl).InitFilterCB;
        end;
        RestoreSearchInfo;
      finally
        LookupLink.Dataset.EndCallbackFreeze;
      end;
    end;
    if LookupLink.Active then
      FLookupPossession := true;
  end;
end;

procedure TIB_CustomLookupEnh.DeactivateLookup;
begin
  FCallbackAbort := true; // make sure any lookup callback is cancelled
  if FLookupPossession and LookupLink.Active and
    (LookupDeactivation or LookupShared) then
  begin
    FLookupPossession := false;         // indicate that we dont want it anymore.
    // Only close the dataset if we are not longer editing
    // or if we are sharing...
    if LookupShared or not (
      (DataLink.State = dssSearch) or
      (DataLink.State = dssInsert) or
      (DataLink.State = dssEdit)) then
    begin
      if LookupLink.Dataset.Fetching then
        LookupLink.Dataset.AbortFetching;
      SaveSearchInfo(true);
      LookupLink.Dataset.Close;
    end;
    if (ComboControl <> nil) then
      ComboControl.CloseUp;
  end
  else
    FLookupPossession := false;
end;

procedure TIB_CustomLookupEnh.SaveSearchInfo(Deactivating: boolean);
begin
  // Save bookmark data for search mode before deactivation
  if LookupLink.Active and (LookupLink.Dataset <> nil) and
    (DataLink.State = dssSearch) then
  begin
    if (LookupLink.Dataset.RowNum = LookupLink.Dataset.BofRowNum) or
      (LookupLink.Dataset.RowNum = LookupLink.Dataset.EofRowNum) then
    begin
      DataLink.AdvSearchData := '';
      // I dont know why I used to assign Text instead of blank string
      // so leave the old item commented in here for a while.  It does
      // not seem to make sense and was causing problems when lookups
      // were not being deactivated.
      DataLink.SetAdvSearchDisplay('');//?? Text);
      DataLink.AdvSearchWhere := '';
    end
    else
    begin
      DataLink.AdvSearchData :=
        LookupLink.Dataset.IB_KeyDataLink.AdvSearchData;
      DataLink.SetAdvSearchDisplay(LookupLink.FieldText);
      // if we are deactivating we need to save the where clause as well
      if Deactivating then
        DataLink.AdvSearchWhere :=
          LookupLink.Dataset.IB_KeyDataLink.AdvSearchWhere
      else
        DataLink.AdvSearchWhere := '';
    end;
  end;
end;

procedure TIB_CustomLookupEnh.RestoreSearchInfo;
begin
  // Use any saved bookmark data to restore the lookup position
  if LookupLink.Active and (LookupLink.Dataset <> nil) and
    (DataLink.State = dssSearch) then
  begin
    LookupLink.Dataset.IB_KeyDataLink.AdvSearchData :=
      DataLink.AdvSearchData;
  end;
end;

//procedure TIB_CustomLookupEnh.LookupStateChanged( Sender: TIB_DataLink; IB_DataSource: TIB_DataSource );
//begin
////
//end;

procedure TIB_CustomLookupEnh.LookupDataChange(Sender: TIB_DataLink;
  IB_DataSource: TIB_DataSource; IB_Field: TIB_Column);
begin
  // If the lookup currently belongs to us then synchronise...
  // Ignore the synchronisation if we are only a paramedit and the dataset
  // is not in a strictly prepared state.
  if FLookupPossession or (FDisplayCol = LookupLink.Field) then
  begin
    SaveSearchInfo(false);
    SysDataChange(DataLink.Field);
  end;
end;

function TIB_CustomLookupEnh.GetFilterHotKeyDefault: TShortCut;
begin
  Result := TextToShortCut('F8');
end;

function TIB_CustomLookupEnh.GetRefreshHotKeyDefault: TShortCut;
begin
  Result := TextToShortCut('F5');
end;

procedure TIB_CustomLookupEnh.EditingChanged(Sender: TIB_DataLink;
  IB_DataSource: TIB_DataSource);
// Copy of function from IB_CustomEdit.IMP
var
  NewRO: boolean;
begin
  NewRO := DataLink.ReadOnly;
  if not NewRO then
  begin
    with DataLink do
    begin
      if not ((State in [dssInsert, dssEdit, dssSearch]) or IsParamEdit) then
      begin
        NewRO := not SearchAlways;
      end;
    end;
  end;
  inherited ReadOnly := NewRO;
end;

procedure TIB_CustomLookupEnh.SetDisplayCol;
begin
  with DataLink do
  begin
    if DataLink.IsParamEdit then
      FDisplayCol := LookupLink.Field
    else
    begin
      if LookupLink.Active and
        ((FLookupPossession and DataLink.Editing) or
        ((not LookupDeactivation) and (not LookupShared))) then
      //if FLookupPossession and LookupLink.Active and DataLink.Editing then
        FDisplayCol := LookupLink.Field
      else
      begin
        if (Dataset <> nil) and (Dataset.Prepared) then
        begin
          if DisplayField = '' then
            FDisplayCol := DataLink.Field
          else
            FDisplayCol := Dataset.FieldByName(DisplayField);
        end
        else
          FDisplayCol := nil;
      end;
    end;
  end;
end;

procedure TIB_CustomLookupEnh.SysDataChange(IB_Field: TIB_Column);
  procedure ClearAndSetDisplay(AStr: string);
  begin
    if IsMasked then
      Mask.Clear;
    DisplayText := AStr;
  end;
begin
  // No point doing anything if we disappearing anyway.
  if csDestroying in ComponentState then
    Exit;
  if FInternalRowChg <> 0 then
    Exit;
  // For param edit we ignore all data changes that occur while we
  // are not in a specifically prepared/search state (otherwise the
  // lookup tends to take on the value of whatever it saw last).
  // We use SearchAlways flag to allow special uses of IsParamEdit
  // to keep active in other modes.
  if DataLink.IsParamEdit then
  begin
    if not SearchAlways and
      not (DataLink.State in [dssPrepared, dssSearch]) then
      Exit;
  end;
  // Allow a default value application to synchronise the lookup
  // (not save to do if lookup is shared).
  if (IB_Field <> nil) and DataLink.Dataset.IsInsertPending and
    not (LookupShared or IB_Field.IsNull) then
    ActivateLookup;
  SetDisplayCol; // Make sure display column is set appropriately
  if ((DataLink.State = dssSearch) or SearchAlways) and
    (FDisplayCol <> LookupLink.Field) then
  begin
    if SearchWithMask then
      Text := DataLink.AdvSearchDisplay
    else
      ClearAndSetDisplay(DataLink.AdvSearchDisplay);
  end
  else if IsMasked and (FDisplayCol <> nil) and FDisplayCol.IsNotNull then
    Text := FDisplayCol.AsString
  else if FDisplayCol <> nil then
    ClearAndSetDisplay(FDisplayCol.AsString)
  else
    ClearAndSetDisplay('');
  if Focused and AutoSelect and not TextMouseDown then
    SelectAll;
  // If the datachange occurs during user input we do not
  // mark the control with modified := false - since we assume the
  // datachange was a result of the user input.
  if not InUserInput then
    Modified := false;
  // For lookups we issue the OnChange event here after a datachange
  // (Note that the CustomComboEnh filters the TextChanged event so that
  // it must be handled by the derivative.)  The idea being that we dont
  // really want change events with every keystroke, but just when a change
  // has actually take place - such as a record more or lookup scroll.
  // This is essentially similar to the CustomCombo is list mode which
  // only issues a change when the selected item changes.
  Change;
  // tell incremental search that a datachange has been done.
  FDataChangeDone := true;
end;

procedure TIB_CustomLookupEnh.SysStateChanged;  { Handles LayoutChange event as well }
var
  tmpstr: string;
begin
  // No point doing anything if we disappearing anyway.
  if csDestroying in ComponentState then
    Exit;
  // Now on with the normal stuff
  with DataLink do
  begin
    if ColorScheme then
    begin
      inherited Color := Color;
    end;
    if not (State in [dssInsert, dssPrepared, dssEdit, dssSearch]) then
      EditEndModify;
    SetDisplayCol;
  end;
  if (FDisplayCol <> nil) then
  begin
    case Datalink.FieldCharCase of
      ccUpper: CharCase := ecUpperCase;
      ccLower: CharCase := ecLowerCase;
      ccProper: CharCase := ecProperCase;
    else
      CharCase := ecNormal;
    end;
    if Alignment <> FDisplayCol.Alignment then
    begin
      // NOTE: Changing alignment causes a recreatewnd call.  If the alignment
      // has changed in the middle of user input we avoid doing this as it
      // can cause all sorts of difficulties.
      if FUserInputFlag = 0 then
        Alignment := FDisplayCol.Alignment;
    end;
    tmpstr := FDisplayCol.EditMask;
    if Length(tmpstr) > 0 then
      Self.SetEditMaskEx(tmpstr)
    else
      Self.EditMaskMode := eemNone;
    // MaxLength is redundant with lookups, since it is always restricted
    // to existing values anyway.  Indeed setting MaxLength can cause problems
    // due to preventing insert of a new character (to search for a different
    // value) and so stopping a search from starting.
    MaxLength := 0;
  end;
  EditingChanged(DataLink, DataSource);
  UpdateTabStop;
  InvalidateButton;
  InvalidateGlyph;
  SyncDropBoxColors;
end;


function TIB_CustomLookupEnh.GetLookupKeyText: string;
var
  i: integer;
begin
  Result := '';
  if LookupLink.Active and
    (LookupLink.Dataset.Fields.RowState <> rsNone) then
    with LookupLink.Dataset do
    begin
      i := KeyLinks.Count - 1;
      if i >= 0 then
        Result := FieldByName(KeyLinks.IndexNames[i]).AsString;
    end;
end;

procedure TIB_CustomLookupEnh.SysUpdateData;
begin
  if Assigned(Field) then
  begin
    if Modified then
    begin
      // ValidateEdit;  Already done so avoid repeating and possible recursion
      // But we still need to force the modify to end so that search criteria
      // is saved (including where clause if lookup is deactivated.
      EditEndModify;
      // We are not editing the field directly unless in ParamEdit
      // Its should come from Lookup synchronisation!
      if DataLink.IsParamEdit then
        DataLink.FieldText := GetLookupKeyText;
    end;
  end;
  Modified := false;
end;

function TIB_CustomLookupEnh.IsMaskStored: boolean;
begin
  // IBO does not store masks with controls, only with the data definition
  Result := false;
end;

function TIB_CustomLookupEnh.GetDBDisplayField: TObject;
var
  Col: TIB_Column;
begin
  // The returned display field must ALWAYS be the source/main dataset display
  // field, which is not necessarily the field we are displaying in the control
  // (which may come from the lookup dataset).  This is because we want the
  // column returned to be valid for use in grid displays etc.
  with DataLink do
  begin
    if (Dataset <> nil) and (Dataset.Prepared) then
    begin
      if DisplayField = '' then
        Result := DataLink.Field
      else
      begin
        Col := nil;
        if (Dataset.Fields.GetByName(DisplayField, Col)) then
          Result := Col
        else
          Result := nil;
      end;
    end
    else
      Result := nil;
  end;
end;

function TIB_CustomLookupEnh.IBG_GetAutoLabel: TObject;
begin
  Result := FAutoLabel;
end;

procedure TIB_CustomLookupEnh.DoAfterValidateEdit(var ErrCode: integer);
begin
  inherited DoAfterValidateEdit(ErrCode);
  if (ErrCode = E_EDITENH_NOERROR) and DataLink.IsParamEdit and
    DataLink.Prepared then
    DataLink.FieldText := GetLookupKeyText;
end;

// Retrieve the display column value appropriate to displaying the *current*
// row of the dataset - because in edit/insert mode the DBDisplayField may
// not have been updated to match the lookup field.
// ??? Commented out because KeyDescLinks should now ensure full sync
// ??? function TIB_CustomLookupEnh.GetCurrentDisplayField: TIB_Column;
// ??? begin
// ??? if LookupLink.Active then
// ???     Result := FDisplayCol
// ??? else begin
// ???     with DataLink do begin
// ???         if (Dataset <> nil) and (Dataset.Prepared) and (DisplayField <> '') then
// ???             Result := Dataset.FieldByName( DisplayField )
// ???         else Result := nil;
// ???         end;
// ???     end;
// ??? end;

procedure TIB_CustomLookupEnh.SetWorking(DspStr: string);
begin
  if (ComboControl <> nil) and ComboControl.Active then
    TIB_ComboLookupListEnh(ComboControl).SetWorking(DspStr);
end;

procedure TIB_CustomLookupEnh.ClrWorking;
begin
  if (ComboControl <> nil) then
    TIB_ComboLookupListEnh(ComboControl).ClrWorking;
end;

procedure TIB_CustomLookupEnh.WorkingCallback(Sender: TIB_Dataset;
  Status: TIB_CallbackStatus; CursorRowNum: integer; var Abort: boolean);
var
  StatusStr: string;
begin
  case Status of
    csInit: FCallbackAbort := false;
    csFinal: ClrWorking;
    csRefresh:
      begin
        if FCallbackAbort then
          Abort := true
        else
        begin
          // need to localise
          StatusStr := ENHSTR_Lookup_Searching + ' (' + IntToStr(CursorRowNum) + ')';
          SetWorking(StatusStr);
        end;
      end;
  end;
end;

function TIB_CustomLookupEnh.EditCanModify: boolean;
begin
  if DataLink.IsParamEdit then
    Result := true
  else
    Result := DataLink.CanModify and
      (DataLink.DataSource.Modifying or DataLink.DataSource.AutoEdit or
      (DataLink.DataSource.AutoInsert and
      (DataLink.Dataset.Fields.RowState = rsNone)));
end;

function TIB_CustomLookupEnh.EditBeginModify: boolean;
begin
  if DataLink.IsParamEdit then
    Result := true
  else
    Result := DataLink.Modify;
  if Result then
  begin
    ActivateLookup;
    Result := LookupLink.Active;
    if Result then
    begin
      if AutoDropDown then
        ShowDropDown;
    end;
  end;
end;

function TIB_CustomLookupEnh.EditEndModify: boolean;
begin
  DeactivateLookup;
  if (DataLink.DataSource <> nil) then
  begin
    Result := not DataLink.DataSource.Modifying;
  end
  else
    Result := true;
end;

procedure TIB_CustomLookupEnh.Reset;
begin
  if Focused and Modified then
  begin
    if Text <> FSavedLookupText then
      Text := FSavedLookupText;
    Modified := false;
  end;
  SelectAll;
end;

procedure TIB_CustomLookupEnh.Change;
begin
  if not MaskChanging then
    with DataLink do
    begin
      // Only mark modified if processing user input, otherwise we assume
      // that the change came as the result of underlying buffer changes
      // (which do not indicate that the control value is modified).
      if InUserInput and
        (DataChangingLevel = 0) and
        (NeedToPost or (State = dssSearch) or SearchAlways or IsParamEdit) then
      begin
        ControlIsModified := true;
        Modified := true;
      end;
    end;
  inherited Change;
end;

//procedure TIB_CustomLookupEnh.CMGotFocus(var Message: TCMGotFocus);
//begin
//inherited;
//SysDataChange( nil );
//end;

procedure TIB_CustomLookupEnh.CMEnter(var Message: TCMEnter);
begin
  //SetFocused( true );
  inherited;
  //ActivateLookup;
  SysDataChange(nil);
  FSavedLookupText := Text;
  // We dont try and activate the lookup on enter, that will
  // happen as needed by a call to EditBeginModify when key or
  // mouse events require it.
end;

procedure TIB_CustomLookupEnh.CMExit(var Message: TCMExit);
begin
  inherited;   // base class does a ValidateEdit
  if not DataLink.IsParamEdit then
    DataLink.UpdateData;
  SysDataChange(nil);
  FSavedLookupText := Text;
end;

procedure TIB_CustomLookupEnh.WMKillFocus(var Message: TMessage);
begin
  inherited;
  DeactivateLookup;
end;

procedure TIB_CustomLookupEnh.KeyDown(var Key: Word; Shift: TShiftState);
begin
  Inc( FUserInputFlag );
  try
    if ShortCut(Key, Shift) = FRefreshHotKey then
    begin
      if not (LookupShared or LookupLink.Active) and
        (DataLink.State in [dssEdit, dssInsert, dssSearch]) then
        ActivateLookup;
      if LookupLink.Active then
      begin
        LookupLink.Dataset.Refresh;
        Key := 0;
        FSkipChar := true;
      end;
    end
    else
      if (not ReadOnly) and EditCanModify and (LookupLink.Dataset <> nil) then
      begin
        FSkipChar := false;
        if (ShortCut(Key, Shift) = FFilterHotKey) then
        begin
          if (ComboControl <> nil) and ComboControl.Active then
          begin
            // do toggle on key up
            //TIB_ComboLookupListEnh(ComboControl).ToggleFilterCB;
            Key := 0;
            FSkipChar := true;
          end;
        end;
        if Shift = [] then
        begin
          case Key of
            VK_DOWN:
              begin
                if EditBeginModify then
                  with LookupLink.Dataset do
                  begin
                    if RowNum = EofRowNum - 1 then
                      ValidateRows(EofRowNum, EofRowNum);
                    if RowNum < EofRowNum - 1 then
                      Next;
                  end;
                Key := 0;
              end;
            VK_UP:
              begin
                if EditBeginModify then
                  with LookupLink.Dataset do
                  begin
                    if RowNum = BofRowNum + 1 then
                      ValidateRows(BofRowNum, BofRowNum);
                    if RowNum > BofRowNum + 1 then
                      Prior;
                  end;
                Key := 0;
              end;
            VK_NEXT:
              begin
                if EditBeginModify then
                  with LookupLink.Dataset do
                  begin
                    DisableControls;
                    try
                      MoveBy( DropDownRows );
                      if Eof and ( BufferRowCount > 0 ) then
                        Prior;
                    finally
                      EnableControls;
                    end;
                  end;
                Key := 0;
              end;
            VK_PRIOR:
              begin
                if EditBeginModify then
                  with LookupLink.Dataset do
                  begin
                    DisableControls;
                    try
                      MoveBy( - DropDownRows );
                      if Bof and ( BufferRowCount > 0 ) then
                        Next;
                    finally
                      EnableControls;
                    end;
                  end;
                Key := 0;
              end;
          end;
        end;
        if Shift = [ssCtrl] then
        begin
          case Key of
            VK_HOME:
              begin
                if EditBeginModify then
                  LookupLink.Dataset.First;
                Key := 0;
              end;
            VK_END:
              begin
                if EditBeginModify then
                  LookupLink.Dataset.Last;
                Key := 0;
              end;
          end;
        end;
      end;
    if (DataLink.DataSource <> nil) and
      (DataLink.DataSource.Modifying or DataLink.IsParamEdit) then
    begin
      if Key = VK_RETURN then
      begin
        SelectAll;
        DoTextChange;
        if (ComboControl <> nil) then
          ComboControl.CloseUp;
        Key := 0;
      end;
    end;
    inherited;
  finally
    Dec(FUserInputFlag);
  end;
end;

procedure TIB_CustomLookupEnh.KeyPress(var Key: char);
begin
  Inc(FUserInputFlag);
  try
    if FSkipChar then // stop beeps being generated with the hotkey
      Key := char(0);
    inherited;
  finally
    Dec(FUserInputFlag);
  end;
end;

procedure TIB_CustomLookupEnh.KeyUp(var Key: Word; Shift: TShiftState);
begin
  Inc(FUserInputFlag);
  try
    if FSkipChar then
    begin
      if (ShortCut(Key, Shift) = FFilterHotKey) then
      begin
        if (ComboControl <> nil) and ComboControl.Active then
        begin
          TIB_ComboLookupListEnh(ComboControl).ToggleFilterCB;
          Key := 0;
          FSkipChar := false;
        end;
      end;
    end;
    inherited;
  finally
    Dec(FUserInputFlag);
  end;
end;

function TIB_CustomLookupEnh.IncrementalSearch(TxtVal: string; ChkLen: integer):
integer;
var
  SrchVal, tmpStr: string;
  tmpPos: integer;
begin
  Result := -1;
  FDataChangeDone := false; // init flag read to test if datachange is done
  if (LookupLink.Dataset <> nil) and (LookupLink.Dataset is TIB_BDataset) and
    (DataLink.DataSource <> nil) and (LookupLink.Active) and
    (DataLink.DataSource.Modifying or DataLink.IsParamEdit) and
    (not DataLink.DataChanging) and (not LookupLink.DataChanging) then
    with LookupLink.Dataset as TIB_BDataset do
    begin
      BeginCallbackFreeze;
      Inc(FUserInputFlag);
      try
      if TxtVal = '' then
      begin
        // Checking OrderingRefinePos here is no good, it stops
        // the user from clearing the selection.
        // if OrderingRefinePos = 0 then
        Inc(FInternalRowChg);
        try
          if (RowNum <> BofRowNum) and (RowNum <> EofRowNum) then
          begin
            //RowNum := BofRowNum;  Not reliable when using refined zones
            GotoNull;
            Change;
          end;
        finally
          Dec(FInternalRowChg);
        end;
        // otherwise just leave all alone till more is typed
      end
      else
      begin
        if (OrderingRefinePos = 0) or (ChkLen >= OrderingRefinePos) or
          (ChkLen = -1) then
        begin
          if ChkLen > 0 then
            SrchVal := Copy(TxtVal, 1, ChkLen)
          else
            SrchVal := TxtVal;
          // temporary setting for BufferRowNum to see if this
          // resolves the problems of unwanted BufferEof = true
          //BufferRowNum := 0; // if this does not work try BufferRowNum := RowNum
          BufferRowNum := RowNum;
          // If a search delimiter has been specified and exists in
          // the given incremental search string the begin a "brute"
          // search for the required value...
          tmpPos := -1;
          if SearchDelimiter <> #0 then
          begin
            tmpPos := Pos(SearchDelimiter, SrchVal);
            if tmpPos > 0 then
            begin
              // first do an incremental upto the delimiter...
              tmpStr := Copy(SrchVal, 1, tmpPos - 1);
              IncSearchString(tmpStr, 1, true); //SeekNearest );
              // then brute force from there
              BruteSearch(SrchVal, true, false); //SeekNearest );
            end;
          end;
          // If not delimiter was specified/found above just do inc search
          if tmpPos < 1 then
          begin
            // first try to satisfy from the buffer only
            if not BruteSearch(SrchVal, true, true) then
              // and if that didnt work do the dataset search
              IncSearchString(SrchVal, 1, true); //SeekNearest );
          end;
          if RowNum > BofRowNum then
          begin
            Result := ChkLen;
          end;
          // Events above can result in breaking synchronisation between
          // the DataLink.ControlIsModified and the control's Modified flag,
          // so re-establish synchronisation...
          if Self.Modified <> Self.DataLink.ControlIsModified then
            Self.DataLink.ControlIsModified := Self.Modified;
          // If no change issued by actual dataset search then force our own
          // change event to be sure appropriate update and change processing
          // is done anyway.
          if (not FDataChangeDone) then
            Self.SysDataChange(nil);
        end
        else
          Result := ChkLen;
      end;
      finally
        Dec(FUserInputFlag);
        EndCallbackFreeze;
      end;
    end;
end;

function TIB_CustomLookupEnh.BruteSearch(SearchVal: string;
  MatchNear, BufOnly: boolean): boolean;
// Expects the dataset to have been validated already in the IncrementalSearch
// procedure that calls this.  Current this is ALWAYS case insensitive!
var
  tmpCmp: integer;
  tmpVal: string;
  tmpCol: TIB_Column;

  function DoCompareText( S1, S2: string ): integer;
  var
    ii, iilim: integer;
  begin
    if Assigned( LookupLink.Dataset.IB_Connection.OnCustomCompareText ) then
      Result :=
        LookupLink.Dataset.IB_Connection.OnCustomCompareText( tmpCol, S1, S2 )
    else
    begin
      // AnsiCompare text does not do character by character comparison, and
      // that is what we require for incremental search comparisons.
      //Result := AnsiCompareText(S1,S2);
      iilim := Length(S1);
      if Length(S2) < iilim then
        iilim := Length(S2);
      ii := 1;
      Result := 0;
      while ii <= iilim do
      begin
        if AnsiUpperCase(S1[ii]) < AnsiUpperCase(S2[ii]) then
        begin
          Result := -1;
          break;
        end else if AnsiUpperCase(S1[ii]) > AnsiUpperCase(S2[ii]) then
        begin
          Result := 1;
          break;
        end;
        Inc(ii);
      end;
      if (Result = 0) and (ii > iilim) then
      begin
        if Length(S1) < Length(S2) then
          Result := -1
        else if Length(S1) > Length(S2) then
          Result := 1;
      end;
    end;
  end;

begin
  Result := false;
  SearchVal := SearchVal; // was ansiupper, no need now using comparetext
  with LookupLink.Dataset as TIB_BDataset do
  begin
    if RowNum <= BofRowNum then
      BufferRowNum := BofRowNum + 1
    else
      BufferRowNum := RowNum;
    tmpCol := nil;
    if not BufferEof then
      tmpCol := BufferFieldByName(LookupLink.Field.FieldName);
    while not BufferEof do
    begin
      tmpVal := tmpCol.DisplayText; // was ansiupper, no need now using compare
      tmpCmp := DoCompareText( SearchVal, tmpVal );
      if tmpCmp > 0 then //SearchVal > tmpVal then
      begin
        if BufOnly and (BufferRowNum >= BufferRowCount) then
          break;
        BufferRowNum := BufferRowNum + 1;
      end
      else
      begin
        if tmpCmp = 0 then //SearchVal = tmpVal then
        begin
          RowNum := BufferRowNum;
          Result := true;
        end
        else
        begin
          tmpVal := Copy(tmpVal, 1, Length(SearchVal));
          if MatchNear and (DoCompareText(SearchVal, tmpVal) = 0) then
            //(SearchVal = tmpVal) then
          begin
            RowNum := BufferRowNum;
            Result := true;
          end;
        end;
        break;
      end;
    end;
  end;
end;

function TIB_CustomLookupEnh.CreateComboControl: TCustomComboControlEnh;
begin
  if IB_ComboLookupEnh_ListBox = nil then
  begin
    IB_ComboLookupEnh_ListBox := TIB_ComboLookupListEnh.Create(nil);
  end;
  Inc(IB_ComboLookupEnh_ListBoxRefCount);
  Result := IB_ComboLookupEnh_ListBox;
end;

procedure TIB_CustomLookupEnh.DestroyComboControl(CurrentControl:
  TCustomComboControlEnh);
begin
  if IB_ComboLookupEnh_ListBox.Parent = Self then
  begin
    IB_ComboLookupEnh_ListBox.FList.DataSource := nil;
    IB_ComboLookupEnh_ListBox.Parent := nil;
  end;
  if CurrentControl = IB_ComboLookupEnh_ListBox then
  begin
    Dec(IB_ComboLookupEnh_ListBoxRefCount);
    if IB_ComboLookupEnh_ListBoxRefCount = 0 then
    begin
      IB_ComboLookupEnh_ListBox.Free;
      IB_ComboLookupEnh_ListBox := nil;
    end;
  end;
end;


initialization;
  IB_ComboLookupEnh_ListBox := nil;
  IB_ComboLookupEnh_ListBoxRefCount := 0;

finalization;
  IB_ComboLookupEnh_ListBox := nil;
  IB_ComboLookupEnh_ListBoxRefCount := 0;

end.
