{ Telesis_IB_MemoEnh.pas                          SearchTag: ~~EnhComponents~~

 Copyright 1999-2001 - Telesis Computing Pty Ltd.  Australia.
Contact: Geoff Worboys - geoff@telesiscomputing.com.au }

{: @summary IBO Aware variation of TCustomMemoEnh.
<br>
@desc ...
<br>
<br>
<strong>History</strong><br>
<br>
22-Jan-2008 - Publish some events from EditEnh.  [internal]
Some clean-up of uses.<br>
<br>
19-Jun-2005 - Trying to synchronise ValidateEdit and UpdateData calls so that
both get done but we minimise duplication and avoid recursion.<br>
<br>
29-Apr-2002 - Addition of DoAfterValidateEdit override and rearranged
UpdateData processing. (Improvements to Parameter edit capability.)<br>
<br>
17-Dec-2001 - Published some properties and events from base class.<br>
<br>
23-Aug-2001 - Additional csDestroying checks.<br>
<br>
02-Aug-2001 - Documentation.<br>
<br>
<br>
<strong>Author: </strong>Geoff Worboys - Telesis Computing Pty Ltd.<br>
<br>}
unit Telesis_IB_MemoEnh;

{$INCLUDE IB_Directives.inc}
{$INCLUDE Telesis.inc}

interface

uses
  Classes,
  Messages,
  Controls,
  IB_FieldDataLink,
  IBC_Label,
  IB_Components,
  Telesis_MemoEnh;

type

  {$DEFINE IBC_MemoEnh}
  {$DEFINE IB_UserClass}
  {$DEFINE IB_FieldControl}

  TIB_CustomMemoEnh = class(TCustomMemoEnh)
    {$DEFINE IB_AutoLabel}
    {$I IBC_CustomMemo.INT}
    {$UNDEF IB_AutoLabel}
  private
  protected
    function IsMaskStored: boolean; override;
    function GetDBDisplayField: TObject; override;
    function IBG_GetAutoLabel: TObject; override;
    procedure DoAfterValidateEdit(var ErrCode: integer); override;

  public
    function EditCanModify: boolean; override;
    function EditBeginModify: boolean; override;

  end;                                  // TIB_CustomMemoEnh


  TIB_MemoEnh = class(TIB_CustomMemoEnh)
    {$DEFINE IB_AutoLabel}
    {$I IB_WinControl.PBL}
    {$UNDEF IB_AutoLabel}
    // Cant use CustomMemo.PBL because it is expecting some of the things only
    // defined to TIB_CustomEdit
  public

  published
    // Stuff we want from CustomEdit.PBL
    property AutoSelect;
    property AutoSize;
    property HideSelection;
    property OEMConvert;
    property OnChange;
    // Stuff we want from CustomMemo.PBL
    property AutoDisplay;
    property MaxLength;
    property ScrollBars;
    property WantReturns;
    property WantTabs;
    property WordWrap;

    // TCustomMemoEnh publications...
    //property Align;
    property Alignment;
    property Button;
    property CharCase;
    property Ctl3DShallow;
    property ScrollBarsAuto;
    property GlyphAlign;
    property GlyphSize;
    property Options;

    // TCustomEditEnh events
    property OnButton1;
    property OnButton2;
    property OnButton3;
    property OnDrawButton;
    property AfterValidateEdit;
    property OnValidateEdit;
    property OnValidateError;
    property OnGlyphClick;
    property OnDrawGlyph;
  end;                                  // TIB_MemoEnh


implementation

uses Windows,
  Forms,
  IB_Constants,
  Telesis_Constants,
  Telesis_EditEnh;


// I was using IB_CustomMemo.IMP but it filters change events which means
// that I cannot find out when I should update the scroll bars.  So I have
// copied IB_CustomMemo.IMP into here and simply update the Change function.
{ $ I IBC_CustomMemo.IMP}

{$DEFINE IB_AutoLabel}
{$I IBC_CustomEdit.IMP}
{$UNDEF IB_AutoLabel}

constructor TIB_CustomMemoEnh.Create(AOwner: TComponent);
begin
  FIB_FieldDataLink := TIB_FieldDataLink.Create(Self);
  FAutoLabel := TIB_AutoLabel.Create(Self);
  inherited Create(AOwner);
  inherited ReadOnly := true;
  ControlStyle := ControlStyle + [csReplicatable];
  FAutoDisplay := true;
  FAllowCRLF := true;
  DataLink.Control := Self;
  DataLink.OnEditingChanged := EditingChanged;
  DataLink.OnStateChanged := StateChanged;
  DataLink.OnDataChange := DataChange;
  DataLink.OnUpdateData := UpdateData;
end;

destructor TIB_CustomMemoEnh.Destroy;
begin
  FAutoLabel.Free;
  FAutoLabel := nil;
  DataSource := nil;
  DataLink.Control := nil;
  DataLink.OnEditingChanged := nil;
  DataLink.OnStateChanged := nil;
  DataLink.OnDataChange := nil;
  DataLink.OnUpdateData := nil;
  inherited Destroy;
end;

procedure TIB_CustomMemoEnh.Loaded;
begin
  inherited Loaded;
  SysStateChanged;
  DataLink.DataChange;
  RemakeAutoLabel;
end;

procedure TIB_CustomMemoEnh.SetAutoDisplay(AValue: Boolean);
begin
  if FAutoDisplay <> AValue then
  begin
    FAutoDisplay := AValue;
    if AValue then
    begin
      LoadMemo;
    end;
  end;
end;

procedure TIB_CustomMemoEnh.LoadMemo;
begin
  if Assigned(Field) and not MemoLoaded then
  begin
    try
      Inc(FIgnoreChangeLevel);
      try
        Field.AssignTo(Lines);
        FMemoLoaded := True;
      except
        Lines.Text := E_Could_Not_Load_Memo;
        Application.HandleException(Self);
      end;
    finally
      Dec(FIgnoreChangeLevel);
    end;
    SysStateChanged;
    Modified := false;
  end;
end;

procedure TIB_CustomMemoEnh.SysDataChange(IB_Field: TIB_Column);
begin
  // No point doing anything if we disappearing anyway.
  if csDestroying in ComponentState then
    Exit;
  Inc(FIgnoreChangeLevel);
  try
    with DataLink do
    begin
      MaxLength := FieldMaxLength;
      FMemoLoaded := false;
      if Assigned(Field) then
      begin
        case State of
          dssBrowse,
            dssDelete,
            dssEdit,
            dssInsert:
            begin
              if Field.IsBlob then
              begin
                if AutoDisplay or Field.IsLoaded then
                begin
                  LoadMemo;
                end
                else
                begin
                  Text := FieldText;
                end;
                FMemoLoaded := Field.IsLoaded;
              end
              else
              begin
                if Lines.Text <> FieldText then
                begin
                  Lines.Text := FieldText;
                end;
                FMemoLoaded := true;
              end;
            end
        else
          begin
            if Lines.Text <> FieldText then
            begin
              Lines.Text := FieldText;
            end;
            FMemoLoaded := true;
          end;
        end;
      end
      else
        if Lines.Text <> FieldText then
        begin
          Lines.Text := FieldText;
        end;
    end;
    Modified := false;
  finally
    Dec(FIgnoreChangeLevel);
  end;
end;

procedure TIB_CustomMemoEnh.SysUpdateData;
begin
  if DataLink.ControlIsModified and Assigned(Field) then
  begin
    if not Validating then
      ValidateEdit;
    case DataLink.State of
      dssEdit,
        dssInsert:
        begin
          if Field.IsBlob then
          begin
            Field.Assign(Lines);
          end
          else
          begin
            DataLink.FieldText := Lines.Text;
          end;
        end
    else
      begin
        DataLink.FieldText := Lines.Text;
      end;
    end;
  end;
end;

procedure TIB_CustomMemoEnh.Change;
begin
  if FIgnoreChangeLevel = 0 then
  begin
    with DataLink do
    begin
      if MemoLoaded then
      begin
        Modify;
      end;
      FMemoLoaded := true;
      if (State in [dssInsert, dssEdit, dssSearch]) or IsParamEdit then
      begin
        ControlIsModified := true;
      end;
    end;
  end;
  if MemoLoaded then
  begin
    inherited Change;
  end
  else
  begin
    if ScrollBarsAuto then
      CheckAutoScrollBars;
  end;
end;

procedure TIB_CustomMemoEnh.KeyDown(var Key: Word; Shift: TShiftState);
var
  Msg: TMsg;
begin
  if (not AllowCRLF) and (Key = VK_RETURN) then
  begin
    PeekMessage(Msg, Handle, WM_CHAR, WM_CHAR, PM_REMOVE);
    inherited KeyDown(Key, Shift);
    Key := 0;
  end
  else
  begin
    inherited KeyDown(Key, Shift);
    if MemoLoaded then
    begin
      if (Key in [8..10, 13, 23, 24]) or
        (Key = VK_DELETE) or
        ((Key = VK_INSERT) and (ssShift in Shift)) then
      begin
        DataLink.Modify;
      end;
    end
    else
      if (DataLink.State in [dssInsert, dssEdit, dssSearch]) then
      begin
        LoadMemo;
      end
      else
      begin
        Key := 0;
      end;
  end;
end;

procedure TIB_CustomMemoEnh.KeyPress(var Key: Char);
begin
  inherited KeyPress(Key);
  if MemoLoaded then
  begin
    if Assigned(Field) and (Key in [#32..#255]) and
      not DataLink.FieldIsValidChar(Key) then
    begin
      MessageBeep(0);
      Key := #0;
    end;
    case Key of
      #32..#255: DataLink.Modify;
      #27: SysDataChange(nil);
    end;
  end
  else
  begin
    if Key = #13 then
    begin
      LoadMemo;
    end;
    Key := #0;
  end;
end;

procedure TIB_CustomMemoEnh.WMLButtonDblClk(var Message: TWMLButtonDblClk);
begin
  if MemoLoaded then
  begin
    inherited;
  end
  else
  begin
    LoadMemo;
  end;
end;

procedure TIB_CustomMemoEnh.WMCut(var Message: TMessage);
begin
  if MemoLoaded and DataLink.Modify then
  begin
    inherited;
  end;
end;

procedure TIB_CustomMemoEnh.WMPaste(var Message: TMessage);
begin
  if DataLink.Modify then
  begin
    if not MemoLoaded then
    begin
      Lines.Clear;
    end;
    inherited;
  end;
end;

procedure TIB_CustomMemoEnh.SysStateChanged;
begin
  // No point doing anything if we disappearing anyway.
  if csDestroying in ComponentState then
    Exit;
  with DataLink do
  begin
    if ColorScheme then
    begin
      inherited Color := Color;
    end;
    case FieldCharCase of
      ccUpper: CharCase := ecUpperCase;
      ccLower: CharCase := ecLowerCase;
      ccProper: CharCase := ecProperCase;
    else
      CharCase := ecNormal;
    end;
    if Alignment <> FieldAlignment then
    begin
      Alignment := FieldAlignment;
    end;
  end;
  UpdateTabStop;
end;

procedure TIB_CustomMemoEnh.CMEnter(var Message: TCMEnter);
begin
  {$IFDEF IBC_Edit}
  SetFocused(True);
  {$ENDIF}
  inherited;
end;

procedure TIB_CustomMemoEnh.CMExit(var Message: TCMExit);
begin
  inherited;   // base class does a ValidateEdit
end;

function TIB_CustomMemoEnh.IsMaskStored: boolean;
begin
  // IBO does not store masks with controls, only with the data definition
  Result := false;
end;

function TIB_CustomMemoEnh.GetDBDisplayField: TObject;
begin
  Result := DataLink.Field;
end;

function TIB_CustomMemoEnh.IBG_GetAutoLabel: TObject;
begin
  Result := FAutoLabel;
end;

procedure TIB_CustomMemoEnh.DoAfterValidateEdit(var ErrCode: integer);
begin
  inherited DoAfterValidateEdit(ErrCode);
  if (ErrCode = E_EDITENH_NOERROR) and not DataLink.DataUpdating then
  begin
    DataLink.UpdateData;
    SysDataChange(nil);
  end;
end;

function TIB_CustomMemoEnh.EditCanModify: boolean;
begin
  if DataLink.IsParamEdit then
    Result := true
  else
    Result := DataLink.CanModify and
      (DataLink.DataSource.Modifying or DataLink.DataSource.AutoEdit or
      (DataLink.DataSource.AutoInsert and
      (DataLink.Dataset.Fields.RowState = rsNone)));
end;

function TIB_CustomMemoEnh.EditBeginModify: boolean;
begin
  if DataLink.IsParamEdit then
    Result := true
  else
    Result := DataLink.Modify;
end;

end.