{ Telesis_IB_SpinEditEnh.pas                      SearchTag: ~~EnhComponents~~

 Copyright 1999-2001 - Telesis Computing Pty Ltd.  Australia.
Contact: Geoff Worboys - geoff@telesiscomputing.com.au }

{: @summary IBO Aware variation of TCustomSpinEditEnh.
<br>
@desc ...
<br>
<br>
<strong>History</strong><br>
<br>
12-Aug-2005 - Rewrite SysDataChange, the old was too messy and unreliable.
Also changes for better SelectAll consistency.<br>
<br>
19-Jun-2005 - Trying to synchronise ValidateEdit and UpdateData calls so that
both get done but we minimise duplication and avoid recursion.<br>
<br>
29-May-2002 - Published some properties and events from base class.<br>
<br>
29-Apr-2002 - Addition of DoAfterValidateEdit override and rearranged
UpdateData processing. (Improvements to Parameter edit capability.)<br>
<br>
17-Dec-2001 - Published some properties and events from base class.  Also
removed any attempt to support Delphi2.<br>
<br>
23-Aug-2001 - Additional csDestroying checks.<br>
<br>
02-Aug-2001 - Documentation.<br>
<br>
<br>
<strong>Author: </strong>Geoff Worboys - Telesis Computing Pty Ltd.<br>
<br>}
unit Telesis_IB_SpinEditEnh;

{$INCLUDE IB_Directives.inc}

interface

uses SysUtils,
  Messages,
  Windows,
  Classes,
  Graphics,
  Menus,
  Controls,
  Forms,
  StdCtrls,
  Mask,
  ExtCtrls,
  IB_Components,
  IB_FieldDataLink,
  Telesis_EditEnh,
  Telesis_SpinEditEnh,
  IBC_Label;


type
  {$DEFINE IBC_SpinEditEnh}
  {$DEFINE IB_UserClass}
  {$DEFINE IB_FieldControl}

{: @summary The IBO-aware variation of TSpinEditEnh.}
  TIB_CustomSpinEditEnh = class(TCustomSpinEditEnh)
    {$DEFINE IB_AutoLabel}
    {$I IBC_CustomEdit.INT}
    {$UNDEF IB_AutoLabel}
  private
    FBorderStyle: TBorderStyle;
    procedure SetBorderStyle(AValue: TBorderStyle);
  protected
    property BorderStyle: TBorderStyle read FBorderStyle
    write SetBorderStyle
      default bsSingle;
  protected
    function IsMaskStored: boolean; override;
    function GetDBDisplayField: TObject; override;
    function IBG_GetAutoLabel: TObject; override;
    procedure DoAfterValidateEdit(var ErrCode: integer); override;
    procedure Loaded; override;
    procedure Change; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Reset; override;
    function EditCanModify: boolean; override;
    function EditBeginModify: boolean; override;
  end;                                  // TIB_CustomSpinEditEnh

{: @summary The published variation of TIB_CustomSpinEditEnh.}
  TIB_SpinEditEnh = class(TIB_CustomSpinEditEnh)
    {$DEFINE IB_CustomEdit}
    {$DEFINE IB_AutoLabel}
    {$I IB_WinControl.PBL}
    {$UNDEF IB_AutoLabel}
    {$UNDEF IB_CustomEdit}
  published

    property Increment;
    property Large_Increment;
    property Maximum;
    property Minimum;
    property WrapValue;

    property AutoSelect;
    property AutoSize;
    property HideSelection;
    property OEMConvert;
    property OnChange;

    // TCustomEditEnh publications...
    //property Alignment;      // only set via dataset/connection
    property Button;
    //property CharCase;       // only set via dataset/connection
    property Ctl3DShallow;
    //property EditMask;       // only set via dataset/connection
    //property EditMaskMode;   // only set via dataset/connection
    property GlyphAlign;
    property GlyphSize;
    property Options;
    //property Text;           // only set via dataset/connection

    property OnButton1;
    property OnButton2;
    property OnButton3;
    property OnDrawButton;
    property AfterValidateEdit;
    property OnValidateEdit;
    property OnValidateError;
    property OnGlyphClick;
    property OnDrawGlyph;
  end;


implementation

uses Telesis_MaskEnh, Telesis_Constants;

{$DEFINE IB_AutoLabel}
{$I IB_WinControl.IMP}
{$UNDEF IB_AutoLabel}

constructor TIB_CustomSpinEditEnh.Create(AOwner: TComponent);
begin
  FIB_FieldDataLink := TIB_FieldDataLink.Create(Self);
  FAutoLabel := TIB_AutoLabel.Create(Self);
  inherited Create(AOwner);
  DataLink.Control := Self;
  DataLink.OnStateChanged := StateChanged;
  DataLink.OnEditingChanged := EditingChanged;
  DataLink.OnDataChange := DataChange;
  DataLink.OnUpdateData := UpdateData;
  // Somebody plays with BorderStyle in the generic units, default it
  // back to where it belongs...
  BorderStyle := bsSingle;
end;

destructor TIB_CustomSpinEditEnh.Destroy;
begin
  FAutoLabel.Free;
  FAutoLabel := nil;
  DataSource := nil;
  DataLink.Control := nil;
  DataLink.OnStateChanged := nil;
  DataLink.OnEditingChanged := nil;
  DataLink.OnDataChange := nil;
  DataLink.OnUpdateData := nil;
  inherited Destroy;
end;

procedure TIB_CustomSpinEditEnh.Loaded;
begin
  inherited Loaded;
  SysStateChanged;
  SysDataChange(nil);
  RemakeAutoLabel;
end;

// Copy of function from IB_CustomEdit.IMP

procedure TIB_CustomSpinEditEnh.EditingChanged(Sender: TIB_DataLink;
  IB_DataSource: TIB_DataSource);
var
  NewRO: boolean;
begin
  //!!! Check for a lost connection.
  NewRO := DataLink.ReadOnly;
  if not NewRO then
  begin
    with DataLink do
    begin
      if not ((State in [dssInsert, dssEdit, dssSearch]) or IsParamEdit) then
      begin
        NewRO := not SearchAlways;
      end;
    end;
  end;
  inherited ReadOnly := NewRO;
end;

procedure TIB_CustomSpinEditEnh.SysDataChange;
  procedure ClearAndSetDisplay(AStr: string);
  begin
    if IsMasked then
      Mask.Clear;
    DisplayText := AStr;
  end;
begin
  // No point doing anything if we disappearing anyway.
  if csDestroying in ComponentState then
    Exit;
  with DataLink do
  begin
    if not Assigned(Field) or not Prepared then
      ClearAndSetDisplay(InvalidText)
    else
    begin
      if (not IsParamEdit) and ((State = dssSearch) or SearchAlways) then
      begin
        if SearchWithMask then
          Self.Text := FieldText
        else
          ClearAndSetDisplay(FieldText);
      end
      else if Self.IsMasked and Field.IsNotNull then
      begin
        if (Field is TIB_ColumnInteger) or
          (Field is TIB_ColumnSmallint) then
          Self.Value := Field.AsInteger
        else
          Self.Text := Field.AsString;
      end
      else
        ClearAndSetDisplay(FieldText);
    end;
  end;
  if Focused and AutoSelect and (not TextMouseDown) and HandleAllocated then
    SelectAll;
  Modified := false;
end;

procedure TIB_CustomSpinEditEnh.SysStateChanged;
var
  tmpstr: string;
begin
  // No point doing anything if we disappearing anyway.
  if csDestroying in ComponentState then
    Exit;
  with DataLink do
  begin
    if ColorScheme then
    begin
      inherited Color := Color;
    end;
    case FieldCharCase of
      ccUpper: CharCase := ecUpperCase;
      ccLower: CharCase := ecLowerCase;
      ccProper: CharCase := ecProperCase;
    else
      CharCase := ecNormal;
    end;
    if Alignment <> FieldAlignment then
    begin
      Alignment := FieldAlignment;
    end;
    tmpstr := FieldEditMask;
    if Length(tmpstr) > 0 then
      Self.SetEditMaskEx(tmpstr)
      // With spin edit we set null masks to the default mask
    else
    begin
      if ((State = dssSearch) or SearchAlways) and SearchWithMask then
        Self.SetEditMaskEx(DefaultMask_SpinEditEnh)
      else
        Self.EditMaskMode := eemNone;
    end;
    MaxLength := FieldMaxLength;
  end;
  EditingChanged(DataLink, DataSource);
  UpdateTabStop;
end;

procedure TIB_CustomSpinEditEnh.SysUpdateData;
begin
  if Assigned(Field) then
  begin
    if Modified then
    begin
      if not Validating then
        ValidateEdit;
      if IsEmpty then
        DataLink.FieldText := ''
      else
        with DataLink do
        begin
          if IsParamEdit then
          begin
            if (Field is TIB_ColumnInteger) or
              (Field is TIB_ColumnSmallint) then
               Field.AsInteger := Value
            else
               Field.AsString := Self.Value;
          end
          else if (State = dssSearch) or SearchAlways then
            SearchBuffer := Self.Value
          else
            case State of
            dssEdit,
            dssInsert:
              begin
                if (Field is TIB_ColumnInteger) or
                  (Field is TIB_ColumnSmallint) then
                   Field.AsInteger := Self.Value
                else
                   Field.AsString := Self.Value;
              end;
          end;
        end;
    end;
  end;
  Modified := false;
end;

function TIB_CustomSpinEditEnh.IsMaskStored: boolean;
begin
  // IBO does not store masks with controls, only with the data definition
  Result := false;
end;

function TIB_CustomSpinEditEnh.GetDBDisplayField: TObject;
begin
  Result := DataLink.Field;
end;

function TIB_CustomSpinEditEnh.IBG_GetAutoLabel: TObject;
begin
  Result := FAutoLabel;
end;

procedure TIB_CustomSpinEditEnh.DoAfterValidateEdit(var ErrCode: integer);
begin
  inherited DoAfterValidateEdit(ErrCode);
  if (ErrCode = E_EDITENH_NOERROR) and not DataLink.DataUpdating then
  begin
    DataLink.UpdateData;
    SysDataChange(nil);
  end;
end;

function TIB_CustomSpinEditEnh.EditCanModify: boolean;
begin
  if DataLink.IsParamEdit then
    Result := true
  else
    Result := DataLink.CanModify and
      (DataLink.DataSource.Modifying or DataLink.DataSource.AutoEdit or
      (DataLink.DataSource.AutoInsert and
      (DataLink.Dataset.Fields.RowState = rsNone)));
end;

function TIB_CustomSpinEditEnh.EditBeginModify: boolean;
begin
  if DataLink.IsParamEdit then
    Result := true
  else
    Result := DataLink.Modify;
end;

procedure TIB_CustomSpinEditEnh.Reset;
begin
  SysDataChange(nil);
  SelectAll;
end;

procedure TIB_CustomSpinEditEnh.Change;
begin
  if not MaskChanging then
    with DataLink do
    begin
      if (DataChangingLevel = 0) and
        (NeedToPost or (State = dssSearch) or IsParamEdit) then
      begin
        ControlIsModified := true;
        Modified := true;
      end;
    end;
  inherited Change;
end;

procedure TIB_CustomSpinEditEnh.CMEnter(var Message: TCMEnter);
begin
  inherited;
  SysDataChange(nil);
end;

procedure TIB_CustomSpinEditEnh.CMExit(var Message: TCMExit);
begin
  inherited;   // base class does a ValidateEdit
end;

end.
