{ Telesis_MaskEnhEditor.pas                       SearchTag: ~~EnhComponents~~

 Copyright 1999-2001 - Telesis Computing Pty Ltd.  Australia.
Contact: Geoff Worboys - geoff@telesiscomputing.com.au }

{: @summary MaskEnh design dialog - including samples and testing facility.
<br>
@desc This unit contains the dialog displayed when the user clicks on the
EditMask property of TEditEnh controls and derivatives.
<br>
<br>
<strong>History</strong><br>
<br>
01-Aug-2001 - Documentation.<br>
<br>
<br>
<strong>Author: </strong>Geoff Worboys - Telesis Computing Pty Ltd.<br>
<br>}
unit Telesis_MaskEnhEditor;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, Telesis_EditEnh, ComCtrls, Buttons,
  Grids, Telesis_MaskEnh;

type
  TMaskEnhEditorForm = class(TForm)
    BackingPnl: TPanel;
    OKBtn: TBitBtn;
    CancelBtn: TBitBtn;
    Pages: TPageControl;
    MaskTestPg: TTabSheet;
    MaskSamplePg: TTabSheet;
    MaskModeRG: TRadioGroup;
    MaskAttribGB: TGroupBox;
    DisplayBlankLbl: TLabel;
    TextBlankLbl: TLabel;
    DisplayBlankEB: TEditEnh;
    TextBlankEB: TEditEnh;
    DisplayBlankDefaultLbl: TLabel;
    TextBlankDefaultLbl: TLabel;
    MaskStringLbl: TLabel;
    MaskTesting1Bvl: TBevel;
    InputTestLbl: TLabel;
    MaskStringEB: TEditEnh;
    InputTestEB: TEditEnh;
    MaskTesting2Bvl: TBevel;
    TextValueLbl: TLabel;
    TextValueEB: TEditEnh;
    SetEditMaskExLbl: TLabel;
    SetEditMaskExEB: TEditEnh;
    SampleBackingPnl: TPanel;
    SampleSizerPnl: TPanel;
    SamplesSG: TStringGrid;
    TestSampleBtn: TButton;
    SampleInfoPnl: TPanel;
    NotePg: TTabSheet;
    SaveLiteralsCB: TCheckBox;
    SampleMaskStringPnl: TPanel;
    SampleMaskStringEB: TMemo;
    NotesEB: TMemo;
    SaveLiteralsLbl: TLabel;
    UserMaskEB: TEditEnh;
    UserMaskLbl: TLabel;
    procedure MaskInputChange(Sender: TObject);
    procedure SamplesSGSelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
    procedure SamplesSGSetEditText(Sender: TObject; ACol, ARow: Integer;
      const Value: String);
    procedure SampleMaskStringEBChange(Sender: TObject);
    procedure TestSampleBtnClick(Sender: TObject);
    procedure SamplesSGKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure SamplesSGEnter(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure OkCancelClick(Sender: TObject);
    procedure StartSizing(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure EndSizing(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure MoveSizing(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure InputTestEBChange(Sender: TObject);
  private
    { Private declarations }
    FInternalUpdate: integer;
    FCellChanged: boolean;
    FSamplesChanged: boolean;
    FPanelSizing: boolean;
    FDoReverseAfterShow: boolean;
    procedure ApplyMaskString;
    procedure ReverseEngineer;
    procedure InitSamplesGrid;
    procedure UpdateSampleFromGrid;
    procedure DoReadRegistry;
    procedure DoWriteRegistry;
    procedure DoRemoveOldRegistry;
  public
    { Public declarations }
    constructor Create( AOwner: TComponent ); override;
    destructor Destroy; override;
    procedure AssignMask( InitMask: TMaskEnh );

  public
    RegistryKey: string;
  end;

var
  MaskEnhEditorForm: TMaskEnhEditorForm;

implementation

{$R *.DFM}

uses Registry;

const
    // Default samples
SamplesDefaultCount = 11;
SamplesDefault: array[1..SamplesDefaultCount] of string = (
    '  Last Used Mask List  | |::SL::|',
    'Credit Card (Standard)|5555 4444 3333 2222|::SM::|"0000 0000 0000 0000;1;_"',
    'Australian Postcode (Standard)|2058|::SM::|"0000;1;_"',
    'Formatted Text (Enhanced)|"abc-def-ghi"|::EM::|"[""ccc[-ccc[-ccc]"";0;_"',
    'Password Entry (Enhanced)|********|::EM::|"cccccccccccc;0;_"',
    'Currency (Numeric)|$123.45|::NM::|"!\$-999],999],990!.00;0;0;0"',
    'Floating Point (Numeric)|3,210.123,345|::NM::|"!+999],999],999],990!.099[,999[,999;0;0;0"',
    'Date and Time (Numeric)|12/12/1999 12:00|::NM::|"!90!/!90!/!0000!\ !90!:!00;0;_;0"',
    'Australian Telephone (Standard List)|(02) 1234 5678|::SL::|"(C0) 0000' +
        ' 0000;0;_","00*=(C000) cccccccccccccccccccccccc;0;_","04*=(C000) ' +
        '000 000;0;_","01*=(C00) 000 000;0;_","13*=C0 0000;0;_","18*=(C000' +
        ') 000 000;0;_","19*=(C000) cccccccccccccccccccccccc;0;_","#*=Cccc' +
        'cccccccccccccccccccccccccc;0;_"',
    'US Zip Code (Enhanced List)|12345-67890|::EL::|"00000[-99999;1;_","??????*=00000-00000;1;_"',
    'Hex Entry (Standard)|0A3F-FE0A|::EM::|">XXXX[-xxxx;0;_;;X[0-9A-F]"'
    );

var
    SampleStringList: TStringList;

constructor TMaskEnhEditorForm.Create( AOwner: TComponent );
var
    i: integer;
begin
inherited;
FInternalUpdate := 0;
FCellChanged := false;
FSamplesChanged := false;
FPanelSizing := false;
FDoReverseAfterShow := false;
RegistryKey := ''; // default to blank, no save/restore to/from registry
SampleStringList := nil;
SampleStringList := TStringList.Create;
for i := 1 to SamplesDefaultCount do begin
    SampleStringList.Add( SamplesDefault[i] );
    end;
end;

destructor TMaskEnhEditorForm.Destroy;
begin
SampleStringList.Free;
SampleStringList := nil;
inherited;
end;

procedure TMaskEnhEditorForm.AssignMask( InitMask: TMaskEnh );
begin
if assigned( InitMask ) and (InitMask.Mode <> eemNone) then begin
    InputTestEB.SetEditMaskEx( InitMask.MaskEx );
    end
else begin
    InputTestEB.EditMaskMode := eemNone;
    InputTestEB.EditMask := '';
    end;
FDoReverseAfterShow := true;
end;

procedure TMaskEnhEditorForm.ApplyMaskString;
var
    db, tb: char;
    sl: boolean;
begin
InputTestEB.Clear;
if SaveLiteralsCB.Checked then
  sl := true
else sl := false;
if Length(DisplayBlankEB.Text) > 0 then
  db := DisplayBlankEB.Text[1]
else db := eeMskDefaultBlank;
if Length(TextBlankEB.Text) > 0 then
  tb := TextBlankEB.Text[1]
else tb := eeMskRealBlank;

InputTestEB.EditMask := MakeMaskEnhString( MaskStringEB.Text, sl,
    db, tb, UserMaskEB.Text );
end;

procedure TMaskEnhEditorForm.ReverseEngineer;
var
    tmpStr, tmpStr2: string;
    DspBlank, TxtBlank: char;
    SaveLit: boolean;
begin
DspBlank := eeMskDefaultBlank;
TxtBlank := eeMskRealBlank;
SaveLit := false;
Inc( FInternalUpdate );
try
    // Mask mode...
    MaskModeRG.ItemIndex := integer(InputTestEB.EditMaskMode);
    if (InputTestEB.EditMaskMode > eemNone) and
        (Length(InputTestEB.EditMask) > 0) then begin
        // Mask string and attributes...
        if ((InputTestEB.EditMaskMode = eemMaskList) or
            (InputTestEB.EditMaskMode = eemMaskEnhList)) and
            assigned(InputTestEB.Mask) then begin
            MaskStringEB.Text := '';
            SamplesSG.Row := 1;
            SampleMaskStringEB.Text := InputTestEB.Mask.MaskList.Text;
            SamplesSG.Cells[2,1] := MaskEnhModeToString( InputTestEB.EditMaskMode );
            UpdateSampleFromGrid;
            end
        else begin
            SplitMaskEnhString( InputTestEB.EditMask, tmpStr, SaveLit,
                DspBlank, TxtBlank, tmpStr2 );
            MaskStringEB.Text := tmpStr;
            SaveLiteralsCB.Checked := SaveLit;
            DisplayBlankEB.Text := DspBlank;
            TextBlankEB.Text := TxtBlank;
            UserMaskEB.Text := tmpStr2;
            end;
        if (InputTestEB.EditMaskMode = eemMaskList) or
            (InputTestEB.EditMaskMode = eemMaskEnhList) then begin
            if assigned( InputTestEB.Mask ) then
                SetEditMaskExEB.Text := MaskEnhModeToString( InputTestEB.EditMaskMode ) +
                    InputTestEB.Mask.MaskList.CommaText;
            end
        else begin
            SetEditMaskExEB.Text := MaskEnhModeToString( InputTestEB.EditMaskMode ) +
                InputTestEB.EditMask;
            end;
        end
    else begin
        UserMaskEB.Text := '';
        MaskStringEB.Text := '';
        SetEditMaskExEB.Text := '';
        end;
finally
    Dec( FInternalUpdate );
    end
end;

procedure TMaskEnhEditorForm.MaskInputChange(Sender: TObject);
begin
if FInternalUpdate = 0 then begin
    Inc( FInternalUpdate );
    try
        if (Sender = MaskModeRG) then begin
            InputTestEB.Clear;
            InputTestEB.EditMask := '';
            InputTestEB.EditMaskMode := TMaskEnhMode(MaskModeRG.ItemIndex);
            ApplyMaskString;
            ReverseEngineer;
            end
        else if (Sender = SaveLiteralsCB) or
            (Sender = DisplayBlankEB) or
            (Sender = TextBlankEB) or
            (Sender = UserMaskEB) or
            (Sender = MaskStringEB) then begin
            ApplyMaskString;
            ReverseEngineer;
            end
        else if (Sender = TextValueEB) then begin
            InputTestEB.Text := TextValueEB.Text;
            end
        else if (Sender = SetEditMaskExEB) then begin
            InputTestEB.SetEditMaskEx( SetEditMaskExEB.Text );
            ReverseEngineer;
            end;
    finally
        Dec( FInternalUpdate );
        end;
    end;
if (Sender = MaskModeRG) then begin
    if (MaskModeRG.ItemIndex = 2) or (MaskModeRG.ItemIndex = 4) then begin
        MaskAttribGB.Enabled := false;
        MaskStringEB.Enabled := false;
        SaveLiteralsLbl.Enabled := false;
        DisplayBlankLbl.Enabled := false;
        TextBlankLbl.Enabled := false;
        MaskStringLbl.Enabled := false;
        end
    else begin
        MaskAttribGB.Enabled := true;
        MaskStringEB.Enabled := true;
        SaveLiteralsLbl.Enabled := true;
        DisplayBlankLbl.Enabled := true;
        TextBlankLbl.Enabled := true;
        MaskStringLbl.Enabled := true;
        end
    end;
end;

procedure TMaskEnhEditorForm.InitSamplesGrid;
var
    i, p: integer;
    tmpStr: string;
begin
with SamplesSG do begin
    RowCount := SampleStringList.Count + 1;
    ColWidths[0] := 200;
    ColWidths[1] := 200;
    ColWidths[2] := 40;
    Cells[0,0] := 'Description';
    Cells[1,0] := 'Sample Value';
    Cells[2,0] := 'Mode';
    for i := 0 to SampleStringList.Count -1 do begin
        tmpStr := SampleStringList.Strings[i];
        p := Pos( '|', tmpStr );
        Cells[0,i+1] := Copy( tmpStr, 1, p - 1 );
        Delete( tmpStr, 1, p );
        p := Pos( '|', tmpStr );
        Cells[1,i+1] := Copy( tmpStr, 1, p - 1 );
        Delete( tmpStr, 1, p );
        p := Pos( '|', tmpStr );
        Cells[2,i+1] := Copy( tmpStr, 1, p - 1 );
        Delete( tmpStr, 1, p );
        end;
    FCellChanged := false;
    SamplesSG.Row := RowCount - 1;
    SamplesSG.Row := 1;
    end;
end;

procedure TMaskEnhEditorForm.UpdateSampleFromGrid;
var
    EMode: TMaskEnhMode;
begin
if FCellChanged and (SamplesSG.Row > 0) then with SamplesSG do begin
    FSamplesChanged := true;
    if not StringToMaskEnhMode( UpperCase(Cells[2,Row]), EMode ) then
        Cells[2,Row] := ''
    else Cells[2,Row] := MaskEnhModeToString( EMode );
    SampleStringList.Strings[Row - 1] :=
        Cells[0,Row] + '|' + Cells[1,Row] + '|' +
        Cells[2,Row] + '|' + SampleMaskStringEB.Lines.CommaText;
    FCellChanged := false;
    end;
end;


procedure TMaskEnhEditorForm.SamplesSGSelectCell(Sender: TObject; ACol,
  ARow: Integer; var CanSelect: Boolean);
var
    i, p: integer;
    tmpStr: string;
begin
UpdateSampleFromGrid;
i := ARow - 1; // allow for title row
if (i >= 0) and (i < SampleStringList.Count) then begin
    with SamplesSG do begin
        tmpStr := SampleStringList.Strings[i];
        p := Pos( '|', tmpStr );
        Cells[0,ARow] := Copy( tmpStr, 1, p - 1 );
        Delete( tmpStr, 1, p );
        p := Pos( '|', tmpStr );
        Cells[1,ARow] := Copy( tmpStr, 1, p - 1 );
        Delete( tmpStr, 1, p );
        p := Pos( '|', tmpStr );
        Cells[2,ARow] := Copy( tmpStr, 1, p - 1 );
        Delete( tmpStr, 1, p );
        SampleMaskStringEB.Lines.CommaText := tmpStr;
        FCellChanged := false;
        end;
    end;
end;

procedure TMaskEnhEditorForm.SamplesSGSetEditText(Sender: TObject; ACol,
  ARow: Integer; const Value: String);
begin
FCellChanged := true;
end;

procedure TMaskEnhEditorForm.SampleMaskStringEBChange(Sender: TObject);
begin
FCellChanged := true;
end;

procedure TMaskEnhEditorForm.TestSampleBtnClick(Sender: TObject);
var
    tmpStr: string;
    EMode: TMaskEnhMode;
begin
if SamplesSG.Row > 0 then begin
    UpdateSampleFromGrid;
    Inc( FInternalUpdate );
    try
        tmpStr := SamplesSG.Cells[2,SamplesSG.Row];
        if StringToMaskEnhMode( tmpStr, EMode ) then begin
            if (EMode = eemMaskList) or (EMode = eemMaskEnhList) then begin
                tmpStr := tmpStr + SampleMaskStringEB.Lines.CommaText;
                end
            else if EMode <> eemNone then begin
                if SampleMaskStringEB.Lines.Count > 0 then
                    tmpStr := tmpStr + SampleMaskStringEB.Lines.Strings[0];
                end;
            InputTestEB.SetEditMaskEx( tmpStr );
            end;
    finally
        Dec( FInternalUpdate );
        end;
    ReverseEngineer;
    Pages.ActivePage := MaskTestPg;
    end;
end;

procedure TMaskEnhEditorForm.SamplesSGKeyDown(Sender: TObject;
  var Key: Word; Shift: TShiftState);
var
    tmpRow: integer;
begin
if Key = VK_RETURN then begin
    // If editing then simply close the editor
    if SamplesSG.EditorMode then begin
        UpdateSampleFromGrid;
        SamplesSG.EditorMode := false;
        SamplesSG.Options := SamplesSG.Options + [goRowSelect];
        end
    // If not editing then select the highlighted mask
    else begin
        TestSampleBtnClick( Sender );
        end;
    end
else if Key = VK_F2 then begin
    // If editing then simply close the editor
    if SamplesSG.EditorMode then begin
        UpdateSampleFromGrid;
        SamplesSG.EditorMode := false;
        SamplesSG.Options := SamplesSG.Options + [goRowSelect] - [goTabs,goAlwaysShowEditor];
        end
    // If not editing then select the highlighted mask
    else begin
        SamplesSG.Options := SamplesSG.Options - [goRowSelect] + [goTabs,goAlwaysShowEditor];
        SamplesSG.EditorMode := true;
        end;
    end
else if ssCtrl in Shift then begin
    if Key = VK_INSERT then begin
        if SamplesSG.Row >= 2 then begin
            tmpRow := SamplesSG.Row - 1;
            SampleStringList.Insert( tmpRow, '' );
            InitSamplesGrid;
            SamplesSG.Row := tmpRow + 1;
            FSamplesChanged := true;
            end;
        end
    else if Key = VK_DELETE then begin
        if SamplesSG.Row >= 2 then begin
            tmpRow := SamplesSG.Row - 1;
            SampleStringList.Delete( tmpRow );
            InitSamplesGrid;
            if tmpRow < SamplesSG.RowCount - 1 then
                SamplesSG.Row := tmpRow + 1
            else SamplesSG.Row := tmpRow;
            FSamplesChanged := true;
            end;
        end;
    end;
end;

procedure TMaskEnhEditorForm.SamplesSGEnter(Sender: TObject);
begin
// Always enable row select on enter
SamplesSG.EditorMode := false;
SamplesSG.Options := SamplesSG.Options + [goRowSelect] - [goTabs,goAlwaysShowEditor];
end;

procedure TMaskEnhEditorForm.FormShow(Sender: TObject);
begin
// If registry key defined then try and restore from registry...
if Length( RegistryKey ) > 0 then
    DoReadRegistry;
InitSamplesGrid;
Pages.ActivePage := MaskTestPg;
if FDoReverseAfterShow then begin
    ReverseEngineer;
    FDoReverseAfterShow := false;
    end;
end;

procedure TMaskEnhEditorForm.DoReadRegistry;
var
    Reg: TRegistry;
    i: integer;
    tmpStr, tmpValID: string;
    tmpStrs: TStringList;
begin
Reg := TRegistry.Create;
tmpStrs := TStringList.Create;
try
    Reg.RootKey := HKEY_CURRENT_USER;
    if Reg.OpenKey( RegistryKey, false ) then begin
        tmpStr := '';
        i := 0;
        while true do begin
            tmpValID := 'MaskSample' + IntToStr(i);
            if Reg.ValueExists( tmpValID ) then begin
                tmpStr := Reg.ReadString( tmpValID );
                tmpStrs.Add( tmpStr );
                end
            else break;
            Inc( i );
            end;
        Reg.CloseKey();
        end;
    if tmpStrs.Count > 0 then
        SampleStringList.Assign( tmpStrs );
finally
    tmpStrs.Free;
    Reg.Free;
    end;
end;

procedure TMaskEnhEditorForm.DoWriteRegistry;
var
    Reg: TRegistry;
    i: integer;
begin
Reg := TRegistry.Create;
try
    Reg.RootKey := HKEY_CURRENT_USER;
    if Reg.OpenKey( RegistryKey, true ) then begin
        for i := 0 to SampleStringList.Count - 1 do begin
            Reg.WriteString( 'MaskSample' + IntToStr(i), SampleStringList.Strings[i] );
            end;
        Reg.CloseKey();
        end;
finally
    Reg.Free;
    end;
end;

procedure TMaskEnhEditorForm.DoRemoveOldRegistry;
var
    Reg: TRegistry;
    i: integer;
    tmpStr, tmpValID: string;
begin
Reg := TRegistry.Create;
try
    Reg.RootKey := HKEY_CURRENT_USER;
    if Reg.OpenKey( RegistryKey, false ) then begin
        tmpStr := '';
        i := 0;
        while true do begin
            tmpValID := 'MaskSample' + IntToStr(i);
            if Reg.ValueExists( tmpValID ) then begin
                Reg.DeleteValue( tmpValID );
                end
            else break;
            Inc( i );
            end;
        Reg.CloseKey();
        end;
finally
    Reg.Free;
    end;
end;

procedure TMaskEnhEditorForm.OkCancelClick(Sender: TObject);
begin
if Sender = OKBtn then begin
    UpdateSampleFromGrid;
    // If registry key defined then try and save to registry...
    if FSamplesChanged and (Length( RegistryKey ) > 0) then begin
        DoRemoveOldRegistry;
        DoWriteRegistry;
        end;
    ModalResult := mrOk;
    end;
end;

procedure TMaskEnhEditorForm.StartSizing(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
FPanelSizing := true;
end;

procedure TMaskEnhEditorForm.EndSizing(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
FPanelSizing := false;
end;

procedure TMaskEnhEditorForm.MoveSizing(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
var
    i: integer;
begin
if FPanelSizing then begin
    if Sender = SampleSizerPnl then begin
        i := SampleMaskStringPnl.Height - Y;
        if i < 20 then i := 20;
        if i > SampleBackingPnl.ClientHeight - 40 then
            i := SampleBackingPnl.ClientHeight - 40;
        SampleMaskStringPnl.Height := i;
        end;
    end;
end;

procedure TMaskEnhEditorForm.InputTestEBChange(Sender: TObject);
begin
Inc( FInternalUpdate );
try
    TextValueEB.Text := InputTestEB.Text;
finally
    Dec( FInternalUpdate );
    end;
end;

end.